/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2007, 2008
 *               Bioinformatics Research Center,
 *               University of Aarhus, Denmark
 *               Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#ifndef PORTABLE_ARCHIVES_COMMON_HH_INCLUDED
#define PORTABLE_ARCHIVES_COMMON_HH_INCLUDED

#include <boost/detail/endian.hpp>



/** \file
 * \brief Contains various method for handeling endianness.
 *
 * Only handles little and big endian systems.
 */



namespace BiRC {
    namespace SNPFile {
        /**
         * \brief Checks the endianness of this system.
         *
         * \returns true iff this system is bigendian.
         */
        inline
        bool isBigEndian()
        {
#ifdef BOOST_BIG_ENDIAN
            return true;
#else
            return false;
#endif
        }



        /**
         * \brief Reverses the order of the bytes in the input.
         *
         * \param t   Data to reverse.
         * \returns   The input t in reverse byte order.
         */
        template <typename T>
        T reverse(T t)
        {
            union {
                T v;
                char a[sizeof(T)];
            } u;
            u.v = t;

            for(size_t i = 0, j = sizeof(T) - 1; i < j; i++, j--)
            {
                char tmp = u.a[i];
                u.a[i] = u.a[j];
                u.a[j] = tmp;
            }

            return u.v;
        }



        /**
         * \brief Reverses the order of the bytes in the input iff
         * this system is big endian.
         *
         * Works as the identity function on little endian systems and
         * as the above reverse function on big endian systems.
         *
         * \param t    Data to maybe reverse.
         * \returns    The input maybe reversed.
         */
        template <typename T> inline
        T maybeReverse(T t)
        {
            if(isBigEndian())
                return reverse(t);
            else
                return t;
        }



        /**
         * \brief Changes data from host format to little endian format.
         *
         * \param t   Data to transform.
         * \returns   Data in little endian format.
         */
        template <typename T> inline
        T toLittleEndian(T t)
        {
            return maybeReverse(t);
        }



        /**
         * \brief Changes data from little endian format to host format.
         *
         * \param t   Data to transform.
         * \returns   Data in host format.
         */
        template <typename T> inline
        T fromLittleEndian(T t)
        {
            return maybeReverse(t);
        }



        /**
         * \brief Typetrait for serialization and deserialization.
         *
         * Specializations of this struct should have the two functions:
         * static void load(InputArchive &ia, T &t)
         * static void save(OutputArchive &oa, const T &t)
         */
        template<typename T> struct SerializationTrait;
    }
}



#endif
