/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus, Denmark
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#undef NDEBUG			// enable asserts always

#include "archives.hh"
#include "type_trait.hh"


#include <cassert>
#include <string>
#include <sstream>
#include <iostream>
#include <map>



using namespace BiRC::SNPFile;



///Used for testing serialization of objects/pointers to objects.
class IntrusiveClass
{
    int i;

public:
    /// Default constructor. Required by serialization.
    IntrusiveClass() : i(0) {}
    /// Constructor with initial value.
    IntrusiveClass(int i) : i(i) {}
    /// Getter. Allows reading of the value of this object.
    int getI() const { return i; }

    /**
     * \brief Does both serialization and unserialization.
     *
     * Serializes this class if archive argument is an output archive.
     * Unserializes this class if archive argument is an input archive. 
     *
     * \param ar  Archive to read from / write to.
     */
    template<typename Archive>
    void serialize(Archive & ar)
    {
        ar | i;
    }
};
BIRC_SNPFILE_INTRUSIVE_SERIALIZATION(IntrusiveClass);


///Used for testing serialization of objects/pointers to objects.
struct NonIntrusiveStruct
{
    int i;
};

template<typename Archive>
void serialize(Archive & ar, NonIntrusiveStruct &s)
{
    ar | s.i;
}
BIRC_SNPFILE_NONINTRUSIVE_SERIALIZATION(NonIntrusiveStruct);

///Used for testing serialization of objects/pointers to objects.
class NonIntrusiveClass
{
    int i;
public:
    NonIntrusiveClass() : i(12) {};
    int getI() const { return i; }
    void setI(int j) { i = j; }
};

namespace BiRC { 
    namespace SNPFile {
	template <>
	struct SerializationTrait<NonIntrusiveClass> {
	    static void load(InputArchive &ia, NonIntrusiveClass &nic)
	    {
		int i;
		ia >> i;
		nic.setI(i);
	    }
	    static void save(OutputArchive &oa, const NonIntrusiveClass &nic)
	    {
		oa << nic.getI();
	    }
	};
    }
}

/// Type used in testing the archieval of maps.
typedef std::map<std::string, double> test_map_t;



/**
 * \brief Exercises and tests the archives.
 *
 * \returns 0 to indicate program success.
 */
int main()
{
    //Test basic portability functions
    uint32_t i = 0x12345678;
    uint32_t expectedI = 0x78563412;
    assert(reverse(i) == expectedI);

    union {
        float f;
        uint32_t i;
    } u;
    u.i = i;
    u.f = reverse(u.f);
    assert(u.i == expectedI);

    char expectedC[] = {0x78, 0x56, 0x34, 0x12};

    union {
        char a[sizeof(i)];
        uint32_t j;
    } littleEndian;

    littleEndian.j = toLittleEndian(i);

    for(size_t j = 0; j < sizeof(expectedC); j++)
        assert(expectedC[j] == littleEndian.a[j]);




    //Test archive
    std::string string;

    int8_t   ei8   = -0x12;
    uint8_t  eui8  = 0x12;
    int16_t  ei16  = -0x1234;
    uint16_t eui16 = 0x1234;
    int32_t  ei32  = -0x12345678;
    uint32_t eui32 = 0x12345678;
    int64_t  ei64  = -((int64_t(0x12345678) << 32) + int64_t(0x90abcdef));
    uint64_t eui64 = (int64_t(0x12345678) << 32) + int64_t(0x90abcdef);
    float ef = 3.14;
    double ed = 2.17;

    int32_t eToBeCast = 31337;

    std::string es = "Hello World";
    const IntrusiveClass em(7);
    NonIntrusiveStruct enis;
    NonIntrusiveClass enic;
    test_map_t emap;
    const test_map_t &const_emap = emap;
    emap["Foo"] = 0.1234;
    emap["Bar"] = 0.5678;


    {
        OutputArchive oa(string);
        oa << ei8 << eui8 << ei16 << eui16;
        oa << ei32 << eui32 << ei64 << eui64;
        oa << ef << ed;
        oa << es;
        oa << em;
	oa << enis;
	oa << enic;
        oa << eToBeCast;
        oa << const_emap;
    }

    int8_t   ri8 ;
    uint8_t  rui8 ;
    int16_t  ri16;
    uint16_t rui16;
    int32_t  ri32;
    uint32_t rui32;
    int64_t  ri64;
    uint64_t rui64;
    int64_t afterCast;
    float rf;
    double rd;
    std::string rs;
    IntrusiveClass rm;
    NonIntrusiveStruct rnis;
    NonIntrusiveClass rnic;
    test_map_t rmap;

    {
        InputArchive ia(string);
        ia >> ri8 >> rui8 >> ri16 >> rui16;
        ia >> ri32 >> rui32 >> ri64 >> rui64;
        ia >> rf >> rd;
        ia >> rs;
        ia >> rm;
	ia >> rnis;
	ia >> rnic;
        ia >> afterCast;
        ia >> rmap;
    }

    assert( ei8  ==  ri8 );
    assert(eui8  == rui8 );
    assert( ei16 ==  ri16);
    assert(eui16 == rui16);
    assert( ei32 ==  ri32);
    assert(eui32 == rui32);
    assert( ei64 ==  ri64);
    assert(eui64 == rui64);
    assert(int64_t(eToBeCast) == afterCast);

    assert(es == rs);
    assert(em.getI() == rm.getI());
    assert(enis.i == rnis.i);
    assert(enic.getI() == rnic.getI());

    test_map_t::const_iterator it;

    for(it = emap.begin(); it != emap.end(); ++it)
        assert(it->second == rmap[it->first]);
}
