/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus, Denmark
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include "matrix.hh"
#include "array_matrix.hh"

/** \file
 * \brief Implementation of ArrayMatrixData.
 */

namespace BiRC
{
    namespace SNPFile
    {

        /**
         * \brief Implementation of MatrixData based on an Array for
         * internal representation.
         *
         * This class actually stores the haplotype data, but the data is
         * usually manipulated through the MatrixView class, that represents
         * sub-matrices.
         *
         * \see MatrixView
         */
        class ArrayMatrixData : public MatrixData
        {
            /// Flag indicating if the data is phased or unphased
            bool i_phased;

            /// Copy-constructor. Not implemented. Do not copy array
            /// matrices.
            ArrayMatrixData(ArrayMatrixData &rhs);

            /// Assignment-operator. Not implemented. Do not assign
            /// array matrices.
            ArrayMatrixData &operator= (ArrayMatrixData &rhs);

            /**
             * \brief Resize matrix to new dimensions.
             *
             * WARNING: resizing does *not* just expand or decrease the
             * dimensions -- the data will not (generally) be preserved by a
             * resizing.
             *
             * \param noRows	Number of rows in the resized matrix.
             * \param noCols	Number of columns in the resized matrix.
             */
            virtual void resizeData(int noRows, int noCols);

            /// Is the data phased?
            virtual bool isPhased() const;
            /// Set phased flag.
            virtual void setPhased(bool phased);

	    virtual void readHiding();
	    virtual void storeHiding();

        public:
            /**
             * \brief Construct a new matrix with the specified dimensions.
             *
             * \param noRows	Number of rows for the new matrix.
             * \param noCols	Number of columns for the new matrix.
             * \param phased	Flag indicating whether the sequences are phased 
             *                  or not.
             */
            ArrayMatrixData(int noRows = 0, int noCols = 0, 
			    bool phased = false);

            /// Virtual destructor.
            virtual ~ArrayMatrixData();

        };


        ArrayMatrixData::ArrayMatrixData(int noRows, int noCols, 
                                         bool phased)
            : MatrixData(noRows, noCols),
              i_phased(phased)
        {
            i_data = new allele_t[size_t(noRows) * size_t(noCols)];
        }



        ArrayMatrixData::~ArrayMatrixData()
        {
            delete[] i_data;
        }



        bool
        ArrayMatrixData::isPhased() const
        {
            return i_phased; 
        }



        void
        ArrayMatrixData::setPhased(bool phased) 
        {
            i_phased = phased;
        }



        void
        ArrayMatrixData::resizeData(int noRows, int noCols)
        {
	    // FIXME: this isn't exception safe!
            delete[] i_data;
            i_data = new allele_t[size_t(noRows) * size_t(noCols)];
        }

	void
	ArrayMatrixData::readHiding()
	{}			// nop
	
	void
	ArrayMatrixData::storeHiding()
	{}			// nop


	Matrix newArrayMatrix(int noRows, int noCols, bool phased)
	{
	    return Matrix(new ArrayMatrixData(noRows, noCols, phased));
	}

    }
}
