/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus, Denmark
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include "snpfile.hh"
#include "matrix.hh"
#include "file_matrix.hh"
#include "metadata_access.hh"
using namespace BiRC::SNPFile;

#include <cstdlib>
#include <cerrno>
#include <fstream>
#include <boost/program_options.hpp>
#include <boost/lexical_cast.hpp>
//#include <boost/lambda/lambda.hpp>
//#include <boost/lambda/if.hpp>
//using namespace boost::lambda;


/// Exception thrown if loading haplotypes failed.
struct ParseError : std::logic_error {
    ParseError() : std::logic_error("Error parsing haplotype data.") {}
};


static void
readHaplotypes(std::istream &is, Matrix &matrix)
{
    //std::vector< std::vector<int> > haplotypes;
    char status;
    int position;
    int marker_count=0;
    
    matrix.resize(matrix.noRows()*2, matrix.noCols());

    while (is >> status >> position)
	{
	    std::string line;
	    getline(is, line);
	    if (line.empty()) continue; // skip empty line
	    std::stringstream ss(line);
	    int k;
	    for(int i=0; ss >> k; i++) {
		matrix(i,marker_count) = k;
	    }
	    marker_count++;
	}
    /*
    if (haplotypes.empty()) throw ParseError();
    int noHaplotypes = haplotypes[0].size();
    int noMarkers    = haplotypes.size();

    matrix.resize(noHaplotypes, noMarkers);
    for (int i = 0; i < noHaplotypes; ++i)
	{
	    std::cout << i << std::endl;
	    for (int j = 0; j < noMarkers; ++j)
		matrix(i,j) = haplotypes[j][i];
	}

    */
    matrix.setPhased(true);
}

namespace {
    template <typename Itr>
    bool isSorted(Itr begin, Itr end)
    {
        if (begin+1 >= end) return true;
        for (Itr i = begin; i < end-1; ++i)
            if (*i > *(i+1)) return false;
        return true;
    }
}


namespace options {
    bool verbose = false;
    std::string haplotypesFileName;
    std::string snpFileName;
}


namespace po = boost::program_options;
static void printUsage(const char *progname,
                       std::ostream &os,
                       po::options_description &options)
{
    os << "Usage: " << progname
       << " [options] phased-haplotype-file snp-file"
       << std::endl
       << options << std::endl;
}

int main(int argc, char** argv)
{
    po::options_description visible("Allowed options");
    po::options_description hidden("Allowed options");
    po::options_description all("Allowed options");
    po::variables_map vm;
    visible.add_options()
        ("help,h", "Produce help message.")
        ("verbose,v","Turns on verbose output.")
        ;
    hidden.add_options()
        ("haplotypes",
         po::value<std::string>(&options::haplotypesFileName),
         "File containing marker data.")
        ("snpfile",
         po::value<std::string>(&options::snpFileName),
         "The SNPFile to convert.")
        ;
    all.add(visible).add(hidden);

    po::positional_options_description p;
    p.add("haplotypes", 1);
    p.add("snpfile", 1);
    p.add("output", 1);

    try {
        po::store(po::command_line_parser(argc, argv)
                  .options(all).positional(p).run(), 
                  vm, false);
        po::notify(vm);
    } catch (std::exception &e) {
        std::cerr << "Error parsing options: " << e.what()
                  << std::endl;
        printUsage(argv[0], std::cerr, visible);
        return EXIT_FAILURE;
    }

    if (vm.count("help")) {
        printUsage(argv[0], std::cout, visible);
        return EXIT_SUCCESS;
    }
    options::verbose = vm.count("verbose") > 0;

    // checking options
    if (options::haplotypesFileName.empty())
	{
  	    std::cerr << "Error: No haplotype file specified."
                  << std::endl << std::endl;
	    printUsage(argv[0], std::cerr, visible);
	    return EXIT_FAILURE;
	}
    if (options::snpFileName.empty())
	{
  	    std::cerr << "Error: No SNP file specified."
		      << std::endl << std::endl;
	    printUsage(argv[0], std::cerr, visible);
	    return EXIT_FAILURE;
	}


    // checking files
    std::ifstream haplotypesFile(options::haplotypesFileName.c_str());
    if (!haplotypesFile)
	{
	    std::cerr << "Error opening "
                  << options::haplotypesFileName << std::endl;
	    return EXIT_FAILURE;
	}

    try
    {
        SNPFile input(options::snpFileName, true);
	Matrix genotypes(newReadWriteFileMatrix(input));

        if (genotypes.isPhased())
        {
            std::cerr << "SNPFile indicates that it is already phased. "
                      << "Phasing it again might not be"
                      << std::endl
                      << "sensible. At least, I don't know how to interpret "
                      << "this. "
                      << std::endl
                      << "If you want this feature, email me and I'll take a "
                      << "look at the problem."
                      << std::endl << std::endl;
            return EXIT_FAILURE;
        }

	if(!input.containsMetadata("column_missing")){
	    std::vector<double> column_missing(genotypes.noCols());
	    for(int i = 0; i<genotypes.noCols(); i++){
		int count  = 0;
		for(int j = 0; j<genotypes.noRows(); j++){
		    if(genotypes(j, i)==MISSING) count++;
		}
		column_missing[i] = ((double) count)/genotypes.noRows();
		
	    }
	    setMetadata(input, "column_missing", column_missing);	    
	}
	
	if(!input.containsMetadata("row_missing")){
	    std::vector<double> row_missing(genotypes.noRows());
	    for(int i = 0; i<genotypes.noRows(); i++){
		int count  = 0;
		for(int j = 0; j<genotypes.noCols(); j++){
		    if(genotypes(i, j)==MISSING) count++;
		}
		row_missing[i] = ((double) count)/genotypes.noCols();
		
	    }
	    setMetadata(input, "row_missing", row_missing);	    
	}
	
	readHaplotypes(haplotypesFile, genotypes);
	
    }
    catch(std::exception &e)
    {
        std::cerr << "Error: " << e.what() << std::endl;
    }

    return EXIT_SUCCESS;
}
