/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2007, 2008
 *               Bioinformatics Research Center,
 *               University of Aarhus, Denmark
 *               Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#ifndef BYTE_ORDER_HANDLER_HH_INCLUDED
#define BYTE_ORDER_HANDLER_HH_INCLUDED

#include <arpa/inet.h>
#include <cstring>



/** \file
 * \brief Declarations of WriteOrderHandler and ReadOrderHandler.
 *
 * Stream for writing and reading from a stream portably.  These are
 * used for portable low-level file I/O.
 */



namespace BiRC {
    namespace SNPFile {

        /**
         * \brief Streams to a char array, while handling byte ordering issues.
         */
        class WriteOrderHandler
        {
            /// Where to write the next byte.
            char *i_position;



        public:
            /**
             * \brief          Construct new WriteOrderHandler writing
             *                 its output to a given position.
             *
             * \param position Where output should be placed.
             */
            WriteOrderHandler(char *position)
                : i_position(position)
            {}



            /**
             * \brief          Resets destination.
             *
             * \param position New destination
             */
            void setPosition(char *position)
            {
                i_position = position;
            }

            /**
             * \brief          Retrieves current destination.
             *
             * \returns        Current destination
             */
            char *getPosition() const
            {
                return i_position;
            }



            /// Write unsigned integer
            WriteOrderHandler &operator<<(uint32_t i);

            /// Write signed integer
            WriteOrderHandler &operator<<(int32_t i);

            /// Write byte
            WriteOrderHandler &operator<<(char c);

            /// Write string
            WriteOrderHandler &operator<<(const char *str);
        };



        /**
         * \brief Streams from a char array, while handling byte ordering issues.
         */
        class ReadOrderHandler
        {
            /// Where to write the next byte.
            const char *i_position;



        public:
            /**
             * \brief          Construct new ReadOrderHandler writing
             *                 its output to a given position.
             *
             * \param position Where output should be placed.
             */
            ReadOrderHandler(const char *position)
                : i_position(position)
            {}



            /**
             * \brief          Resets destination.
             *
             * \param position New destination
             */
            void setPosition(const char *position)
            {
                i_position = position;
            }

            /**
             * \brief          Retrieves current destination.
             *
             * \returns        Current destination
             */
            const char *getPosition()
            {
                return i_position;
            }



            /// Read unsigned integer
            ReadOrderHandler &operator>>(uint32_t &i);

            /// Read signed integer
            ReadOrderHandler &operator>>(int32_t &i);

            /// Read byte
            ReadOrderHandler &operator>>(char &c);

            /// Read string
            ReadOrderHandler &operator>>(char *dest);
        };
    }
}



#endif
