/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus, Denmark
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include "snpfile.hh"
#include "matrix.hh"
#include "file_matrix.hh"
#include "metadata_access.hh"
using namespace BiRC::SNPFile;

#include <cstdlib>
#include <cerrno>
#include <fstream>
#include <boost/program_options.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/if.hpp>
using namespace boost::lambda;




/// Exception thrown if positions read in are not in increasing order.
struct UnsortedPositions : std::logic_error {
    UnsortedPositions() : std::logic_error("Positions must be sorted.") {}
};

/// Exception thrown if loading haplotypes failed.
struct ParseError : std::logic_error {
    ParseError() : std::logic_error("Error parsing haplotype data.") {}
};



static void
readHaplotypes(std::istream &is, Matrix &matrix)
{
    std::vector< std::vector<int> > haplotypes;
    while (is)
	{
	    std::string line;

	    getline(is, line);
	    if (line.empty()) continue; // skip empty line

	    std::vector<int> haplotype(line.size());
	    std::transform(line.begin(), line.end(), 
                       haplotype.begin(),
                       if_then_else_return(_1 == '0', 0, 1));

	    if (haplotypes.size() > 0)
            if (haplotypes[0].size() != haplotype.size())
                throw ParseError();
	    haplotypes.push_back(haplotype);
	}

    if (haplotypes.empty()) throw ParseError();
    int noHaplotypes = haplotypes.size();
    int noMarkers    = haplotypes[0].size();

    matrix.resize(noHaplotypes, noMarkers);
    for (int i = 0; i < noHaplotypes; ++i)
	{
	    for (int j = 0; j < noMarkers; ++j)
            matrix(i,j) = haplotypes[i][j];
	}
    matrix.setPhased(true);
}

namespace {
    template <typename Itr>
    bool isSorted(Itr begin, Itr end)
    {
        if (begin+1 >= end) return true;
        for (Itr i = begin; i < end-1; ++i)
            if (*i > *(i+1)) return false;
        return true;
    }
}

static void
readMarkers(std::istream &is, 
            std::vector<std::string> &markerNames,
            std::vector<int>         &positions)
{
    while (is) {
        std::string name;
        int position;
        is >> name >> position;
        if (!name.empty()) {
            markerNames.push_back(name);
            positions.push_back(position);
        }
    }

    if (!isSorted(positions.begin(), positions.end())) 
        throw UnsortedPositions();
}

namespace options {
    bool verbose = false;
    std::string haplotypesFileName;
    std::string markersFileName;
    std::string phenotypesFileName;
    std::string outputFileName;
    bool quantitative = false;
}


namespace po = boost::program_options;
static void printUsage(const char *progname,
                       std::ostream &os,
                       po::options_description &options)
{
    os << "Usage: " << progname
       << " [options] genotype-file markers-file phenotype-file output-file"
       << std::endl
       << options << std::endl;
}

int main(int argc, char** argv)
{
    po::options_description visible("Allowed options");
    po::options_description hidden("Allowed options");
    po::options_description all("Allowed options");
    po::variables_map vm;
    visible.add_options()
        ("help,h", "Produce help message.")
        ("verbose,v","Turns on verbose output.")
        ("quantitative,q", "QTL mapping (default false).")
        ;
    hidden.add_options()
        ("genotypes",
         po::value<std::string>(&options::haplotypesFileName),
         "File containing marker data.")
        ("markers",
         po::value<std::string>(&options::markersFileName),
         "File containing marker names and positions.")
        ("phenotypes",
         po::value<std::string>(&options::phenotypesFileName),
         "File containing phenotypes.")
        ("output",
         po::value<std::string>(&options::outputFileName),
         "Output file.")
        ;
    all.add(visible).add(hidden);

    po::positional_options_description p;
    p.add("genotypes", 1);
    p.add("markers", 1);
    p.add("phenotypes", 1);
    p.add("output", 1);

    try {
        po::store(po::command_line_parser(argc, argv)
                  .options(all).positional(p).run(), 
                  vm, false);
        po::notify(vm);
    } catch (std::exception &e) {
        std::cerr << "Error parsing options: " << e.what()
                  << std::endl;
        printUsage(argv[0], std::cerr, visible);
        return EXIT_FAILURE;
    }

    if (vm.count("help")) {
        printUsage(argv[0], std::cout, visible);
        return EXIT_SUCCESS;
    }
    options::verbose = vm.count("verbose") > 0;
    options::quantitative = vm.count("quantitative") > 0;

    // checking options
    if (options::haplotypesFileName.empty())
	{
  	    std::cerr << "Error: No haplotype file specified."
                  << std::endl << std::endl;
	    printUsage(argv[0], std::cerr, visible);
	    return EXIT_FAILURE;
	}
    if (options::markersFileName.empty())
	{
  	    std::cerr << "Error: No markers file specified."
                  << std::endl << std::endl;
	    printUsage(argv[0], std::cerr, visible);
	    return EXIT_FAILURE;
	}
    if (options::phenotypesFileName.empty())
	{
  	    std::cerr << "Error: No phenotype file specified."
                  << std::endl << std::endl;
	    printUsage(argv[0], std::cerr, visible);
	    return EXIT_FAILURE;
	}
    if (options::outputFileName.empty())
	{
  	    std::cerr << "Error: No output file specified."
                  << std::endl << std::endl;
	    printUsage(argv[0], std::cerr, visible);
	    return EXIT_FAILURE;
	}


    // checking files
    std::ifstream haplotypesFile(options::haplotypesFileName.c_str());
    if (!haplotypesFile)
	{
	    std::cerr << "Error opening "
                  << options::haplotypesFileName << std::endl;
	    return EXIT_FAILURE;
	}
    std::ifstream markersFile(options::markersFileName.c_str());
    if (!markersFile)
	{
	    std::cerr << "Error opening " 
                  << options::markersFileName << std::endl;
	    return EXIT_FAILURE;
	}
    std::ifstream phenotypesFile(options::phenotypesFileName.c_str());
    if (!phenotypesFile)
	{
	    std::cerr << "Error opening " 
                  << options::phenotypesFileName << std::endl;
	    return EXIT_FAILURE;
	}

    std::auto_ptr<SNPFile> outputFile;
    try {
        SNPFile *ofile = new SNPFile(options::outputFileName, true);
        outputFile = std::auto_ptr<SNPFile>(ofile);
    } catch (std::exception &e) {
        std::cerr << "Error accessing " << options::outputFileName
                  << " : " << e.what() << std::endl;
        return EXIT_FAILURE;
    }
    

    try
    {
        // ----- MARKER NAMES --------------------------------------------------
        // ----- MARKER POSITIONS ----------------------------------------------
        std::vector<std::string> markerNames;
        std::vector<int> positions;
        readMarkers(markersFile, markerNames, positions);

        setMetadata(*outputFile, "positions", positions);
        setMetadata(*outputFile, "markerNames", markerNames);

        // ----- PHASED GENOTYPES ----------------------------------------------
	Matrix matrix(newReadWriteFileMatrix(*outputFile));
        readHaplotypes(haplotypesFile, matrix);

        // ----- PHENOTYPES ----------------------------------------------------
        if (options::quantitative)
        {
            std::vector<double> values;
            std::istream_iterator<double> instream(phenotypesFile), eof;
            for (; instream != eof; ++instream)
                values.push_back(*instream);
            if (static_cast<int>(values.size()) != matrix.noRows() / 2)
            {
                std::cerr << "The number of rows does not match the number of phenotypes." << std::endl
                          << "I expected twice as many rows as phenotypes." << std::endl << std::endl
                          << "Number of rows: " << matrix.noRows() << std::endl
                          << "Number of phenotypes: " << values.size() << std::endl
                          << std::endl;
                return EXIT_FAILURE;
            }
            setMetadata(*outputFile, "phenotype", values);
            setMetadata(*outputFile, "quantitative", true);
            setMetadata(*outputFile, "case/control", false);
        }
        else // case/control
        {
            std::vector<bool> status;
            std::istream_iterator<bool> instream(phenotypesFile), eof;
            for (; instream != eof; ++instream)
                status.push_back(*instream);
            if (static_cast<int>(status.size()) != matrix.noRows() / 2)
            {
                std::cerr << "The number of rows does not match the number of phenotypes." << std::endl
                          << "I expected twice as many rows as phenotypes." << std::endl;
                return EXIT_FAILURE;
            }
            setMetadata(*outputFile, "phenotype", status);
            setMetadata(*outputFile, "quantitative", false);
            setMetadata(*outputFile, "case/control", true);
        }
    }
    catch(std::exception &e)
    {
        std::cerr << "Error: " << e.what() << std::endl;

        if(unlink(options::outputFileName.c_str()) == -1 && errno != ENOENT)
            std::cerr << "Unable to remove file "
                      << options::outputFileName
                      << std::endl;
    }

    return EXIT_SUCCESS;
}
