/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus, Denmark
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include "matrix.hh"
#include "file_matrix.hh"

#include "snpfile.hh"
#include "metadata_access.hh"

#include "metadata.hh"
#include <boost/signals/trackable.hpp>
#include <boost/signals/connection.hpp>

#include <boost/bind.hpp>

#include <cassert>



/** \file
 * \brief Implementation of classes for accessing matrices in SNPFile's.
 */



namespace BiRC {
    namespace SNPFile {

        /**
         * \brief Implementation of MatrixData where the data is
         * stored in a SNPFile.
         *
         * This class only allows reading of data, however you shold
         * NOT expect it to be immutable, since the file backing it
         * may change.  This class may be used with SNPFiles in
         * read/write mode, however it is primarily meant to be used
         * with read-only files.
         */
        class ReadOnlyFileMatrixData : public MatrixData
        {
            /// Const reference to file backing this matrix.
            const SNPFile &i_const_file;

            /// Connection to the const file signalling changes in underlying pointer.
            boost::signals::connection i_file_connection;

            /// Copy-constructor. Not implemented. Do not copy file matrices.
            ReadOnlyFileMatrixData(ReadOnlyFileMatrixData &rhs);

            /// Assignment-operator. Not implemented. Do not assign
            /// file matrices.
            ReadOnlyFileMatrixData &operator= (ReadOnlyFileMatrixData &rhs);


	    // FIXME: it is really an error to call these on a
	    // read-only file matrix...
            /// Resizes the data.
            virtual void resizeData(int rows, int cols) {};
	    /// Set phased flag.
            virtual void setPhased(bool phased) {};
	    


        public:
            /// Key indicating which metadata contains the phased variable.
            static const std::string KEY_PHASED;
            /// Key indicating which metadata contains the hidden rows.
            static const std::string KEY_HIDDEN_ROWS;
            /// Key indicating which metadata contains the hidden columns.
            static const std::string KEY_HIDDEN_COLUMNS;

            /**
             * \brief Construct a new matrix from the data stored in the file.
             *
             * \param file	SNPFile object containing the matrix data.
             */
            ReadOnlyFileMatrixData(const SNPFile &file);

            /// Virtual destructor.
            virtual ~ReadOnlyFileMatrixData();

            /**
             * \brief Is the data phased?
             *
             * \throws MetadataError If the phased metadata have been removed from wrapped file.
             *                       Upon construction it is ensured this data exists, so this will
             *                       not happen unless somebody explicitely removes it.
             */
            virtual bool isPhased() const;

            /**
             * \brief Reads the sets of hidden rows and columns from file metadata.
             *
             * The current sets of hidden rows and columns in this matrix is replaced 
             * with that read from file metadata.
             */
            virtual void readHiding();
	    virtual void storeHiding();

            /**
             * \brief Used to signal changes to underlying file.
             *
             * Part of an observer pattern. Do not call this explicitely.
             */
            virtual void onFileUpdate();
        };



        /**
         * \brief Implementation of MatrixData where the data is
         * stored in a SNPFile.
         *
         * This class should only be used if the backing file is in
         * read/write mode.
         */
        class ReadWriteFileMatrixData : public ReadOnlyFileMatrixData
        {
            /// Reference to file backing this matrix.
            SNPFile &i_file;

            /// Connection to the file signalling changes in underlying pointer.
            boost::signals::connection i_file_connection;

            /// Copy-constructor. Not implemented. Do not copy file matrices.
            ReadWriteFileMatrixData(ReadWriteFileMatrixData &rhs);

            /// Assignment-operator. Not implemented. Do not assign
            /// file matrices.
            ReadWriteFileMatrixData &operator= (ReadWriteFileMatrixData &rhs);

        protected:
            /**
             * \brief Resize matrix to new dimensions.
             *
             * WARNING: resizing does *not* just expand or decrease the
             * dimensions -- the data will not (generally) be preserved by a
             * resizing.
             *
             * \param noRows	Number of rows in the resized matrix.
             * \param noCols	Number of columns in the resized matrix.
             */
            virtual void resizeData(int noRows, int noCols);

        public:
            /**
             * \brief Construct a new matrix from the data stored in the file.
             *
             * \param file	SNPFile object containing the matrix data.
             */
            ReadWriteFileMatrixData(SNPFile &file);

            /// Virtual destructor.
            virtual ~ReadWriteFileMatrixData();

            /// Set phased flag
            virtual void setPhased(bool phased);

            /**
             * \brief Stores the current sets of hidden rows and columns.
             * 
             * The sets are stored as file metadata. Any previously
             * stored values will be overwritten.
             */
            void storeHiding();

            /**
             * \brief Used to signal changes to underlying file.
             *
             * Part of an observer pattern. Do not call this explicitely.
             */
            virtual void onFileUpdate();
        };


        //******************** ReadOnlyFileMatrixData definitions ********************

        const std::string ReadOnlyFileMatrixData::KEY_PHASED         = "BiRC::SNPFile::FileMatrixData::isPhased";
        const std::string ReadOnlyFileMatrixData::KEY_HIDDEN_ROWS    = "BiRC::SNPFile::hiddenRows";
        const std::string ReadOnlyFileMatrixData::KEY_HIDDEN_COLUMNS = "BiRC::SNPFile::hiddenColumns";



        ReadOnlyFileMatrixData::ReadOnlyFileMatrixData(const SNPFile &file)
            : MatrixData(file.noRows(), file.noCols()),
              i_const_file(file)
        {
            i_file_connection = i_const_file.addMatrixDataListener(boost::bind(&ReadOnlyFileMatrixData::onFileUpdate, this));
            onFileUpdate();
        }



        ReadOnlyFileMatrixData::~ReadOnlyFileMatrixData() 
        {
            i_file_connection.disconnect();
        }



        bool
        ReadOnlyFileMatrixData::isPhased() const
        {
            if(!i_const_file.containsMetadata(KEY_PHASED))
                return false;
            else
                return fetchMetadata<bool>(i_const_file, KEY_PHASED); 

        }



        void
        ReadOnlyFileMatrixData::readHiding()
        {
            std::set<int> hiddenRows;
            std::set<int> hiddenColumns;

            if(i_const_file.containsMetadata(KEY_HIDDEN_ROWS))
                getMetadata(i_const_file, KEY_HIDDEN_ROWS, hiddenRows);

            if(i_const_file.containsMetadata(KEY_HIDDEN_COLUMNS))
                getMetadata(i_const_file, KEY_HIDDEN_COLUMNS, hiddenColumns);

            unhideRows(i_hiddenRows.begin(), i_hiddenRows.end());
            hideRows(hiddenRows.begin(), hiddenRows.end());

            const std::set<int> &previousHiddenColumns = getHiddenCols();
            unhideCols(previousHiddenColumns.begin(), previousHiddenColumns.end());
            hideCols(hiddenColumns.begin(), hiddenColumns.end());
        }


	void
	ReadOnlyFileMatrixData::storeHiding()
	{}			// nop


        void
        ReadOnlyFileMatrixData::onFileUpdate()
        {
	    i_data = const_cast<allele_t*>(i_const_file.accessMatrixData());
        }



        //******************** ReadWriteFileMatrixData definitions ********************

        class SNPFileReadOnlyError;

        namespace
        {
            inline void onReadOnlyShouldBeReadWrite()
            {
                throw SNPFileReadOnlyError("Cannot create a ReadWriteFileMatrix on an read-only file.");
            }
        }

        ReadWriteFileMatrixData::ReadWriteFileMatrixData(SNPFile &file)
            : ReadOnlyFileMatrixData(file),
              i_file(file)
        {
            if(i_file.isReadOnly())
                onReadOnlyShouldBeReadWrite();

            i_file_connection = i_file.addMatrixDataListener(boost::bind(&ReadWriteFileMatrixData::onFileUpdate, this));
            onFileUpdate();
        }



        ReadWriteFileMatrixData::~ReadWriteFileMatrixData() 
        {
            i_file_connection.disconnect();
        }



        void
        ReadWriteFileMatrixData::setPhased(bool phased)
        {
            setMetadata<bool>(i_file, KEY_PHASED, phased);
        }



        void
        ReadWriteFileMatrixData::storeHiding()
        {
            setMetadata(i_file, KEY_HIDDEN_ROWS, i_hiddenRows);
            setMetadata(i_file, KEY_HIDDEN_COLUMNS, i_hiddenCols);
        }



        void
        ReadWriteFileMatrixData::resizeData(int noRows, int noCols)
        {
            i_file.resizeMatrix(noRows, noCols);
        }



        void
        ReadWriteFileMatrixData::onFileUpdate()
        {
            if(i_file.isReadOnly())
                onReadOnlyShouldBeReadWrite();
	    i_data = i_file.accessMatrixData();
        }





	// Factory methods
	ImmutableMatrix newReadOnlyFileMatrix(const SNPFile &file)
	{
	    return ImmutableMatrix(new ReadOnlyFileMatrixData(file));
	}

	Matrix newReadWriteFileMatrix(SNPFile &file)
	{
	    return Matrix(new ReadWriteFileMatrixData(file));
	}

    }
}
