/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus, Denmark
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */


#undef NDEBUG			// enable asserts always


#include "matrix.hh"
#include "file_matrix.hh"
#include "snpfile.hh"
#include "metadata_access.hh"

#include "test_util.hh"

#include <iostream>
#include <sstream>

#include <vector>



using namespace BiRC::SNPFile;



/**
 * \brief Main method. Test handling of moving matrix data.
 *
 * When metadata is added to a file and it is committet, the file
 * is mmapped again, and thus its data pointer may change.
 * This tests that these changed are properly propagated to the
 * FileMatrixData class
 *
 * \returns 0 indicating program success.
 */
int main()
{
    // ********** Begin: Test moving data pointers **********

    int fd1;
    const char *fileName1 = makeTempFile(fd1);

    SNPFile file1(fileName1, true);

    int fd2;
    const char *fileName2 = makeTempFile(fd2);

    SNPFile file2(fileName2, true);

    std::ostringstream strstrm;
    std::string expected;

    Matrix m1(newReadWriteFileMatrix(file1));
    m1.resize(4, 4);

    m1(0, 0) = 1;
    m1(1, 1) = 2;
    m1(2, 2) = 3;
    m1(3, 3) = 4;

    expected = 
        "1 0 0 0 \n"
        "0 2 0 0 \n"
        "0 0 3 0 \n"
        "0 0 0 4 \n";
    strstrm << m1;
    assert(expected == strstrm.str());
    strstrm.str("");


    Matrix m2(newReadWriteFileMatrix(file2));
    m2.resize(12, 12);
    m2(3, 3) = 14;

    std::vector<int> v;
    for(int i = 0; i < 4096; i++)
        v.push_back(i);



    setMetadata(file1, "key", v);
    file1.commit();



    strstrm << m1;
    assert(expected == strstrm.str());
    strstrm.str("");

    assert(m2(3,3) == 14);

    removeTempFile(fd1, fileName1);
    removeTempFile(fd2, fileName2);

    // **********  End: Test moving data pointers  **********

    // ************* Begin: Test read-only file *************

    int fd3;
    const char *fileName3 = makeTempFile(fd3);

    {
        SNPFile file3(fileName3, true);
	Matrix m3(newReadWriteFileMatrix(file3));

        assert(m3.isPhased() == false);
        m3.resize(2, 2);
        m3(0, 0) = 1;
        m3(0, 1) = 2;
        m3(1, 0) = 3;
        m3(1, 1) = 4;
        file3.close();
    }

    {
        SNPFile file3(fileName3);

        {
	    ImmutableMatrix m3(newReadOnlyFileMatrix(file3));

            assert(m3.isPhased() == false);
            assert(m3(0, 0) == 1);
            assert(m3(0, 1) == 2);
            assert(m3(1, 0) == 3);
            assert(m3(1, 1) == 4);
            file3.close();
        }

        try
        {
	    Matrix m3(newReadWriteFileMatrix(file3));
            assert(!"An exception should have been thrown.");
        }
        catch(SNPFileReadOnlyError &e)
        {
        }
    }

    removeTempFile(fd3, fileName3);

    // *************  End: Test read-only file  *************

    // ********* Begin: Test persistent data-hiding *********

    int fd4;
    const char *fileName4 = makeTempFile(fd4);
    
    {
        SNPFile file(fileName4, true);
	Matrix m(newReadWriteFileMatrix(file));
        m.resize(4, 4);
        m(0, 0) = 1;
        m(1, 1) = 2;
        m(2, 2) = 3;
        m(3, 3) = 4;

        m.readHiding();

        expected =
            "1 0 0 0 \n"
            "0 2 0 0 \n"
            "0 0 3 0 \n"
            "0 0 0 4 \n";
        strstrm << m.view();
        assert(expected == strstrm.str());
        strstrm.str("");

        m.hideRow(2);
        m.hideCol(1);

        expected =
            "1 0 0 \n"
            "0 0 0 \n"
            "0 0 4 \n";
        strstrm << m.view();
        assert(expected == strstrm.str());
        strstrm.str("");

        m.storeHiding();

        file.close();
    }

    {
        SNPFile file(fileName4);
	ImmutableMatrix m(newReadOnlyFileMatrix(file));
        
        expected =
            "1 0 0 0 \n"
            "0 2 0 0 \n"
            "0 0 3 0 \n"
            "0 0 0 4 \n";
        strstrm << m.view();
        assert(expected == strstrm.str());
        strstrm.str("");

        m.hideCol(0);
        
        expected =
            "0 0 0 \n"
            "2 0 0 \n"
            "0 3 0 \n"
            "0 0 4 \n";
        strstrm << m.view();
        assert(expected == strstrm.str());
        strstrm.str("");

        m.readHiding();

        expected =
            "1 0 0 \n"
            "0 0 0 \n"
            "0 0 4 \n";
        strstrm << m.view();
        assert(expected == strstrm.str());
        strstrm.str("");

        file.close();
    }

    removeTempFile(fd4, fileName4);

    // *********  End: Test persistent data-hiding  *********
    return 0;
}
