/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2007, 2008
 *               Bioinformatics Research Center,
 *               University of Aarhus, Denmark
 *               Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#ifndef INPUT_ARCHIVE_HH_INCLUDED
#define INPUT_ARCHIVE_HH_INCLUDED

#include <string>

#include "archive_common.hh"



/** \file
 * \brief Archive for portable de-serialization of arbitrary data.
 */



namespace BiRC {
    namespace SNPFile {
        /**
         * \brief An archive for reading data in a portable way.
         */
        class InputArchive
        {
            std::string::const_iterator it;

        public:
            /**
             * \brief Constructs an archive reading its data from a given string.
             *
             * \param source String containing serialized data.
             */
            InputArchive(const std::string &source)
                : it(source.begin())
            {
            }

            /**
             * \brief Load binary data directly to a
             * destination. Beware of portability.
             *
             * \param dst Where to write bytes.
             * \param count How many bytes to read and write.
             */
            void loadBytes(void *dst, int count)
            {
                char *charDst = static_cast<char *>(dst);
                for(int i = 0; i < count; i++)
                    charDst[i] = *(it++);
            }

        private:

            ///Loads an intgral type.
            template<typename T>
            void loadIntegral(T &i)
            {
                uint64_t tmp = 0;
                loadBytes(&tmp, sizeof(uint64_t));
                i = T(fromLittleEndian(tmp));
            }

            ///Loads a floating point type.
            template<typename T>
            void loadFloat(T &f)
            {
                T tmp = 0;
                loadBytes(&tmp, sizeof(T));
                f = fromLittleEndian(tmp);
            }

        public:

            /**
             * \brief Loads an arbitrary type from this archive.
             *
             * The loading will be done by BiRC::SNPFile::SerializationTrait<T>.load.
             * Refine that to deserialize userdefined types.
             *
             * \param t Where to write result.
             */
            template<typename T>
            void load(T &t);

            ///Convenience operator. Streams data using the load function.
            template <typename T>
            InputArchive &operator>> (T &t)
            {
                load(t);
                return *this;
            }

            ///Convenience operator. Streams data using the load function.
            template <typename T>
            InputArchive &operator| (T &t)
            {
                return operator>>(t);
            }
        };



        template<typename T>
        void InputArchive::load(T &t)
        {
            SerializationTrait<T>().load(*this, t);
        }

#define BIRC_SNPFILE_MAKE_INTEGRAL_LOAD_FUNCTION(T)  \
        template<> inline                            \
        void InputArchive::load(T &t)                \
        {                                            \
            loadIntegral(t);                         \
        }

        BIRC_SNPFILE_MAKE_INTEGRAL_LOAD_FUNCTION(bool);
        BIRC_SNPFILE_MAKE_INTEGRAL_LOAD_FUNCTION(char);
        BIRC_SNPFILE_MAKE_INTEGRAL_LOAD_FUNCTION(unsigned char);
        BIRC_SNPFILE_MAKE_INTEGRAL_LOAD_FUNCTION(signed char);
        BIRC_SNPFILE_MAKE_INTEGRAL_LOAD_FUNCTION(unsigned short);
        BIRC_SNPFILE_MAKE_INTEGRAL_LOAD_FUNCTION(signed short);
        BIRC_SNPFILE_MAKE_INTEGRAL_LOAD_FUNCTION(unsigned int);
        BIRC_SNPFILE_MAKE_INTEGRAL_LOAD_FUNCTION(signed int);
        BIRC_SNPFILE_MAKE_INTEGRAL_LOAD_FUNCTION(unsigned long);
        BIRC_SNPFILE_MAKE_INTEGRAL_LOAD_FUNCTION(signed long);
        BIRC_SNPFILE_MAKE_INTEGRAL_LOAD_FUNCTION(unsigned long long);
        BIRC_SNPFILE_MAKE_INTEGRAL_LOAD_FUNCTION(signed long long);

#undef BIRC_SNPFILE_MAKE_INTEGRAL_LOAD_FUNCTION

#define BIRC_SNPFILE_MAKE_FLOAT_LOAD_FUNCTION(T)     \
        template<> inline                            \
        void InputArchive::load(T &t)                \
        {                                            \
            loadFloat(t);                            \
        }

        BIRC_SNPFILE_MAKE_FLOAT_LOAD_FUNCTION(float);
        BIRC_SNPFILE_MAKE_FLOAT_LOAD_FUNCTION(double);
        BIRC_SNPFILE_MAKE_FLOAT_LOAD_FUNCTION(long double);

#undef BIRC_SNPFILE_MAKE_FLOAT_LOAD_FUNCTION
    }
}



#endif
