/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus, Denmark
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#ifndef ITERATORS_HH_INCLUDED
#define ITERATORS_HH_INCLUDED



#include "matrix.hh"

#include <vector>

#include <boost/scoped_ptr.hpp>



namespace BiRC
{
    namespace SNPFile
    {
        namespace detail
        {
            /**
             * \brief Iterator over the columns or rows of a matrix.
             *
             * WARNING: Since a matrix is not actually represented as
             * a list of columns or rows internally this does NOT
             * actually return references to objects contained in the
             * matrix but temporary representations instead and thus
             * do NOT implement the complete interface demanded by
             * STL. It is however close enough to be usable in most
             * instances.
             */
            template <typename Derived>
            class WindowIteratorBase
            {
            private:
                const MatrixView *base;
                int windowSize;
                int nextIdx;



            public:
                /// Signal to STL about capabilities of this iterator
                typedef std::random_access_iterator_tag   iterator_category;
                /// Types of values returned by dereferencing this iterator
                typedef MatrixView                        value_type;
                /// Type this iterator expects size to be represented by
                typedef int                               difference_type;
                /// Type of pointers to values represented by this iterator
                typedef value_type*                       pointer;
                /// Type of references to values represented by this iterator
                typedef value_type&                       reference;



                /// Constructs an iterator that is always at the end.
                WindowIteratorBase()
                    : base(),
                      windowSize(0),
                      nextIdx(0)
                {}



                /**
                 * \brief Constructs iterator with given index and size.
                 *
                 * \param data       Matrix to iterate over.
                 * \param windowSize Size of windows iterated over.
                 * \param nextIdx    Index of column/row to point to.
                 */
                WindowIteratorBase(const MatrixView &base, int windowSize = 1, int nextIdx = 0)
                    : base(&base),
                      windowSize(windowSize),
                      nextIdx(nextIdx)
                {}



                /**
                 * \brief Constructs iterator with given index and size.
                 *
                 * \param data       Matrix to iterate over.
                 * \param windowSize Size of windows iterated over.
                 * \param nextIdx    Index of column/row to point to.
                 */
                WindowIteratorBase(const MatrixView * base, int windowSize = 1, int nextIdx = 0)
                    : base(base),
                      windowSize(windowSize),
                      nextIdx(nextIdx)
                {}



                /// Returns the View this iterator iterates over.
                const MatrixView &getBaseView() const
                {
                    return *base;
                }



                /// Returns the index of the column/row dereference returns.
                int getNextStartIndex() const
                {
                    return nextIdx;
                }


            
                /// Returns the width of the window.
                int getWindowWidth() const
                {
                    return windowSize;
                }



                /// Dereference operator. Not implemented. Derived class should implement this:
                MatrixView operator*();



                /// Moves iterator to next column/row.
                Derived operator++ ()
                {
                    return Derived(this, ++nextIdx);
                }

                /// Moves iterator to next column/row.
                Derived operator++ (int)
                {
                    return Derived(this, nextIdx++);
                }

                /// Moves iterator to previous column/row.
                Derived operator-- ()
                {
                    return Derived(this, --nextIdx);        
                }

                /// Moves iterator to previous column/row.
                Derived operator-- (int)
                {
                    return Derived(this, nextIdx--);
                }

                /// Moves iterator a number of columns/rows.
                Derived operator += (int i)
                {
                    nextIdx += i;
                    return Derived(this, nextIdx);
                }

                /// Moves iterator a number of columns/rows.
                Derived operator -= (int i)
                {
                    return (*this) += -i;
                }

                /// Creates a new iterator at a offset of this.
                Derived operator+ (int i) const
                {
                    return Derived(this, nextIdx + i);
                }

                /// Creates a new iterator at a offset of this.
                Derived operator- (int i) const
                {
                    return *this + (-i);
                }

                /// Computes distance between iterators.
                int operator- (const Derived &rhs) const
                {
                    assert(windowSize == rhs.windowSize);
                    assert(base == rhs.base);
                    return nextIdx - rhs.nextIdx;
                }

                /// Compares iterators. Iterator are compared by position.
                bool operator< (const Derived &rhs) const
                {
                    return base == rhs.base
                        && windowSize == rhs.windowSize
                        && nextIdx < rhs.nextIdx;
                }

                /// Iterators are equal iff they represent same position in same matrix and have the same window-size.
                bool operator== (const Derived &rhs) const
                {
                    if(base == 0)
                    {
                        if(rhs.base == 0)
                            return true;
                        else
                            return Derived::isAtEnd(rhs);
                    }
                    else if(rhs.base == 0)
                    {
                        return Derived::isAtEnd(*this);
                    }
                    else
                    {
                        return base == rhs.base
                            && windowSize == rhs.windowSize
                            && nextIdx == rhs.nextIdx;
                    }
                }

                /// Iiterators are <= iff they share the matrix and window-size and their position is <=.
                bool operator<= (const Derived &rhs) const
                {
                    return (*this < rhs) || (*this == rhs);
                }

                /// Iterator are different if their matrix, window-size or position is different.
                bool operator!= (const Derived &rhs) const
                {
                    return !(*this == rhs);
                }

                /// Iterators are > if they share matrix and window-size, but their position is >.
                bool operator> (const Derived &rhs) const
                {
                    return rhs < *this;
                }

                /// Iterators are >= if they share matrix and window-size, but their position is >=.
                bool operator>= (const Derived &rhs) const
                {
                    return rhs <= *this;
                }
            };







            template <typename Derived>
            class ColumnWindowIteratorBase : public WindowIteratorBase<Derived>
            {
            public:
                /// Constructs an iterator that is always at the end.
                ColumnWindowIteratorBase() : WindowIteratorBase<Derived>() {}

                /**
                 * \brief Constructs iterator with given index and size.
                 *
                 * \param data       Matrix to iterate over.
                 * \param windowSize Size of windows iterated over.
                 * \param nextIdx    Index of column to point to.
                 */
                ColumnWindowIteratorBase(const MatrixView &base, int windowSize, int idx)
                    : WindowIteratorBase<Derived>(base, windowSize, idx) {}

                /**
                 * \brief Constructs iterator with given index and size.
                 *
                 * \param data       Matrix to iterate over.
                 * \param windowSize Size of windows iterated over.
                 * \param nextIdx    Index of column to point to.
                 */
                ColumnWindowIteratorBase(const MatrixView *base, int windowSize, int idx)
                    : WindowIteratorBase<Derived>(base, windowSize, idx) {}

                /// Returns a MatrixView of the columns in current window.
                MatrixView operator* ()
                {
                    return MatrixView(this->getBaseView(),
                                      0, this->getBaseView().noRows(),
                                      this->getNextStartIndex(), this->getNextStartIndex() + this->getWindowWidth());
                }

                /// Returns true iff it points past end of base matrix view.
                static bool isAtEnd(const detail::WindowIteratorBase<Derived> &it)
                {
                    return it.getBaseView().noCols() < it.getNextStartIndex() + it.getWindowWidth();
                }
            };







            template <typename Derived>
            class RowWindowIteratorBase : public WindowIteratorBase<Derived>
            {
            public:
                /// Constructs an iterator that is always at the end.
                RowWindowIteratorBase()
                    : WindowIteratorBase<Derived>() {}

                /**
                 * \brief Constructs iterator with given index and size.
                 *
                 * \param data       Matrix to iterate over.
                 * \param windowSize Size of windows iterated over.
                 * \param nextIdx    Index of row to point to.
                 */
                RowWindowIteratorBase(const MatrixView &base, int windowSize = 1, int idx = 0)
                    : WindowIteratorBase<Derived>(base, windowSize, idx) {}

                /**
                 * \brief Constructs iterator with given index and size.
                 *
                 * \param data       Matrix to iterate over.
                 * \param windowSize Size of windows iterated over.
                 * \param nextIdx    Index of row to point to.
                 */
                RowWindowIteratorBase(const MatrixView *base, int windowSize = 1, int idx = 0)
                    : detail::WindowIteratorBase<Derived>(base, windowSize, idx) {}

                /// Returns a MatrixView of the rows in current window.
                MatrixView operator* ()
                {
                    return MatrixView(this->getBaseView(),
                                      this->getNextStartIndex(), this->getNextStartIndex() + this->getWindowWidth(),
                                      0, this->getBaseView().noCols());
                }

                /// Returns true iff it points past end of base matrix view.
                static bool isAtEnd(const detail::WindowIteratorBase<Derived> &it)
                {
                    return it.getBaseView().noRows() < it.getNextStartIndex() + it.getWindowWidth();
                }
            };
        }







        class ColumnWindowIterator : public detail::ColumnWindowIteratorBase<ColumnWindowIterator>
        {
            friend class detail::WindowIteratorBase<ColumnWindowIterator>;

            ColumnWindowIterator(const detail::WindowIteratorBase<ColumnWindowIterator> *copy, int newIdx)
                : detail::ColumnWindowIteratorBase<ColumnWindowIterator>(copy->getBaseView(), copy->getWindowWidth(), newIdx) {}



        public:
            /// Constructs an iterator that is always at the end.
            ColumnWindowIterator()
                : detail::ColumnWindowIteratorBase<ColumnWindowIterator>() {}

            /**
             * \brief Constructs iterator with given index and size.
             *
             * \param data       Matrix to iterate over.
             * \param windowSize Size of windows iterated over.
             * \param nextIdx    Index of column to point to.
             */
            ColumnWindowIterator(const MatrixView &base, int windowSize = 1, int idx = 0)
                : detail::ColumnWindowIteratorBase<ColumnWindowIterator>(base, windowSize, idx) {}

            /**
             * \brief Constructs iterator with given index and size.
             *
             * \param data       Matrix to iterate over.
             * \param windowSize Size of windows iterated over.
             * \param nextIdx    Index of column to point to.
             */
            ColumnWindowIterator(const MatrixView *base, int windowSize = 1, int idx = 0)
                : detail::ColumnWindowIteratorBase<ColumnWindowIterator>(base, windowSize, idx) {}
        };

        /**
         * \brief Operator overloading of + for ColumnWindowIterator. Does what you would expect.
         */
        ColumnWindowIterator operator+ (int i, const ColumnWindowIterator& rhs)
        {
            return rhs + i;
        }







        class ColumnIterator : public detail::ColumnWindowIteratorBase<ColumnIterator>
        {
            friend class detail::WindowIteratorBase<ColumnIterator>;

            ColumnIterator(const detail::WindowIteratorBase<ColumnIterator> *copy, int newIdx)
                : detail::ColumnWindowIteratorBase<ColumnIterator>(copy->getBaseView(), 1, newIdx) {}



        public:
            /// Constructs an iterator that is always at the end.
            ColumnIterator()
                : detail::ColumnWindowIteratorBase<ColumnIterator>() {}

            /**
             * \brief Constructs iterator with given index and size.
             *
             * \param data       Matrix to iterate over.
             * \param nextIdx    Index of column to point to.
             */
            ColumnIterator(const MatrixView &base, int idx = 0)
                : detail::ColumnWindowIteratorBase<ColumnIterator>(base, 1, idx) {}

            /**
             * \brief Constructs iterator with given index and size.
             *
             * \param data       Matrix to iterate over.
             * \param nextIdx    Index of column to point to.
             */
            ColumnIterator(const MatrixView *base, int idx = 0)
                : detail::ColumnWindowIteratorBase<ColumnIterator>(base, 1, idx) {}
        };

        /**
         * \brief Operator overloading of + for ColumnWindowIterator. Does what you would expect.
         */
        ColumnIterator operator+ (int i, const ColumnIterator& rhs)
        {
            return rhs + i;
        }







        class RowWindowIterator : public detail::RowWindowIteratorBase<RowWindowIterator>
        {
            friend class detail::WindowIteratorBase<RowWindowIterator>;

            RowWindowIterator(const detail::WindowIteratorBase<RowWindowIterator> *copy, int nextIdx)
                : detail::RowWindowIteratorBase<RowWindowIterator>(copy->getBaseView(), copy->getWindowWidth(), nextIdx) {}



        public:
            /// Constructs an iterator that is always at the end.
            RowWindowIterator()
                : detail::RowWindowIteratorBase<RowWindowIterator>() {}

            /**
             * \brief Constructs iterator with given index and size.
             *
             * \param data       Matrix to iterate over.
             * \param windowSize Size of windows iterated over.
             * \param nextIdx    Index of row to point to.
             */
            RowWindowIterator(const MatrixView &base, int windowSize = 1, int idx = 0)
                : detail::RowWindowIteratorBase<RowWindowIterator>(base, windowSize, idx) {}

            /**
             * \brief Constructs iterator with given index and size.
             *
             * \param data       Matrix to iterate over.
             * \param windowSize Size of windows iterated over.
             * \param nextIdx    Index of row to point to.
             */
            RowWindowIterator(const MatrixView *base, int windowSize = 1, int idx = 0)
                : detail::RowWindowIteratorBase<RowWindowIterator>(base, windowSize, idx) {}
        };

        /**
         * \brief Operator overloading of + for ColumnWindowIterator. Does what you would expect.
         */
        RowWindowIterator operator+ (int i, const RowWindowIterator& rhs)
        {
            return rhs + i;
        }







        class RowIterator : public detail::RowWindowIteratorBase<RowIterator>
        {
            friend class detail::WindowIteratorBase<RowIterator>;

            RowIterator(const detail::WindowIteratorBase<RowIterator> *copy, int newIdx)
                : detail::RowWindowIteratorBase<RowIterator>(copy->getBaseView(), 1, newIdx) {}



        public:
            /// Constructs an iterator that is always at the end.
            RowIterator() : detail::RowWindowIteratorBase<RowIterator>() {}

            /**
             * \brief Constructs iterator with given index and size.
             *
             * \param data       Matrix to iterate over.
             * \param nextIdx    Index of row to point to.
             */
            RowIterator(const MatrixView &base, int idx = 0)
                : detail::RowWindowIteratorBase<RowIterator>(base, 1, idx) {}

            /**
             * \brief Constructs iterator with given index and size.
             *
             * \param data       Matrix to iterate over.
             * \param nextIdx    Index of row to point to.
             */
            RowIterator(const MatrixView *base, int idx = 0)
                : detail::RowWindowIteratorBase<RowIterator>(base, 1, idx) {}
        };

        /**
         * \brief Operator overloading of + for RowWindowIterator. Does what you would expect.
         */
        RowIterator operator+ (int i, const RowIterator& rhs)
        {
            return rhs + i;
        }
    }
}



#endif
