/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus, Denmark
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include "matrix.hh"
#include <cmath>



/** \file
 * \brief Implementation of functions on MutableMatrixData and MatrixView.
 */



namespace BiRC {
    namespace SNPFile {

	MatrixData::MatrixData()
	    : i_noRows(0),
	      i_noCols(0),
	      i_data(0)
	{
	}
	
	MatrixData::MatrixData(int noRows, int noCols)
	    : i_noRows(noRows),
	      i_noCols(noCols),
	      i_data(0)
	{
	}

	MatrixData::~MatrixData()
	{
	    //DO NOT delete i_data; subclasses handles that.
	}


        void Matrix::overwrite(const ImmutableMatrix &rhs)
        {
	    // FIXME: this isn't exception safe...
	    // FIXME: it isn't safe it rhs contains the same i_data either
            resize(rhs.noRows(), rhs.noCols());
            memcpy(i_data->i_data, rhs.i_data->i_data, 
		   sizeof(allele_t)*noRows()*noCols());
            setPhased(rhs.isPhased());
        }


        MatrixView::MatrixView(const ImmutableMatrix &matrix, 
                               int rowStart, int rowEnd, 
                               int colStart, int colEnd)
            : i_matrix(matrix), i_rowIndices(rowEnd-rowStart),
              i_colIndices(colEnd-colStart)
        {
            for (int i = 0; i < noRows(); ++i)
                i_rowIndices[i] = i + rowStart;
            for (int i = 0; i < noCols(); ++i)
                i_colIndices[i] = i + colStart;

            removeHiddenRowsAndCols();
        }



        MatrixView::MatrixView(const MatrixView &view,
                               int rowStart, int rowEnd, 
                               int colStart, int colEnd)
            : i_matrix(view.i_matrix), i_rowIndices(rowEnd-rowStart),
              i_colIndices(colEnd-colStart)
        {
            for (int i = 0; i < noRows(); ++i)
                i_rowIndices[i] = view.i_rowIndices[rowStart + i];
            for (int i = 0; i < noCols(); ++i)
                i_colIndices[i] = view.i_colIndices[colStart + i];
        }



        namespace {
            struct SetContains : std::unary_function<bool, int>
            {
                const std::set<int> &set;
                SetContains(const std::set<int> &set)
                    : set(set) {}
                bool operator() (int i) { return set.find(i) != set.end(); }
            };

            void listRemove(std::vector<int> &list, std::set<int> toRemove)
            {
                std::vector<int>::iterator newEnd = remove_if(list.begin(), list.end(), SetContains(toRemove));
                list.resize(newEnd - list.begin());
            }
        }



        void
        MatrixView::removeHiddenRowsAndCols()
        {
            listRemove(i_colIndices, i_matrix.getHiddenCols());
            listRemove(i_rowIndices, i_matrix.getHiddenRows());
        }
    }
}
