/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2006 by Thomas Mailund <mailund@mailund.dk>
 *                    and Søren Besenbacher <besen@birc.au.dk>
 *
 * Copyright (C) 2007, 2008
 *               Bioinformatics Research Center,
 *               University of Aarhus, Denmark
 *               Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#ifndef MATRIX_HH_INCLUDED
#define MATRIX_HH_INCLUDED

#include <cstring>
#include <vector>
#include <set>
#include <algorithm>
#include <iostream>
#include <cassert>

#include <boost/shared_ptr.hpp>

#include "common.hh"



/** \file
 * \brief Representation of genotype data in matrix form.
 */


namespace BiRC {
    namespace SNPFile {

        class MatrixView;

        /// Interpretation of alleles -- not the *actual* type since
        /// we want to ensure we know the size and bit-interpretation
        /// of the type.
        enum allele_interpretation_t
        {
            HOMOZYGOTE_0 = 0, 
            HOMOZYGOTE_1 = 1, 
            HETEROZYGOTE = 2, 
            MISSING = -9,
        };

        /**
         * \brief Datastructure containing genotype data.
         *
         * This class is a base for classes that actually stores the
         * haplotype data.  The matrix operations are implemented in
         * the handler classes ImmutableMatrix and Matrix, but the
         * data is usually read through the MatrixView class, that
         * represents sub-matrices.
         *
         * For efficiency this class contains a pointer to its data
         * and allows access to the data directly through this. It is
         * the responsibility of derived class to allocate and free
         * the store pointed to by this pointer.
	 *
	 * Through the interface specified here, it is possible to
	 * implement different representation of matrices, such as RAM
	 * or file based matrices.  Through the handler classes, the
	 * different data representations will have a uniform and
	 * abstract interface, where all implementation details can be
	 * hidden.  Through the pointer access to the data, the
	 * handlers can implement the matrix operations with minimal
	 * performance overhead.
         *
	 * \see ImmutableMatrix
	 * \see Matrix
         * \see MatrixView
         */
        class MatrixData
        {
	    friend class ImmutableMatrix;
	    friend class Matrix;

	protected:
            /// Number of rows;
            int i_noRows;

            /// Number of columns.
            int i_noCols;

            /// Set of hidden rows;
            std::set<int> i_hiddenRows;
            /// Set of hidden columns;
            std::set<int> i_hiddenCols;

            /// actual data...
	    allele_t *i_data;

            /**
             * \brief Hides a given column.
             *
             * \warning This does not affect direct accesses to this
             * matrix, but only access to MatrixViews made on this
             * matrix. Secondly, once a matrix view has been created
             * any changes will not affect it.
             *
             * \param j Which column to hide.
             */
            void hideCol(int j) { i_hiddenCols.insert(j); }

            /**
             * \brief Hide a range of columns.
             *
             * \warning This does not affect direct accesses to this
             * matrix, but only access to MatrixViews made on this
             * matrix. Secondly, once a matrix view has been created
             * any changes will not affect it.
             *
             * \param begin Beginning of columns to hide.
             * \param end   End of columns to hide.
             */
            template<typename Itr>
            void hideCols(Itr begin, Itr end);

            /**
             * \brief Unhides a given column.
             *
             * \warning This does not affect direct accesses to this
             * matrix, but only access to MatrixViews made on this
             * matrix. Secondly, once a matrix view has been created
             * any changes will not affect it.
             *
             * \param j Which column to hide.
             */
            void unhideCol(int j) { i_hiddenCols.erase(j); }

            /**
             * \brief Unhide a range of columns.
             *
             * \warning This does not affect direct accesses to this
             * matrix, but only access to MatrixViews made on this
             * matrix. Secondly, once a matrix view has been created
             * any changes will not affect it.
             *
             * \param begin Beginning of columns to unhide.
             * \param end   End of columns to unhide.
             */
            template<typename Itr>
            void unhideCols(Itr begin, Itr end);

            /// Get the set of the hidden cols;
            const std::set<int> &getHiddenCols() const {
		return i_hiddenCols;
	    }


            /**
             * \brief Hides a given row.
             *
             * \warning This does not affect direct accesses to this
             * matrix, but only access to MatrixViews made on this
             * matrix. Secondly, once a matrix view has been created
             * any changes will not affect it.
             *
             * \param j Which row to hide.
             */
            void hideRow(int j) { i_hiddenRows.insert(j); }

            /**
             * \brief Hide a range of rows.
             *
             * \warning This does not affect direct accesses to this
             * matrix, but only access to MatrixViews made on this
             * matrix. Secondly, once a matrix view has been created
             * any changes will not affect it.
             *
             * \param begin Beginning of rows to hide.
             * \param end   End of rows to hide.
             */
            template<typename Itr>
            void hideRows(Itr begin, Itr end);

            /**
             * \brief Unhides a given row.
             *
             * \warning This does not affect direct accesses to this
             * matrix, but only access to MatrixViews made on this
             * matrix. Secondly, once a matrix view has been created
             * any changes will not affect it.
             *
             * \param j Which row to hide.
             */
            void unhideRow(int j) { i_hiddenRows.erase(j); }

            /**
             * \brief Unhide a range of rows.
             *
             * \warning This does not affect direct accesses to this
             * matrix, but only access to MatrixViews made on this
             * matrix. Secondly, once a matrix view has been created
             * any changes will not affect it.
             *
             * \param begin Beginning of rows to unhide.
             * \param end   End of rows to unhide.
             */
            template<typename Itr>
            void unhideRows(Itr begin, Itr end);


	private:
            // not implemented... don't copy MatrixData lightly...
            MatrixData(const MatrixData&);
            MatrixData &operator=(const MatrixData&);

            /**
             * \brief Translate pair-index into a single index, to look up a
             * cell in the internal representation.
             *
             * \param i		Row index.
             * \param j		Column index.
             *
             * \returns		The corresponding index in i_data.
             */
            size_t translateIndex(int i, int j) const
            {
                return size_t(i) + size_t(j) * i_noRows;
            }

            ///Resets stored dimensions, does not actually resize matrix.
            void resetDimensions(int noRows, int noCols)
            {
                i_noRows = noRows;
                i_noCols = noCols;
            }

        protected:

            /// Resizes the data.
            virtual void resizeData(int rows, int cols) = 0;

            /// Is the data phased?
            virtual bool isPhased() const = 0;

	    /// Set phased flag.
            virtual void setPhased(bool phased) = 0;

            /**
             * \brief Reads the sets of hidden rows and columns from
             * metadata (where applicable).
             *
             * The current sets of hidden rows and columns in this
             * matrix is replaced with that read from metadata.
             */
	    virtual void readHiding() = 0;

            /**
             * \brief Stores the current sets of hidden rows and
             * columns in metadata (where applicable).
             * 
             * The sets are stored as file metadata. Any previously
             * stored values will be overwritten.
             */
	    virtual void storeHiding() = 0;

            /**
             * \brief Default constructor.
             *
             * Constructs empty matrix, with 0 rows and 0 columns
             */
            MatrixData();

            /**
             * \brief Constructs a matrix with given dimensions.
             *
             * \param noRows Number of rows to initialize matrix with
             * \param noCols Number of columns to initialize matrix with
             */
            MatrixData(int noRows, int noCols);

	public:
            /// Virtual destructor.
            virtual ~MatrixData();
	};

	/**
	 * \brief Interface to an immutable matrix
	 *
	 * \see MatrixData
	 * \see Matrix
	 * \see MatrixView
	 */
	class ImmutableMatrix {
	    friend class Matrix;
	    boost::shared_ptr<MatrixData> i_data;

        public:
	    explicit ImmutableMatrix(MatrixData *data) : i_data(data) {};

	    bool operator==(const ImmutableMatrix &rhs) const {
		return i_data == rhs.i_data;
	    }
	    bool operator!=(const ImmutableMatrix &rhs) const {
		return !(*this == rhs);
	    }

            /// Get the number of rows for the matrix.
            int noRows() const { return i_data->i_noRows; }

            /// Get the number of columns for the matrix.
            int noCols() const { return i_data->i_noCols; }

            /**
             * \brief Hides a given column.
             *
             * \warning This does not affect direct accesses to this
             * matrix, but only access to MatrixViews made on this
             * matrix. Secondly, once a matrix view has been created
             * any changes will not affect it.
             *
             * \param j Which column to hide.
             */
            void hideCol(int j) { i_data->hideCol(j); }

            /**
             * \brief Hide a range of columns.
             *
             * \warning This does not affect direct accesses to this
             * matrix, but only access to MatrixViews made on this
             * matrix. Secondly, once a matrix view has been created
             * any changes will not affect it.
             *
             * \param begin Beginning of columns to hide.
             * \param end   End of columns to hide.
             */
            template<typename Itr>
            void hideCols(Itr begin, Itr end) { i_data->hideCols(begin,end); }
		

            /**
             * \brief Unhides a given column.
             *
             * \warning This does not affect direct accesses to this
             * matrix, but only access to MatrixViews made on this
             * matrix. Secondly, once a matrix view has been created
             * any changes will not affect it.
             *
             * \param j Which column to hide.
             */
            void unhideCol(int j) { i_data->unhideCol(j); }

            /**
             * \brief Unhide a range of columns.
             *
             * \warning This does not affect direct accesses to this
             * matrix, but only access to MatrixViews made on this
             * matrix. Secondly, once a matrix view has been created
             * any changes will not affect it.
             *
             * \param begin Beginning of columns to unhide.
             * \param end   End of columns to unhide.
             */
            template<typename Itr>
            void unhideCols(Itr begin, Itr end)
	    { i_data->unhideCols(begin,end); }

            /// Get the set of the hidden cols;
            const std::set<int> &getHiddenCols() const {
		return i_data->i_hiddenCols;
	    }


            /**
             * \brief Hides a given row.
             *
             * \warning This does not affect direct accesses to this
             * matrix, but only access to MatrixViews made on this
             * matrix. Secondly, once a matrix view has been created
             * any changes will not affect it.
             *
             * \param j Which row to hide.
             */
            void hideRow(int j) { i_data->hideRow(j); }

            /**
             * \brief Hide a range of rows.
             *
             * \warning This does not affect direct accesses to this
             * matrix, but only access to MatrixViews made on this
             * matrix. Secondly, once a matrix view has been created
             * any changes will not affect it.
             *
             * \param begin Beginning of rows to hide.
             * \param end   End of rows to hide.
             */
            template<typename Itr>
            void hideRows(Itr begin, Itr end) { i_data->hideRows(begin,end); }

            /**
             * \brief Unhides a given row.
             *
             * \warning This does not affect direct accesses to this
             * matrix, but only access to MatrixViews made on this
             * matrix. Secondly, once a matrix view has been created
             * any changes will not affect it.
             *
             * \param j Which row to hide.
             */
            void unhideRow(int j) { i_data->unhideRow(j); }

            /**
             * \brief Unhide a range of rows.
             *
             * \warning This does not affect direct accesses to this
             * matrix, but only access to MatrixViews made on this
             * matrix. Secondly, once a matrix view has been created
             * any changes will not affect it.
             *
             * \param begin Beginning of rows to unhide.
             * \param end   End of rows to unhide.
             */
            template<typename Itr>
            void unhideRows(Itr begin, Itr end)
	    { i_data->unhideRows(begin,end); }

	    /// Get the set of the hidden rows;
	    const std::set<int> &getHiddenRows() const { 
		return i_data->i_hiddenRows;
	    }

            /**
             * \brief Reads the sets of hidden rows and columns from
             * file metadata (where applicable).
             *
             * The current sets of hidden rows and columns in this
             * matrix is replaced with that read from metadata.
             */
	    void readHiding() { i_data->readHiding(); }


            /// Is the data phased?
            bool isPhased() const { return i_data->isPhased(); }


            /**
             * \brief Look up a cell in the matrix.
             *
             * \param i		Row in the cell to access.
             * \param j		Column in the cell to access.
             * \returns		The cell at row \a i and column \a j.
             */
            const allele_t &operator()(int i, int j) const 
            {
                assert(0 <= i); assert(i < noRows());
                assert(0 <= j); assert(j < noCols());
                return i_data->i_data[i_data->translateIndex(i,j)]; 
            }

            /**
             * \brief Extract a submatrix.
             *
             * Extract a (contiguous) sub-matrix as a view.  The
             * sub-matrix is not copied, but what is obtained is a
             * view of a sub-part of this matrix.
             *
             * \param rowStart		The first row in the sub-matrix.
             * \param rowEnd		One past the last row in the
             *                          sub-matrix.
             * \param colStart		The first column in the sub-matrix.
             * \param colEnd		One past the last column in the
             *                          sub-matrix.
             *
             * \returns A view of a sub-matrix.
             */
            MatrixView subMatrix(int rowStart, int rowEnd,
				 int colStart, int colEnd) const;



            /**
             * \brief Extract a submatrix with all rows but a contiguous
             * subset of columns.
             *
             * Extract a (contiguous) sub-matrix as a view.  The sub-matrix
             * will contain all rows of this matrix, but only a subset of the
             * columns. The sub-matrix is not copied, but what is obtained is
             * a view of a sub-part of this matrix.
             *
             * \param colStart		The first column in the sub-matrix.
             * \param colEnd		One past the last column in the
             *                          sub-matrix.
             *
             * \returns A view of a sub-matrix.
             */
            MatrixView subMatrixColumns(int colStart, int colEnd) const;


            /**
             * \brief Extract a submatrix.
             *
             * Extract a sub-matrix as a view.  The sub-matrix is not copied,
             * but what is obtained is a view of a sub-part of this matrix.
             * The sub-matrix consists of a subset of this matrix rows; this
             * subset is specified through a sequence given by a pair of
             * iterators.
             *
             * \param rowsBegin		Iterator to the first element in the 
             *                          sequence of selected rows.
             * \param rowsEnd		Iterator to one past the last element
             *                          in the sequence of selected rows.
             *
             * \returns A view of a sub-matrix.
             */
            template <typename Itr>
            MatrixView view(Itr rowsBegin, Itr rowsEnd) const;

            /// Explicitly translate this matrix into a matrix view
            MatrixView view() const;
        };

	/**
	 * \brief Interface to a mutable matrix
	 *
	 * \see MatrixData
	 * \see ImmutableMatrix
	 * \see MatrixView
	 */
	class Matrix : public ImmutableMatrix {
	public:
	    explicit Matrix(MatrixData *data) : ImmutableMatrix(data) {};


            /// Set phased flag
	    void setPhased(bool phased) { i_data->setPhased(phased); }


            /**
             * \brief Stores the current sets of hidden rows and
             * columns in metadata (where applicable).
             * 
             * The sets are stored as file metadata. Any previously
             * stored values will be overwritten.
             */
	    void storeHiding() { i_data->storeHiding(); }

            /**
             * \brief Resize matrix to new dimensions.
             *
             * WARNING: resizing does *not* just expand or decrease
             * the dimensions -- the data will not (generally) be
             * preserved by a resizing.
             *
             * \param noRows	Number of rows in the resized matrix.
             * \param noCols	Number of columns in the resized matrix.
             */
            void resize(int noRows, int noCols)
            {
                i_data->resizeData(noRows, noCols);
                i_data->resetDimensions(noRows, noCols);
            }

            /** 
             * \brief Completely replace the contents of this matrix
             * with that from input.
             *
             * WARNING: May be a very expensive operation, time-vise.
             *
             * \param rhs Matrix to copy from.
             */
            void overwrite(const ImmutableMatrix &rhs);

	    using ImmutableMatrix::operator();
            /**
             * \brief Look up a cell in the matrix.
             *
             * \param i		Row in the cell to access.
             * \param j		Column in the cell to access.
             * \returns		The cell at row \a i and column \a j.
             */
	    allele_t &operator()(int i, int j)
            {
                assert(0 <= i); assert(i < noRows());
                assert(0 <= j); assert(j < noCols());
                return i_data->i_data[i_data->translateIndex(i,j)]; 
            }
	};


        /**
         * \brief A view of a sub-matrix.
         *
         * This class does not contain data itself, but simply
         * represents a (const) view on a data matrix. It is not
         * possible to modify the matrix through a view, though any
         * changes to the matrix will be reflected in the view.
         *
         * \see MatrixView
         */
        class MatrixView {
            /// The matrix containing the data
	    ImmutableMatrix i_matrix;
            /// The rows in the real data matrix this view considers
            /// its own rows.
            std::vector<int> i_rowIndices;
            /// The columns in the real data matrix this view
            /// considers its own columns.
            std::vector<int> i_colIndices;

        public:
            /**
             * \brief Construct a view on a matrix.
             *
             * Construct a (contiguous) sub-matrix view.  The
             * sub-matrix is not copied, but what is obtained is a
             * view of a sub-part of this matrix.
             *
             * \param matrix		The data matrix this will be a view on.
             * \param rowStart		The first row in the sub-matrix.
             * \param rowEnd		One past the last row in the
             *                          sub-matrix.
             * \param colStart		The first column in the sub-matrix.
             * \param colEnd		One past the last column in the
             *                          sub-matrix.
             */
            MatrixView(const ImmutableMatrix &matrix, 
                       int rowStart, int rowEnd, 
                       int colStart, int colEnd);

            /**
             * \brief Construct a view on a matrix.
             *
             * Construct a sub-matrix view from another view.  The
             * sub-matrix is not copied, but what is obtained is a
             * view of a sub-part of this matrix.
             *
             * \param view		The matrix view this will be a
             *                          (sub-)view on.
             * \param rowStart		The first row in the sub-matrix.
             * \param rowEnd		One past the last row in the
             *                          sub-matrix.
             * \param colStart		The first column in the sub-matrix.
             * \param colEnd		One past the last column in the
             *                          sub-matrix.
             */
            MatrixView(const MatrixView &view, 
                       int rowStart, int rowEnd, 
                       int colStart, int colEnd);


            /**
             * \brief Construct a view on a matrix.
             *
             * Construct a sub-matrix view from a matrix.  The
             * sub-matrix is not copied, but what is obtained is a
             * view of a sub-part of this matrix.  The sub-matrix
             * consists of a subset of this matrix rows; this subset
             * is specified through a sequence given by a pair of
             * iterators.
             *
             * \param matrix		The data matrix this will be a view on.
             * \param rowsBegin		Iterator to the first element in the
             *                          sequence of selected rows.
             * \param rowsEnd		Iterator to one past the last element 
             *                          in the sequence of selected rows.
             */
            template <typename Itr>
            MatrixView(const ImmutableMatrix &matrix, 
		       Itr rowsBegin, Itr rowsEnd);

            /// Number of rows in this matrix view.
            int noRows() const { return i_rowIndices.size(); }
            /// Number of columns in this matrix view.
            int noCols() const { return i_colIndices.size(); }

            /**
             * \brief Look up a cell in the matrix.
             *
             * \param i		Row in the cell to access.
             * \param j		Column in the cell to access.
             * \returns		The cell at row \a i and column \a j.
             */
            allele_t const operator()(int i, int j) const 
            {
                assert(0 <= i); assert(i < noRows());
                assert(0 <= j); assert(j < noCols());
                return i_matrix(i_rowIndices[i],i_colIndices[j]);
            }

            /**
             * \brief Get the row in the original matrix, that a given
             * row in the view correspond to.
             *
             * \param i		Row index in the view.
             * \returns		The row index in the underlying 
             *                  matrix corresponding to \a i.
             */
            int mappedRow(int i) const { return i_rowIndices.at(i); }
            /**
             * \brief Get the column in the original matrix, that a
             * given column in the view correspond to.
             *
             * \param j		Column index in the view.
             * \returns		The column index in the underlying
             *                  matrix corresponding to \a j.
             */
            int mappedCol(int j) const { return i_colIndices.at(j); }

            /**
             * \brief Get a vector mapping row indices into rows of the
             * underlying matrix.
             *
             * \warning This returns a mutable reference to the
             * internal data of this matrixview. Modifying this vector
             * modifies this MatrixView.
             *
             * \returns a reference to the internal vector keeping track of row
             *          in this MatrixView.
             */
            std::vector<int> &getRowMappings() { return i_rowIndices; }

            /**
             * \brief Get a vector mapping column indices into columns of the
             * underlying matrix.
             *
             * \warning This returns a mutable reference to the
             * internal data of this matrixview. Modifying this vector
             * modifies this MatrixView.
             *
             * \returns a reference to the internal vector keeping
             *          track of column in this MatrixView.
             */
            std::vector<int> &getColumnMappings() { return i_colIndices; }

            /**
             * \brief Extract a submatrix.
             *
             * Extract a sub-matrix from this view.  The sub-matrix is
             * not copied, but what is obtained is a view of a
             * sub-part of this matrix.
             *
             * \param rowStart		The first row in the sub-matrix.
             * \param rowEnd		One past the last row in the
             *                          sub-matrix.
             * \param colStart		The first column in the sub-matrix.
             * \param colEnd		One past the last column in the
             *                          sub-matrix.
             *
             * \returns A view of a sub-matrix.
             */
            MatrixView subMatrix (int rowStart, int rowEnd,
                                  int colStart, int colEnd);

        private:
            /// Method used in constructors.
            void removeHiddenRowsAndCols();
        };



        //******************** MatrixData members ********************

        template<typename Itr>
        void MatrixData::hideCols(Itr begin, Itr end)
        {
	    while (begin != end)
		{
		    hideCol(*begin);
		    ++begin;
		}
        }

        template<typename Itr>
        void MatrixData::unhideCols(Itr begin, Itr end)
        {
	    while (begin != end)
		{
		    unhideCol(*begin);
		    ++begin;
		}
        }

        template<typename Itr>
        void MatrixData::hideRows(Itr begin, Itr end)
        {
	    while (begin != end)
		{
		    hideRow(*begin);
		    ++begin;
		}
        }

        template<typename Itr>
        void MatrixData::unhideRows(Itr begin, Itr end)
        {
	    while (begin != end)
		{
		    unhideRow(*begin);
		    ++begin;
		}
        }

        inline MatrixView
        ImmutableMatrix::subMatrix(int rowStart, int rowEnd,
			  int colStart, int colEnd) const
        {
            return MatrixView(*this, rowStart, rowEnd, colStart, colEnd);
        }

        inline MatrixView
        ImmutableMatrix::subMatrixColumns(int colStart, int colEnd) const
        {
            return subMatrix(0, noRows(), colStart, colEnd);
        }

        inline MatrixView
        ImmutableMatrix::view() const
        {
            return subMatrix(0,noRows(),0,noCols()); 
        }

        template <typename Itr>
        inline MatrixView 
	ImmutableMatrix::view(Itr rowsBegin, Itr rowsEnd) const
        {
            return MatrixView(*this, rowsBegin, rowsEnd);
        }



        //******************** MatrixView members ********************

        template <typename Itr>
        MatrixView::MatrixView(const ImmutableMatrix &matrix, 
			       Itr rowStart, Itr rowEnd)
            : i_matrix(matrix), i_rowIndices(rowStart, rowEnd),
              i_colIndices(i_matrix.noCols())
        {
            for(int i =0; i<matrix.noCols(); i++){
                i_colIndices[i] = i;
            }

            removeHiddenRowsAndCols();
        }


        inline MatrixView
        MatrixView::subMatrix(int rowStart, int rowEnd,
                              int colStart, int colEnd)
        {
            return MatrixView(*this, rowStart, rowEnd, colStart, colEnd);
        }



        //******************** Printing matrices ********************

        /**
         * \brief Template printing something looking like a matrix
         *
         * \param os outputstream to print to
         * \param m  Matrix-look-alike to print
         */
        template <typename MatrixType>
        void printMatrix(std::ostream &os, const MatrixType &m)
        {
            for (int i = 0; i < m.noRows(); ++i)
            {
                for (int j = 0; j < m.noCols(); ++j)
                    os << int(m(i,j)) << ' ';
                os << '\n';
            }		    
        }

        /**
         * \brief Overloaded streaming operator for printing ImmutableMatrix
         *
         * \param os outputstream to print to
         * \param m  ImmutableMatrix to print
         * \returns  outputstream after m has been printed to it
         */
        inline std::ostream &
        operator<<(std::ostream &os, const ImmutableMatrix &m)
        {
            printMatrix(os, m);
            return os;
        }


        /**
         * \brief Overloaded streaming operator for printing Matrix
         *
         * \param os outputstream to print to
         * \param m  Matrix to print
         * \returns  outputstream after m has been printed to it
         */
        inline std::ostream &
        operator<<(std::ostream &os, const Matrix &m)
        {
            printMatrix(os, m);
            return os;
        }


        /**
         * \brief Overloaded streaming operator for printing MatrixView
         *
         * \param os outputstream to print to
         * \param m  MatrixView to print
         * \returns  outputstream after m has been printed to it
         */
        inline std::ostream &
        operator<<(std::ostream &os, const MatrixView &m)
        {
            printMatrix(os, m);
            return os;
        }
    }
}
#endif
