/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus, Denmark
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#undef NDEBUG
#include "snpfile.hh"
#include "matrix.hh"
#include "file_matrix.hh"
#include "timing.hh"

#include <fstream>

#include <cstdlib>



using namespace BiRC::SNPFile;



namespace
{
    std::string filename = "/tmp/benchmark_test.snp";
}



/** \file
 * \brief Simple test of access time on matrices.
 */



std::pair<TimeInterval, TimeInterval> readMatrixByColumn(const ImmutableMatrix &m)
{
    std::cout << "Testing " << m.noRows() << "x" << m.noCols() << " matrix." << std::endl;

    TimePoint start;

    long long checkSum = 0;
    //Read numbers back.
    for(int i = 0; i < m.noCols(); i++)
    {
        for(int j = 0; j < m.noRows(); j++)
            checkSum += m(j, i);
    }

    TimePoint mid;

    std::cout << "Checksum1 = " << checkSum << std::endl;

    checkSum = 0;
    //Read numbers back.
    for(int i = 0; i < m.noCols(); i++)
    {
        for(int j = 0; j < m.noRows(); j++)
            checkSum += m(j, i);
    }

    TimePoint end;

    std::cout << "Checksum2 = " << checkSum << std::endl;

    std::cout << (end - start) << std::endl;
    return std::make_pair(mid - start, end - mid);
}



std::pair<TimeInterval, TimeInterval> readMatrixByRow(const ImmutableMatrix &m)
{
    std::cout << "Testing " << m.noRows() << "x" << m.noCols() << " matrix." << std::endl;

    TimePoint start;

    long long checkSum = 0;
    //Read numbers back.
    for(int i = 0; i < m.noRows(); i++)
    {
        for(int j = 0; j < m.noCols(); j++)
            checkSum += m(i, j);
    }

    TimePoint mid;

    std::cout << "Checksum1 = " << checkSum << std::endl;

    checkSum = 0;
    //Read numbers back.
    for(int i = 0; i < m.noRows(); i++)
    {
        for(int j = 0; j < m.noCols(); j++)
            checkSum += m(i, j);
    }

    TimePoint end;

    std::cout << "Checksum2 = " << checkSum << std::endl;

    std::cout << (end - start) << std::endl;
    return std::make_pair(mid - start, end - mid);
}



std::pair<TimeInterval, TimeInterval> benchmarkMatrix(std::string accessTypeArg, const ImmutableMatrix &m)
{
    if(accessTypeArg == "row")
        return readMatrixByRow(m);
    else
        return readMatrixByColumn(m);
}



int main(int argc, char** argv)
{
    if(argc != 3)
    {
        std::cout << "Usage " << argv[0] << " resultFile (row|col)" << std::endl;
        return 1;
    }

    std::ofstream resultFile(argv[1], std::ios::app);

    std::string accessTypeArg(argv[2]);
    if(accessTypeArg != "row" && accessTypeArg != "col")
    {
        std::cout << "Illegal access pattern argument" << std::endl;
        return 1;
    }

    SNPFile snpFile(filename, true);
    ImmutableMatrix m = newReadOnlyFileMatrix(snpFile);
    std::pair<TimeInterval, TimeInterval> result = benchmarkMatrix(accessTypeArg, m);
    
    resultFile << accessTypeArg                             << " "
               << m.noRows()                                << " "
               << m.noCols()                                << " "
               << (size_t(m.noRows()) * size_t(m.noCols())) << " "
               << result.first.getUserTime()                << " "
               << result.first.getNiceTime()                << " "
               << result.first.getSystemTime()              << " "
               << result.first.getIdleTime()                << " "
               << result.first.getIOWaitTime()              << " "
               << result.first.getHarwareIRQTime()          << " "
               << result.first.getSoftwareIRQTime()         << " "
               << result.second.getUserTime()               << " "
               << result.second.getNiceTime()               << " "
               << result.second.getSystemTime()             << " "
               << result.second.getIdleTime()               << " "
               << result.second.getIOWaitTime()             << " "
               << result.second.getHarwareIRQTime()         << " "
               << result.second.getSoftwareIRQTime()        << std::endl;

    snpFile.close();

    resultFile.close();

    return 0;
}
