/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus, Denmark
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */


#undef NDEBUG			// enable asserts always


#include "matrix.hh"
#include "array_matrix.hh"
#include "file_matrix.hh"
#include "snpfile.hh"

#include "test_util.hh"

#include <iostream>
#include <sstream>



using namespace BiRC::SNPFile;



int
test(Matrix &m,
     Matrix &m2,
     Matrix &m3,
     Matrix &m4,
     Matrix &m5,
     Matrix &m6)
{
    std::ostringstream strstrm;
    std::string expected;

    expected = 
	"0 0 0 0 0 \n"
	"0 0 0 0 0 \n"
	"0 0 0 0 0 \n"
	"0 0 0 0 0 \n"
	"0 0 0 0 0 \n";
    strstrm << m;
    assert(expected == strstrm.str());
    strstrm.str("");

    m(0,1) = 1; 
    m(0,2) = 1; 
    m(1,0) = 1; 
    m(2,4) = 1;
    m(4,3) = 1;

    expected =
	"0 1 1 0 0 \n"
	"1 0 0 0 0 \n"
	"0 0 0 0 1 \n"
	"0 0 0 0 0 \n"
	"0 0 0 1 0 \n";
    strstrm << m;
    assert(expected == strstrm.str());
    strstrm.str("");


    MatrixView v = m.subMatrix(0,5,0,1);

    expected =
	"0 \n"
	"1 \n"
	"0 \n"
	"0 \n"
	"0 \n";
    strstrm << v;
    assert(expected == strstrm.str());
    strstrm.str("");



    MatrixView w = m.subMatrix(0,1,0,5);

    expected =
	"0 1 1 0 0 \n";
    strstrm << w;
    assert(expected == strstrm.str());
    strstrm.str("");

    expected =
	"0 1 1 \n"
	"1 0 0 \n";
    strstrm << m.subMatrix(0,2,0,3);
    assert(expected == strstrm.str());
    strstrm.str("");

    expected =
	"0 0 0 \n"
	"0 1 0 \n";
    strstrm << m.subMatrix(3,5,2,5);
    assert(expected == strstrm.str());
    strstrm.str("");

    expected =
	"0 \n"
	"0 \n";
    strstrm << m.subMatrix(3,5,2,5).subMatrix(0,2,2,3);
    assert(expected == strstrm.str());
    strstrm.str("");


    v = m.subMatrix(0,m.noRows(), 0,m.noCols());
    expected =
	"0 1 1 0 0 \n"
	"1 0 0 0 0 \n"
	"0 0 0 0 1 \n"
	"0 0 0 0 0 \n"
	"0 0 0 1 0 \n";
    strstrm << v;
    assert(expected == strstrm.str());
    strstrm.str("");

    v = m.subMatrix(0,3,          0,m.noCols());

    expected =
	"0 1 1 0 0 \n"
	"1 0 0 0 0 \n"
	"0 0 0 0 1 \n";
    strstrm << v;
    assert(expected == strstrm.str());
    strstrm.str("");

    w = m.subMatrix(3,m.noRows(), 0,m.noCols());
    expected =
	"0 0 0 0 0 \n"
	"0 0 0 1 0 \n";
    strstrm << w;
    assert(expected == strstrm.str());
    strstrm.str("");




    v = m.subMatrix(0,m.noRows(), 0,2);
    expected =
	"0 1 \n"
	"1 0 \n"
	"0 0 \n"
	"0 0 \n"
	"0 0 \n";
    strstrm << v;
    assert(expected == strstrm.str());
    strstrm.str("");

    w = m.subMatrix(0,m.noRows(), 2,m.noCols());
    expected =
	"1 0 0 \n"
	"0 0 0 \n"
	"0 0 1 \n"
	"0 0 0 \n"
	"0 1 0 \n";
    strstrm << w;
    assert(expected == strstrm.str());
    strstrm.str("");




    int rows[] = { 1, 3 };
    v = m.view(rows, rows+2);
    expected =
	"1 0 0 0 0 \n"
	"0 0 0 0 0 \n";
    strstrm << v;
    assert(expected == strstrm.str());
    strstrm.str("");


    // Test overwrite method.


    m4(0, 0) = 1;
    m4(1, 1) = 1;
    m4.setPhased(true);

    expected = 
        "1 0 \n"
        "0 1 \n";

    strstrm << m4;
    assert(expected == strstrm.str());
    strstrm.str("");
    assert(m4.isPhased());
   
    m5(1, 2) = 1;
    m5(2, 1) = 1;
    m5.setPhased(false);
   
    expected =
        "0 0 0 0 \n"
        "0 0 1 0 \n"
        "0 1 0 0 \n"
        "0 0 0 0 \n";
    
    strstrm << m5;
    assert(expected == strstrm.str());
    strstrm.str("");
    assert(!m5.isPhased());

    m4.overwrite(m5);

    strstrm << m4;
    assert(expected == strstrm.str());
    strstrm.str("");
    assert(!m4.isPhased());


    //
    // Test hiding of columns and rows
    //

    m6(0, 0) = 1;
    m6(1, 1) = 1;
    m6(2, 2) = 1;
    m6(3, 3) = 1;
    m6(4, 4) = 1;

    expected =
        "1 0 0 0 0 \n"
        "0 1 0 0 0 \n"
        "0 0 1 0 0 \n"
        "0 0 0 1 0 \n"
        "0 0 0 0 1 \n";

    strstrm << m6;
    assert(expected == strstrm.str());
    strstrm.str("");
    
    MatrixView v6 = m6.view();

    strstrm << v6;
    assert(expected == strstrm.str());
    strstrm.str("");


    m6.hideCol(1);
    v6 = m6.view();
    expected =
        "1 0 0 0 \n"
        "0 0 0 0 \n"
        "0 1 0 0 \n"
        "0 0 1 0 \n"
        "0 0 0 1 \n";
    strstrm << v6;
    assert(expected == strstrm.str());
    strstrm.str("");

    int h1[] = {3, 4, 5, 6};
    m6.hideCols(h1, h1 + 4);
    v6 = m6.view();
    expected =
        "1 0 \n"
        "0 0 \n"
        "0 1 \n"
        "0 0 \n"
        "0 0 \n";
    strstrm << v6;
    assert(expected == strstrm.str());
    strstrm.str("");

    m6.hideRow(1);
    v6 = m6.view();
    expected =
        "1 0 \n"
        "0 1 \n"
        "0 0 \n"
        "0 0 \n";
    strstrm << v6;
    assert(expected == strstrm.str());
    strstrm.str("");

    m6.hideRows(h1, h1 + 4);
    v6 = m6.view();
    expected =
        "1 0 \n"
        "0 1 \n";
    strstrm << v6;
    assert(expected == strstrm.str());
    strstrm.str("");

    int h2[] = {0, 1, 2, 4};
    m6.unhideCols(h2, h2 + 4);
    v6 = m6.view();
    expected =
        "1 0 0 0 \n"
        "0 0 1 0 \n";
    strstrm << v6;
    assert(expected == strstrm.str());
    strstrm.str("");

    m6.unhideRow(4);
    v6 = m6.view();
    expected =
        "1 0 0 0 \n"
        "0 0 1 0 \n"
        "0 0 0 1 \n";
    strstrm << v6;
    assert(expected == strstrm.str());
    strstrm.str("");

    m6.unhideCol(3);
    v6 = m6.view();
    expected =
        "1 0 0 0 0 \n"
        "0 0 1 0 0 \n"
        "0 0 0 0 1 \n";
    strstrm << v6;
    assert(expected == strstrm.str());
    strstrm.str("");

    int h3[] = {0, 1, 2};
    m6.unhideRows(h3, h3 + 3);
    v6 = m6.view();
    expected =
        "1 0 0 0 0 \n"
        "0 1 0 0 0 \n"
        "0 0 1 0 0 \n"
        "0 0 0 0 1 \n";
    strstrm << v6;
    assert(expected == strstrm.str());
    strstrm.str("");

    return 0;
}



int
main()
{
    int result = 0;
    Matrix am1(newArrayMatrix(5, 5));
    Matrix am2(newArrayMatrix(5, 5));
    Matrix am3(newArrayMatrix(6, 5));
    Matrix am4(newArrayMatrix(2, 2));
    Matrix am5(newArrayMatrix(4, 4));
    Matrix am6(newArrayMatrix(5, 5));
    result |= test(am1, am2, am3, am4, am5, am6);

    // testing slicing...
    ImmutableMatrix im1(am1);
    ImmutableMatrix im2(am2);
    ImmutableMatrix im3(am3);
    ImmutableMatrix im4(am4);
    ImmutableMatrix im5(am5);
    ImmutableMatrix im6(am6);


    int fd1 = 0;
    char* fn1 = makeTempFile(fd1);
    SNPFile f1(fn1, true);
    Matrix fm1(newReadWriteFileMatrix(f1));
    fm1.resize(5, 5);

    int fd2 = 0;
    char* fn2 = makeTempFile(fd2);
    SNPFile f2(fn2, true);
    Matrix fm2(newReadWriteFileMatrix(f2));
    fm2.resize(5, 5);

    int fd3 = 0;
    char* fn3 = makeTempFile(fd3);
    SNPFile f3(fn3, true);
    Matrix fm3(newReadWriteFileMatrix(f3));
    fm3.resize(6, 5);

    int fd4 = 0;
    char* fn4 = makeTempFile(fd4);
    SNPFile f4(fn4, true);
    Matrix fm4(newReadWriteFileMatrix(f4));
    fm4.resize(2, 2);

    int fd5 = 0;
    char* fn5 = makeTempFile(fd5);
    SNPFile f5(fn5, true);
    Matrix fm5(newReadWriteFileMatrix(f5));
    fm5.resize(4, 4);

    int fd6 = 0;
    char* fn6 = makeTempFile(fd6);
    SNPFile f6(fn6, true);
    Matrix fm6(newReadWriteFileMatrix(f6));
    fm6.resize(5, 5);

    result |= test(fm1, fm2, fm3, fm4, fm5, fm6);

    removeTempFile(fd6, fn6);
    removeTempFile(fd5, fn5);
    removeTempFile(fd4, fn4);
    removeTempFile(fd3, fn3);
    removeTempFile(fd2, fn2);
    removeTempFile(fd1, fn1);

    return result;
}
