/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2007,2008
 *               Bioinformatics Research Center,
 *               University of Aarhus, Denmark
 *               Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#ifndef METADATA_HH_INCLUDED
#define METADATA_HH_INCLUDED

#include <stdexcept>
#include <string>
#include <cstring>

/** \file
 * \brief Declaration of metadata concepts.
 */



namespace BiRC {
    namespace SNPFile {
        /**
         * \brief Class for storing one unit of metadata.
         */
        class MetadataBuffer {
            /// Number of bytes used.
            size_t i_size;
            /// Pointer to data buffer.
            char *i_data;
            /// Type of data
            std::string i_type;


        public:
            /**
             * \brief Construct new empty buffer with a given size.
             *
             * \param size Size of new buffer.
             */
            MetadataBuffer(size_t size);

            /**
             * \brief Construct new buffer with given data and size.
             *
             * \param data Pointer to new data.
             * \param size Size to copy.
             */
            MetadataBuffer(const char *data, size_t size);

            /**
             * \brief Destructs class. Removes buffer data.
             */
            ~MetadataBuffer();

            /**
             * \brief      Reset the size of this buffer.
             *             All data contained in this buffer will 
             *             (generally) be lost.
             *
             * \param size New size of buffer.
             */  
            void resize(size_t size);

            /**
             * \brief      Reset this buffer to new data and size.
             *             All previous data will be lost.
             *
             * \param data New data.
             * \param size Size of data.
             */  
            void reset(const char *data, size_t size);

            /**
             * \brief      Gets the size of the data in this buffer.
             *             Measured in bytes.
             *
             * \returns    Size of data in this buffer.
             */
            size_t size() const
            {
                return i_size;
            }

            /**
             * \brief      Gets a pointer to this buffers data.
             *             Use size, to find the data size.
             *
             * \returns    Pointer to data.
             */
            char *data()
            {
                return i_data;
            }

            /**
             * \brief      Gets a pointer to this buffers data.
             *             Use size, to find the data size.
             *
             * \returns    Pointer to data.
             */
            const char *data() const
            {
                return i_data;
            }

            /**
             * \brief      Set the type of metadata.
             * 
             * \param type Type of metadata

             */
            void setType(const std::string &type)
            {
                i_type = type;
            }

            /**
             * \brief      Get the type of metadata.
             * 
             */
            std::string getType() const
            {
                return i_type;
            }

        };





        /**
         * \brief Class for indirection between return type of
         *        metadataKeysBegin() and actual storage of keys.
         */
        class MetadataKeyInnerIterator
        {
        public:
            /// Virtual destructor.
            virtual ~MetadataKeyInnerIterator() {}
            /// Returns a pointer to a new copy of this iterator.
            virtual MetadataKeyInnerIterator *copy() const = 0;
            /// Returns value currently pointed at.
            virtual const std::string &operator*() const = 0;
            /// Makes this iterator point to next element.
            virtual MetadataKeyInnerIterator &operator++() = 0;
            /// Tests for equality.
            virtual bool operator==(const MetadataKeyInnerIterator &rhs) const = 0;
        };





        /**
         * \brief Iterator for iterating over metadata keys in a
         *        MetadataAccessor.
         *
         * This class actually wraps a MetadataKeyInnerIterator for
         * generalization.
         */
        class MetadataKeyIterator
        {
            /// Wrapped iterator.
            MetadataKeyInnerIterator *inner;

        public:
            /// Type of values iterated over.
            typedef std::string value_type;
            /// Type of distance between iterators.
            typedef int distance_type;

            /**
             * \brief Creates a new MetadataKeyIterator wrapping a
             *        given MetadataKeyInnerIterator.
             *
             * \param inner MetadataKeyInnerIterator to wrap.
             */
            MetadataKeyIterator(const MetadataKeyInnerIterator &inner)
                : inner(inner.copy())
            {
            }

            /// Copy constructor.
            MetadataKeyIterator(const MetadataKeyIterator &copy)
                : inner(copy.inner->copy())
            {
            }

            /// Destructor.
            ~MetadataKeyIterator()
            {
                delete inner;
            }

            /// Assignment operator.
            MetadataKeyIterator &operator=(const MetadataKeyIterator &rhs)
            {
                delete inner;
                inner = rhs.inner->copy();
                return *this;
            }

            /// Returns reference to value currently pointed to.
            const std::string &operator*() const
            {
                return inner->operator*();
            }

            /// Makes this iterator point to the next value.
            MetadataKeyIterator& operator++ ()
            {
                inner->operator++();
                return *this;
            }
 
            /// Makes a copy of this iterator. Makes this iterator
            /// point to next value. Then returns copy.
            MetadataKeyIterator operator++ (int)
            {
                MetadataKeyIterator ans = *this;
                ++(*this);
                return ans;
            }

            /// Test for equality.
            bool operator==(const MetadataKeyIterator &rhs) const
            {
                return *inner == *(rhs.inner);
            }

            /// Test for difference.
            bool operator!=(const MetadataKeyIterator &rhs) const
            {
                return !(*this == rhs);
            }
        };





        /**
         * \brief Interface for interaction with container storing
         * metadata.
         */
        class MetadataAccessor {
        public:
            ///Virtual destructor.
            virtual ~MetadataAccessor();


            /**
             * \brief Get a MetadataBuffer associated with a given key.
             *
             * \param key   Which buffer to get.
             *
             * \returns A metadatabuffer associated to given key.
             *
             * \exception MetadataMissingError If no metadata is associated
             * with given key.
             */
            virtual const MetadataBuffer&
            getMetadataBuffer(const std::string &key) const = 0;

            /**
             * \brief Get a MetadataBuffer associated with a given
             * key.
             *
             * If no buffer is associated with the given key an empty
             * one is created.
             *
             * \param key   Which buffer to get.
             *
             * \returns A metadatabuffer associated to given key.
             */
            virtual MetadataBuffer&
            getMetadataBuffer(const std::string &key) = 0;
    
            /**
             * \brief       Checks if metadata exists.
             *
             * \param key   What metadata to test.
             *
             * \returns     true if key has a valid mapping.
             */
            virtual bool containsMetadata(const std::string &key) const = 0;

            /**
             * \brief       Gets the size, in bytes, of a buffer.
             *
             * \param key   Which buffer to access.
             *
             * \returns     Size of metadata in bytes.
             *
             * \exception MetadataMissingError If no metadata is associated
             * with given key.
             */
            virtual size_t getMetadataSize(const std::string &key) const = 0;

            /**
             * \brief        Removes metadata from file.
             *
             * \param key    What data to remove.
             *
             * \exception MetadataMissingError If no metadata is associated
             * with given key.
             */
            virtual void removeMetadata(const std::string &key) = 0;

            /// Returns iterator pointing to beginning of metadata keys.
            virtual MetadataKeyIterator metadataKeysBegin() const = 0;

            /// Returns iterator pointing past end of metadata keys.
            virtual MetadataKeyIterator metadataKeysEnd() const = 0;
        };



        /**
         * \brief Exception used when problems occur while accessing metadata
         */
        class MetadataError : public std::runtime_error
        {
        public:
            /**
             * \brief Constructs a MetadataError.
             *
             * \param arg Description of why exceptions was thrown.
             */
            MetadataError(const std::string &arg)
                : std::runtime_error(arg)
            {}

            /**
             * \brief Constructs a MetadataError.
             *
             * \param what Description of what failed.
             * \param why  Reason why it failed.
             */
            MetadataError(const std::string &what, const std::string &why)
                : std::runtime_error(what + ": " + why)
            {}
        };



        /**
         * \brief Exception used trying to access non-existing metadata.
         */
        class MetadataMissingError : public MetadataError
        {
        public:
            /**
             * \brief Constructs a MetadataMissingError.
             *
             * \param arg Description of why exceptions was thrown.
             */
            MetadataMissingError(const std::string &arg)
                : MetadataError(arg)
            {}


            /**
             * \brief Constructs a MetadataMissingError.
             *
             * \param what Description of what failed.
             * \param why  Reason why it failed.
             */
            MetadataMissingError(const std::string &what, const std::string &why)
                : MetadataError(what, why)
            {}
        };
    }
}

#endif
