/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2007, 2008
 *               Bioinformatics Research Center,
 *               University of Aarhus, Denmark
 *               Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#ifndef METADATA_ACCESS_HH_INCLUDED
#define METADATA_ACCESS_HH_INCLUDED


#include "metadata.hh"
#include "archives.hh"
#include "type_trait.hh"
#include "serialization_trait.hh"


/** \file
 * \brief Methods for easy access to metadata.
 *
 * Uses boost serialization for archival. To support serialization
 * of a given class, see the documentation for boost. Alternatively
 * you can use template specialization on the templates in this file.
 */

namespace BiRC {
    namespace SNPFile {
        /**
         * \brief Allows retrieval of metadata.
         *
         * The only difference between this and fetchMetadata is the
         * way the result is returned.  getMetadata uses a reference
         * and fetchMetadata uses a return value. Using fetchMetadata
         * will probably be prettier in most situations, however
         * fetchMetadata may require the returned metadata to be
         * copied an extra time (depending on what optimizations the
         * compiler do), which may be very time consuming, depending
         * on the size of the return value. If you are handeling large
         * metadata you may want to use getMetadata.
         *
         * \param acc    Which metadataAccessor contains the metadata.
         * \param key    Index of desired metadata.
         * \param dest   Where to write metadata.
         *
         * \exception MetadataMissingError If no metadata is associated with key.
         */
        template<typename T>
        void getMetadata(const MetadataAccessor &acc,
			 const std::string &key, 
			 T &dest)
        {
            const MetadataBuffer &buffer = acc.getMetadataBuffer(key);
            const std::string str(buffer.data(), buffer.size());
            InputArchive ia(str);
            ia >> dest;
        }



        /**
         * \brief Allows retrieval of metadata.
         *
         * The only difference between this and getMetadata is the way
         * the result is returned.  getMetadata uses a reference and
         * fetchMetadata uses a return value. Using fetchMetadata will
         * probably be prettier in most situations, however
         * fetchMetadata may require the returned metadata to be
         * copied an extra time (depending on what optimizations the
         * compiler do), which may be very time consuming, depending
         * on the size of the return value.
         *
         * \param acc    Which MetadataAccessor to retrieve data from.
         * \param key    Index of desired metadata.
         * \returns      The desired metadata.
         *
         * \exception MetadataMissingError If no metadata is
         * associated with key.
         */
        template<typename T>
        T fetchMetadata(const MetadataAccessor &acc, const std::string &key)
        {
            T t;
            getMetadata(acc, key, t);
            return t;
        }



        /**
         * \brief Writes metadata.
         *
         * \param acc    Which metadata accessor to write to.
         * \param key    Which index to store metadata under.
         * \param src    What data to write.
         */
        template<typename T>
        void setMetadata(MetadataAccessor &acc,
			 const std::string &key,
			 const T &src)
        {
            std::string str;
            OutputArchive oa(str);
            oa << src;
            MetadataBuffer &buffer = acc.getMetadataBuffer(key);
            buffer.reset(str.c_str(), str.size());
            buffer.setType(TypeTrait<T>::typeName());
        }



        /**
         * \brief       Checks if metadata exists.
         *
         * \param acc   Metadata accessor to perform operation on.
         * \param key   Key to test for.
         *
         * \returns     true if key has a valid mapping.
         */
        inline bool containsMetadata(const MetadataAccessor &acc, const std::string &key)
        {
            return acc.containsMetadata(key);
        }



        /**
         * \brief       Gets the size, in bytes, of metadata (after serialization).
         *
         * \param acc   Metadata accessor to perform operation on.
         * \param key   Key of metadata whose size to measure.
         *
         * \returns     Size of metadata in bytes.
         *
         * \exception   MetadataMissingError If no metadata is associated
         * with given key.
         */
        inline size_t getMetadataSize(const MetadataAccessor &acc, const std::string &key)
        {
            return acc.getMetadataSize(key);
        }



        /**
         * \brief       Removes metadata from file.
         *
         * \param acc   Metadata accessor to perform operation on.
         * \param key   Key of data to remove.
         *
         * \exception   MetadataMissingError If no metadata is associated
         * with given key.
         */
        inline void removeMetadata(MetadataAccessor &acc, const std::string &key)
        {
            acc.removeMetadata(key);
        }



        /**
         * \brief       Returns an iterator indicating the first metadata key in the file.
         *
         * \param acc   Metadata accessor to iterate over.
         * \returns     Am iterator pointing to the beginning of metadata keys.
         */
        inline MetadataKeyIterator metadataKeysBegin(const MetadataAccessor &acc)
        {
            return acc.metadataKeysBegin();
        }



        /**
         * \brief       Returns an iterator indicating the end of metadata keys in the file.
         *
         * \param acc   Metadata accessor to iterate over.
         * \returns     Am iterator pointing just past the end of metadata keys.
         */
        inline MetadataKeyIterator metadataKeysEnd(const MetadataAccessor &acc)
        {
            return acc.metadataKeysEnd();
        }



        /**
         * \brief       Returns a string representation of the type of the metadata stored under the key.
         *
         * \param acc   MetadataAccessor containing metadata.
         * \param key   Key to metadata to get type of.
         *
         * \exception MetadataMissingError If no metadata is associated with key.
         */
        inline std::string metadataType(const MetadataAccessor &acc, const std::string key)
        {
            return acc.getMetadataBuffer(key).getType();
        }
    }
}

#endif
