#include <snpfile.hh>
#include <metadata_access.hh>

#include <map>
#include <string>

#include <iostream>

#include <unistd.h>

#include <boost/shared_ptr.hpp>



namespace BiRC
{
    namespace SNPFile
    {
        struct MetadataSerializer
        {
            virtual ~MetadataSerializer() {}

            virtual void serialize(SNPFile &file,
                                   const std::string &key,
                                   void *metadata) const = 0;
        };



        template<typename T>
        struct BoostMetadataSerializer : MetadataSerializer
        {
            virtual void serialize(SNPFile &file,
                                   const std::string &key,
                                   void *metadata) const
            {
                setMetadata(file, key, *reinterpret_cast<T *>(metadata));
            }
        };



        class MetadataCachePrototype
        {
            typedef boost::shared_ptr<void> voidPtr_t;
            typedef boost::shared_ptr<MetadataSerializer> serPtr_t;
            typedef std::pair<voidPtr_t, serPtr_t> cacheData_t;

            SNPFile file;
            std::map<std::string, cacheData_t > cacheMap;



        public:
            MetadataCachePrototype(const std::string &snpFilename)
                : file(snpFilename, true, false)
            {
            }



            ~MetadataCachePrototype();



            template <typename T>
            void cache(const std::string &key);



            template<typename T>
            T &get(const std::string &key);
        };



        template <typename T>
        void MetadataCachePrototype::cache(const std::string &key)
        {
            T *e = new T();

            if(file.containsMetadata(key))
                getMetadata(file, key, *e);

            cacheMap[key] = std::make_pair<voidPtr_t, serPtr_t>
                (voidPtr_t(e),
                 serPtr_t(new BoostMetadataSerializer<T>()));
        }



        template<typename T>
        T &MetadataCachePrototype::get(const std::string &key)
        {
            if(cacheMap.find(key) == cacheMap.end())
                cache<T>(key);

            return *reinterpret_cast<T *>(cacheMap[key].first.get());
        }



        MetadataCachePrototype::~MetadataCachePrototype()
        {
            std::map<std::string, cacheData_t>::const_iterator it;

            for(it = cacheMap.begin(); it != cacheMap.end(); ++it)
                it->second.second->serialize(file, it->first, it->second.first.get());

            file.close();
        }
    }
}



//Test...
int main(int argc, char **argv)
{
    using namespace BiRC::SNPFile;

    const std::string filename = "test_snpfile.snp";

    {   //First create a snpfile.
        SNPFile snpfile(filename, true);
        snpfile.close();
    }

    {   //Put some data into it.
        MetadataCachePrototype cache(filename);

        std::vector<int> &v = cache.get< std::vector<int> >("testInts");
        v.clear();
        v.push_back(42);
        v.push_back(1337);

        bool &b = cache.get< bool >("testBool");
        b = true;

        std::vector<int> &w = cache.get< std::vector<int> >("testInts");
        w.push_back(1);
        w.push_back(2);
        w.push_back(3);
    }

    {   //Read some data from it.
        MetadataCachePrototype cache(filename);

        std::vector<int> &v = cache.get< std::vector<int> >("testInts");
        bool &b = cache.get< bool >("testBool");

        std::cout << "Some ints..." << std::endl;

        std::vector<int>::const_iterator it;
        for(it = v.begin(); it != v.end(); ++it)
            std::cout << *it << std::endl;

        std::cout << "A bool..." << std::endl;

        std::cout << b << std::endl;
    }

    unlink(filename.c_str());    
}
