/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2007, 2008
 *               Bioinformatics Research Center,
 *               University of Aarhus, Denmark
 *               Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#ifndef METADATA_FILTERING_HH_INCLUDED
#define METADATA_FILTERING_HH_INCLUDED



#include "matrix.hh"



/** \file
 * Utility functions for removing metadata corresponding to
 * hidden rows and columns in matrices.
 */



namespace BiRC
{
    namespace SNPFile
    {
        /**
         * \brief Copies a collection while removing certain entries.
         *
         * \param filter Set containing indices to remove.
         * \param input Collection to copy from.
         * \param output Collection to copy to.
         */
        template<typename FilterSet,
                 typename InputCollection,
                 typename OutputCollection>
        void filterOnIndex(const FilterSet &filter,
                           const InputCollection &input,
                           OutputCollection &output)
        {
            typename FilterSet::value_type i = 0;
            typename InputCollection::const_iterator it = input.begin();

            for(; it != input.end(); ++it, ++i)
                if(filter.find(i) == filter.end())
                    output.insert(output.end(), *it);
        }

        /**
         * \brief Copies a collection while removing certain entries.
         *
         * \param filter Set containing indices to remove.
         * \param input Collection to copy from.
         * \return Copied collection.
         */
        template<typename FilterSet,
                 typename Collection>
        Collection filterOnIndex(const FilterSet &filter,
                                 const Collection &input)
        {
            Collection result;
            filterOnIndex(filter, input, result);
            return result;
        }



        /**
         * \brief Copies a collection while removing entries with
         * index corresponding to hidden rows.
         *
         * \param m Matrix with hidden rows.
         * \param input Collection to copy from.
         * \param output Collection to copy to.
         */
        template<typename InputCollection,
                 typename OutputCollection>
        void filterOnIndexInHiddenCols(const ImmutableMatrix &m,
                                       const InputCollection &input,
                                       OutputCollection &output)
        {
            filterOnIndex(m.getHiddenCols(), input, output);
        }

        /**
         * \brief Copies a collection while removing entries with
         * index corresponding to hidden rows.
         *
         * \param m Matrix with hidden rows.
         * \param input Collection to copy from.
         * \return Copied collection.
         */
        template<typename Collection>
        Collection filterOnIndexInHiddenCols(const ImmutableMatrix &m,
                                             const Collection &input)
        {
            return filterOnIndex(m.getHiddenCols(), input);
        }



        /**
         * \brief Copies a collection while removing entries with
         * index corresponding to hidden columns.
         *
         * \param m Matrix with hidden columns.
         * \param input Collection to copy from.
         * \param output Collection to copy to.
         */
        template<typename InputCollection,
                 typename OutputCollection>
        void filterOnIndexInHiddenRows(const ImmutableMatrix &m,
                                       const InputCollection &input,
                                       OutputCollection &output)
        {
            filterOnIndex(m.getHiddenRows(), input, output);
        }

        /**
         * \brief Copies a collection while removing entries with
         * index corresponding to hidden columns.
         *
         * \param m Matrix with hidden columns.
         * \param input Collection to copy from.
         * \return Copied collection.
         */
        template<typename Collection>
        Collection filterOnIndexInHiddenRows(const ImmutableMatrix &m,
                                             const Collection &input)
        {
            return filterOnIndex(m.getHiddenRows(), input);
        }



        





        /**
         * \brief Copies a collection while removing certain entries.
         *
         * \param filter Set containing values to remove.
         * \param input Collection to copy from.
         * \param output Collection to copy to.
         */
        template<typename FilterSet,
                 typename InputCollection,
                 typename OutputCollection>
        void filterOnContents(const FilterSet &filter,
                              const InputCollection &input,
                              OutputCollection &output)
                              
        {
            typename InputCollection::const_iterator it = input.begin();
            for(; it != input.end(); ++it)
                if(filter.find(*it) == filter.end())
                    output.insert(output.end(), *it);
        }

        /**
         * \brief Copies a collection while removing certain entries.
         *
         * \param filter Set containing values to remove.
         * \param input Collection to copy from.
         * \return Copied collection.
         */
        template<typename FilterSet,
                 typename Collection>
                 
        Collection filterOnContents(const FilterSet &filter,
                                    const Collection &input)
                                    
        {
            Collection result;
            filterOnContents(filter, input, result);
            return result;
        }



        /**
         * \brief Copies a collection while removing entries
         * corresponding to hidden rows.
         *
         * \param m Matrix with hidden rows.
         * \param input Collection to copy from.
         * \param output Collection to copy to.
         */
        template<typename InputCollection,
                 typename OutputCollection>
        void filterOnContentsInHiddenCols(const ImmutableMatrix &m,
                                          const InputCollection &input,
                                          OutputCollection &output)
                                          
        {
            filterOnContents(m.getHiddenCols(), input, output);
        }

        /**
         * \brief Copies a collection while removing entries
         * corresponding to hidden rows.
         *
         * \param m Matrix with hidden rows.
         * \param input Collection to copy from.
         * \return Copied collection.
         */
        template<typename Collection>
        Collection filterOnContentsInHiddenCols(const ImmutableMatrix &m,
                                                const Collection &input)
                                                
        {
            return filterOnContents(m.getHiddenCols(), input);
        }



        /**
         * \brief Copies a collection while removing entries
         * corresponding to hidden columns.
         *
         * \param m Matrix with hidden columns.
         * \param input Collection to copy from.
         * \param output Collection to copy to.
         */
        template<typename InputCollection,
                 typename OutputCollection>
        void filterOnContentsInHiddenRows(const ImmutableMatrix &m,
                                          const InputCollection &input,
                                          OutputCollection &output)
                                          
        {
            filterOnContents(m.getHiddenRows(), input, output);
        }

        /**
         * \brief Copies a collection while removing entries
         * corresponding to hidden columns.
         *
         * \param m Matrix with hidden columns.
         * \param input Collection to copy from.
         * \return Copied collection.
         */
        template<typename Collection>
        Collection filterOnContentsInHiddenRows(const ImmutableMatrix &m,
                                                const Collection &input)
                                                
        {
            return filterOnContents(m.getHiddenRows(), input);
        }









        /**
         * \brief Copies a map while removing entries with certain keys.
         *
         * \param filter Set containing keys to remove.
         * \param input Map to copy from.
         * \param output Map to copy to.
         */
        template<typename FilterSet,
                 typename InputPairAssociativeContainer, 
                 typename OutputPairAssociativeContainer>
        void filterOnKeys(const FilterSet &filter,
                          const InputPairAssociativeContainer &input,
                          OutputPairAssociativeContainer &output)
                          
        {
            typename InputPairAssociativeContainer::const_iterator it = input.begin();
            for(;it != input.end(); ++it)
                if(filter.find(it->first) == filter.end())
                    output[it->first] = it->second;
        }

        /**
         * \brief Copies a map while removing entries with certain keys.
         *
         * \param filter Set containing keys to remove.
         * \param input Map to copy from.
         * \return Copied map.
         */
        template<typename FilterSet,
                 typename PairAssociativeContainer>
        PairAssociativeContainer filterOnKeys(const FilterSet &filter,
                                              const PairAssociativeContainer &input)
                                              
        {
            PairAssociativeContainer result;
            filterOnKeys(filter, input, result);
            return result;
        }



        /**
         * \brief Copies a map while removing entries with
         * key corresponding to hidden columns.
         *
         * \param m Matrix with hidden columns.
         * \param input Map to copy from.
         * \param output Map to copy to.
         */
        template<typename InputPairAssociativeContainer,
                 typename OutputPairAssociativeContainer>
        void filterOnKeysInHiddenCols(const ImmutableMatrix &m,
                                      const InputPairAssociativeContainer &input,
                                      OutputPairAssociativeContainer &output)
                                      
        {
            filterOnKeys(m.getHiddenCols(), input, output);
        }

        /**
         * \brief Copies a map while removing entries with
         * key corresponding to hidden columns.
         *
         * \param m Matrix with hidden columns.
         * \param input Map to copy from.
         * \return Copied map.
         */
        template<typename PairAssociativeContainer>
        PairAssociativeContainer filterOnKeysInHiddenCols(ImmutableMatrix &m,
                                                          const PairAssociativeContainer &input)
                                                          
        {
            return filterOnKeys(m.getHiddenCols(), input);
        }



        /**
         * \brief Copies a map while removing entries with
         * key corresponding to hidden rows.
         *
         * \param m Matrix with hidden rows.
         * \param input Map to copy from.
         * \param output Map to copy to.
         */
        template<typename InputPairAssociativeContainer,
                 typename OutputPairAssociativeContainer>
        void filterOnKeysInHiddenRows(const ImmutableMatrix &m,
                                      const InputPairAssociativeContainer &input,
                                      OutputPairAssociativeContainer &output)
                                      
        {
            filterOnKeys(m.getHiddenRows(), input, output);
        }

        /**
         * \brief Copies a map while removing entries with
         * key corresponding to hidden rows.
         *
         * \param m Matrix with hidden rows.
         * \param input Map to copy from.
         * \return Copied map.
         */
        template<typename PairAssociativeContainer>
        PairAssociativeContainer filterOnKeysInHiddenRows(ImmutableMatrix &m,
                                                          const PairAssociativeContainer &input)
                                                          
        {
            return filterOnKeys(m.getHiddenRows(), input);
        }
    }
}



#endif
