/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus, Denmark
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#undef NDEBUG   //Always enable assertions



#include <cassert>

#include "metadata_access.hh"
#include "snpfile.hh"
#include "test_util.hh"

#include <vector>
#include <map>
#include <string>

#include <iostream>



using namespace BiRC::SNPFile;



int main()
{
    int fd = 0;
    char *fn = makeTempFile(fd);

    {
        SNPFile file(fn, true);
    
        int i = 31337;
        long l = 42;
        char c = 'h';
        std::vector<int> vi;
        for(int j = 0; j < 65535; j++)
            vi.push_back(j);

        std::vector<bool> vb;
        for(int j = 0; j < 300; j++)
            vb.push_back(j % 3 == 1);

        std::map<std::string, unsigned int> mm;
        mm["foo"] = 7;
        mm["bar"] = 9;
        mm["baz"] = 13;

        const std::string key0 = "Key 0";
        const std::string key1 = "Key 1";
        const std::string key2 = "Key 2";
        const std::string key3 = "Key 3";
        const std::string key4 = "Key 4";
        const std::string key5 = "Key 5";

        setMetadata(file, key0, i);
        setMetadata(file, key1, l);
        setMetadata(file, key2, c);
        setMetadata(file, key3, vi);
        setMetadata(file, key4, vb);
        setMetadata(file, key5, mm);

        assert(!file.containsMetadata(BiRC::SNPFile::TYPE_INFO_KEY));
        file.commit();
        assert(file.containsMetadata(BiRC::SNPFile::TYPE_INFO_KEY));
    
        std::map<std::string, std::string> types;
        getMetadata(file, BiRC::SNPFile::TYPE_INFO_KEY, types);
        assert(isTypeName<int>(types["Key 0"]));
        assert(isTypeName<long>(types["Key 1"]));
        assert(isTypeName<char>(types["Key 2"]));
        assert(isTypeName< std::vector< int > >(types["Key 3"]));
        assert(isTypeName< std::vector< bool > >(types["Key 4"]));
        bool assertTemp = isTypeName< std::map< std::string, unsigned int > >(types["Key 5"]);
        assert(assertTemp);
        assert(types.size() == 7);
    
        file.close();
    }

    {    
        SNPFile file2(fn, true);
    
        assert(file2.containsMetadata(BiRC::SNPFile::TYPE_INFO_KEY));
    
        std::map<std::string, std::string> types2;
        getMetadata(file2, BiRC::SNPFile::TYPE_INFO_KEY, types2);
        assert(isTypeName<int>(types2["Key 0"]));
        assert(isTypeName<long>(types2["Key 1"]));
        assert(isTypeName<char>(types2["Key 2"]));
        assert(isTypeName< std::vector< int > >(types2["Key 3"]));
        assert(isTypeName< std::vector< bool > >(types2["Key 4"]));
        bool assertTemp = isTypeName< std::map< std::string, unsigned int > >(types2["Key 5"]);
        assert(assertTemp);
        assert(types2.size() == 7);

        int i2 = 7;
        setMetadata(file2, "Key 0", i2);

        file2.close();
    }

    {    
        SNPFile file3(fn);
        assert(file3.containsMetadata(BiRC::SNPFile::TYPE_INFO_KEY));
    
        std::map<std::string, std::string> types3;
        getMetadata(file3, BiRC::SNPFile::TYPE_INFO_KEY, types3);
        assert(isTypeName<int>(types3["Key 0"]));
        assert(isTypeName<long>(types3["Key 1"]));
        assert(isTypeName<char>(types3["Key 2"]));
        assert(isTypeName< std::vector< int > >(types3["Key 3"]));
        assert(isTypeName< std::vector< bool > >(types3["Key 4"]));
        bool assertTemp = isTypeName< std::map< std::string, unsigned int > >(types3["Key 5"]);
        assert(assertTemp);
        assert(types3.size() == 7);

        const SNPFile &constFile3 = file3;

        assert(isTypeName<int>(constFile3.getMetadataBuffer("Key 0").getType()));
        assert(isTypeName<long>(constFile3.getMetadataBuffer("Key 1").getType()));
        assert(isTypeName<char>(constFile3.getMetadataBuffer("Key 2").getType()));
        assert(isTypeName< std::vector<int> >(constFile3.getMetadataBuffer("Key 3").getType()));
        assert(isTypeName< std::vector<bool> >(constFile3.getMetadataBuffer("Key 4").getType()));
        //Temp needed because C macros doesn't like ",".
        bool temp = isTypeName< std::map<std::string, unsigned int> >(constFile3.getMetadataBuffer("Key 5").getType());
        assert(temp);
    
        file3.close();
    }

    removeTempFile(fd, fn);

    return 0;
}
