/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2007, 2008
 *               Bioinformatics Research Center,
 *               University of Aarhus, Denmark
 *               Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#ifndef OUTPUT_ARCHIVE_HH_INCLUDED
#define OUTPUT_ARCHIVE_HH_INCLUDED

#include <string>

#include "archive_common.hh"



/** \file 
 * \brief Declaration of archive for efficient and portable
 * serialization of arbitrary data.
 */



namespace BiRC {
    namespace SNPFile {
        /**
         * \brief Archive for storing data efficiently in a portable way.
         */
        class OutputArchive
        {
            std::string &v;

        public:
            /**
             * \brief Constructs a archive storing data in a given string.
             *
             * \param dest String in which to store serialized data.
             */
            OutputArchive(std::string &dest)
                : v(dest)
            {
            }

            /**
             * \brief Saves bytes directly to stream. Beware of portability.
             *
             * \param src Where to read data from.
             * \param count Amount of data to read and write.
             */
            void saveBytes(const void *src, int count)
            {
                const char *charSrc = static_cast<const char *>(src);
                for(int i = 0; i < count; i++)
                    v.push_back(*(charSrc++));                    
            }

        private:
            ///Saves an integral type to this stream.
            template<typename T>
            void saveIntegral(T i)
            {
                uint64_t tmp = uint64_t(toLittleEndian(i));
                saveBytes(&tmp, sizeof(uint64_t));
            }

            ///Saves a floating point type to this stream.
            template<typename T>
            void saveFloat(T f)
            {
                T tmp = toLittleEndian(f);
                saveBytes(&tmp, sizeof(T));
            }
        public:

            /**
             * \brief Saves arbitrary data to this stream.
             *
             * The data is serialized using
             * BiRC::SNPFile::SerializationTrait<T>.save.
             * Refine that to serialize userdefined types.
             *
             * \param t Data to store.
             */
            template <typename T>
            void save(const T &t);
            
            ///Convenience operator. Streams data using the save function.
            template <typename T>
            OutputArchive &operator<< (const T& t)
            {
                save(t);
                return *this;
            }
            
            /**
             * \brief Convenience operator. Streams data using the
             * save function.
             *
             * Note the input is NOT const. This is to allow this
             * operator to be used in a common serialization function
             * for both serialization and deserialization. Use the <<
             * operator if you require constness.
             *
             * \param t Data to store.
             */
            template <typename T>
            OutputArchive &operator| (T& t)
            {
                return operator<<(t);
            }
        };

        template <typename T>
        void OutputArchive::save(const T &t)
        {
            SerializationTrait<T>().save(*this, t);
        }

#define BIRC_SNPFILE_MAKE_INTEGRAL_SAVE_FUNCTION(T)  \
        template<> inline                            \
        void OutputArchive::save(const T &t)         \
        {                                            \
            saveIntegral(t);                         \
        }

        BIRC_SNPFILE_MAKE_INTEGRAL_SAVE_FUNCTION(bool);
        BIRC_SNPFILE_MAKE_INTEGRAL_SAVE_FUNCTION(char);
        BIRC_SNPFILE_MAKE_INTEGRAL_SAVE_FUNCTION(unsigned char);
        BIRC_SNPFILE_MAKE_INTEGRAL_SAVE_FUNCTION(signed char);
        BIRC_SNPFILE_MAKE_INTEGRAL_SAVE_FUNCTION(unsigned short);
        BIRC_SNPFILE_MAKE_INTEGRAL_SAVE_FUNCTION(signed short);
        BIRC_SNPFILE_MAKE_INTEGRAL_SAVE_FUNCTION(unsigned int);
        BIRC_SNPFILE_MAKE_INTEGRAL_SAVE_FUNCTION(signed int);
        BIRC_SNPFILE_MAKE_INTEGRAL_SAVE_FUNCTION(unsigned long);
        BIRC_SNPFILE_MAKE_INTEGRAL_SAVE_FUNCTION(signed long);
        BIRC_SNPFILE_MAKE_INTEGRAL_SAVE_FUNCTION(unsigned long long);
        BIRC_SNPFILE_MAKE_INTEGRAL_SAVE_FUNCTION(signed long long);

#undef BIRC_SNPFILE_MAKE_INTEGRAL_SAVE_FUNCTION

#define BIRC_SNPFILE_MAKE_FLOAT_SAVE_FUNCTION(T)     \
        template<> inline                            \
        void OutputArchive::save(const T &t)         \
        {                                            \
            saveFloat(t);                            \
        }

        BIRC_SNPFILE_MAKE_FLOAT_SAVE_FUNCTION(float);
        BIRC_SNPFILE_MAKE_FLOAT_SAVE_FUNCTION(double);
        BIRC_SNPFILE_MAKE_FLOAT_SAVE_FUNCTION(long double);

#undef BIRC_SNPFILE_MAKE_FLOAT_SAVE_FUNCTION
    }
}
#endif
