/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus, Denmark
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */


#undef NDEBUG			// enable asserts always


#include "matrix.hh"
#include "array_matrix.hh"
#include "file_matrix.hh"
#include "snpfile.hh"

#include "test_util.hh"

#include <iostream>



using namespace BiRC::SNPFile;



/**
 * \brief Test write capability of matrix.
 *
 * \param m Matrix to test.
 */
void testMutableMatrix(Matrix &m)
{
    m.resize(2, 3);
    m(0, 0) = 1;
    m(0, 1) = 0;
    m(0, 2) = 3;
    m(1, 0) = 0;
    m(1, 1) = 1;
    m(1, 2) = 3;

    assert(m.noRows() == 2);
    assert(m.noCols() == 3);
    assert(m(0, 0) == 1);
    assert(m(0, 1) == 0);
    assert(m(0, 2) == 3);
    assert(m(1, 0) == 0);
    assert(m(1, 1) == 1);
    assert(m(1, 2) == 3);

    m.setPhased(false);
    assert(!m.isPhased());
    
    m.resize(3, 2);
    m(0, 0) = 1;
    m(0, 1) = 6;
    m(1, 0) = 2;
    m(1, 1) = 5;
    m(2, 0) = 3;
    m(2, 1) = 4;

    assert(m.noRows() == 3);
    assert(m.noCols() == 2);
    assert(m(0, 0) == 1);
    assert(m(0, 1) == 6);
    assert(m(1, 0) == 2);
    assert(m(1, 1) == 5);
    assert(m(2, 0) == 3);
    assert(m(2, 1) == 4);

    m.setPhased(true);
    assert(m.isPhased());
}



/**
 * \brief Test read capability of matrix.
 *
 * Assumes matrix has been filled with data by
 * testMutableMatrix.
 *
 * \param m Matrix to test.
 */
void testImmutableMatrix(const ImmutableMatrix &m)
{
    assert(m.noRows() == 3);
    assert(m.noCols() == 2);
    assert(m(0, 0) == 1);
    assert(m(0, 1) == 6);
    assert(m(1, 0) == 2);
    assert(m(1, 1) == 5);
    assert(m(2, 0) == 3);
    assert(m(2, 1) == 4);
    assert(m.isPhased());

}



/**
 * \brief Main method. Test different read/write capability of different types of matrices.
 *
 * \returns 0 indicating program success.
 */
int main()
{
    Matrix arrayMatrix(newArrayMatrix(0, 0));
    testMutableMatrix(arrayMatrix);
    testImmutableMatrix(arrayMatrix);


    int fd1;
    const char *fileName1 = makeTempFile(fd1);

    {
        //File read/write, matrix read/write => ok.
        SNPFile file1(fileName1, true);
	Matrix fileMatrix1(newReadWriteFileMatrix(file1));
        testMutableMatrix(fileMatrix1);
        testImmutableMatrix(fileMatrix1);
        file1.close();
    }

    {
        //File read/write, matrix read-only => ok, but read-only
        SNPFile file1(fileName1, true);
	ImmutableMatrix fileMatrix1(newReadOnlyFileMatrix(file1));
        testImmutableMatrix(fileMatrix1);
        file1.close();
    }

    {
        //File read-only, matrix read-only => ok, but read-only
        SNPFile file1(fileName1);
	ImmutableMatrix fileMatrix1(newReadOnlyFileMatrix(file1));
        testImmutableMatrix(fileMatrix1);
        file1.close();
    }

    {
        //File read-only, matrix read/write => fail.
        SNPFile file1(fileName1);
        try
        {
	    Matrix fileMatrix1(newReadWriteFileMatrix(file1));
            assert(!"Should not be possible to create read/write matrix on read-only file.");
        }
        catch(SNPFileReadOnlyError)
        {
        }
        file1.close();
    }
    
    removeTempFile(fd1, fileName1);

    return 0;
}
