/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus, Denmark
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include "metadata_access.hh"
#include "matrix.hh"
#include "array_matrix.hh"
#include "file_matrix.hh"
#include "test_util.hh"
#include "snpfile.hh"

#include <string>
#include <iterator>
#include <map>
#include <algorithm>
#include <numeric>

#include <iostream>



using namespace BiRC::SNPFile;



/**
 * \brief Class representing a column in a matrix
 */
class MatrixColumn
{
    Matrix i_data;
    int i_columnNu;



public:

    /**
     * \brief Creates a new matrix column from a given index into a given matrix.
     *
     * \param data     Matrix to get column from
     * \param columnNu Index of column
     */
    MatrixColumn(Matrix &data, int columnNu)
        : i_data(data),
          i_columnNu(columnNu)
    {}


    /// Copy constructor
    MatrixColumn(const MatrixColumn &rhs)
        : i_data(rhs.i_data),
          i_columnNu(rhs.i_columnNu)
    {
    }

    /// Subscript operator. Allows access to cells in column
    allele_t &operator[] (int row)
    {
        return i_data(row, i_columnNu);
    }

    /// Const subscript operator. Allow read-only access to cells in column
    const allele_t operator[] (int row) const
    {
        return i_data(row, i_columnNu);
    }

    /// Two columns are equal iff all cells are equal.
    bool operator== (const MatrixColumn &rhs) const
    {
        if (i_data != rhs.i_data)
	    return false;

        for (size_t i = 0; i < length(); i++)
            if (operator[](i) != rhs[i])
                return false;

        return true;
    }

    /// Two columns are different iff any cells are different.
    bool operator!= (const MatrixColumn &rhs) const
    {
        return !(*this == rhs);
    }

    /// Copies contents from one column to another.
    MatrixColumn operator= (const MatrixColumn &rhs)
    {
        assert(i_data == rhs.i_data);

        if(i_columnNu != rhs.i_columnNu)
            for(size_t i = 0; i < length(); i++)
                operator[](i) = rhs[i];

        return MatrixColumn(*this);
    }

    /// Height of this column.
    size_t length() const
    {
        return i_data.noRows();
    }
};



/**
 * \brief Iterator over the columns of a matrix.
 *
 * WARNING: Since a matrix is not actually represented as a list of
 * columns internally this does NOT actually return references to objects
 * contained in the matrix but temporary representations instead and thus
 * do NOT implement the complete interface demanded by STL. It is however
 * close enough to be usable in most instances.
 */
class MatrixColumnIterator
{
    mutable Matrix i_data;
    int i_next;

public:
    /// Signal to STL about capabilities of this iterator
    typedef std::random_access_iterator_tag   iterator_category;
    /// Types of values returned by dereferencing this iterator
    typedef MatrixColumn                      value_type;
    /// Type this iterator expects size to be represented by
    typedef size_t                            difference_type;
    /// Type of pointers to values represented by this iterator
    typedef value_type*                       pointer;
    /// Type of references to values represented by this iterator
    typedef value_type&                       reference;



    /// Default constructor. Required by STL.
    MatrixColumnIterator()
        : i_data(NULL),
          i_next(0)
    {}

    /**
     * \brief Constructs iterator for given column in given matrix
     *
     * \param data     Matrix to iterate over
     * \param columnNo Column in matrix
     */
    MatrixColumnIterator(Matrix &data, int columnNo)
        : i_data(data),
          i_next(columnNo)
    {}

    /**
     * \brief Dereference operator.
     *
     * WARNING: STL demands this is a reference to the actual
     * column in the matrix, however this is not true in this case,
     * simply because there are no such column object. Instead a
     * temporary representation is returned.
     */
    MatrixColumn operator* () const
    {
        return MatrixColumn(i_data, i_next);
    }

    /// Moves iterator to next column
    MatrixColumnIterator operator++ ()
    {
        return MatrixColumnIterator(i_data, ++i_next);        
    }

    /// Moves iterator to next column
    MatrixColumnIterator operator++ (int)
    {
        return MatrixColumnIterator(i_data, i_next++);
    }

    /// Moves iterator to previous column
    MatrixColumnIterator operator-- ()
    {
        return MatrixColumnIterator(i_data, --i_next);        
    }

    /// Moves iterator to previous column
    MatrixColumnIterator operator-- (int)
    {
        return MatrixColumnIterator(i_data, i_next--);
    }

    /// Moves iterator a number of columns
    MatrixColumnIterator operator += (int i)
    {
        i_next += i;
        return MatrixColumnIterator(*this);
    }

    /// Moves iterator a number of columns
    MatrixColumnIterator operator -= (int i)
    {
        return (*this) += -i;
    }

    /// Creates a new iterator at a offset of this
    MatrixColumnIterator operator+ (int i) const
    {
        return MatrixColumnIterator(i_data, i_next + i);
    }

    /// Creates a new iterator at a offset of this
    MatrixColumnIterator operator- (int i) const
    {
        return *this + (-i);
    }

    /// Creates a new iterator at a offset of this
    int operator- (const MatrixColumnIterator &rhs) const
    {
        return i_next - rhs.i_next;
    }

    /// Compares iterators. Iterator are compared by position
    bool operator< (const MatrixColumnIterator &rhs) const
    {
        return i_data == rhs.i_data && i_next < rhs.i_next;
    }

    /// Iterators are equal iff they represent same position in same matrix
    bool operator== (const MatrixColumnIterator &rhs) const
    {
        return i_data == rhs.i_data && i_next == rhs.i_next;
    }

    /// Iiterators are <= iff they share the matrix and their position is <=
    bool operator<= (const MatrixColumnIterator &rhs) const
    {
        return (*this < rhs) || (*this == rhs);
    }

    /// Iterator are different if thier matrix or their position is different
    bool operator!= (const MatrixColumnIterator &rhs) const
    {
        return !(*this == rhs);
    }

    /// Iterators are > if they share matrix but their position is >
    bool operator> (const MatrixColumnIterator &rhs) const
    {
        return rhs < *this;
    }

    /// Iterators are >= if they share matrix but their position is >=
    bool operator>= (const MatrixColumnIterator &rhs) const
    {
        return rhs <= *this;
    }
};



/**
 * \brief Operator overloading of + for MatrixColumnIterator. Does what you would expect.
 */
MatrixColumnIterator operator+ (int i, const MatrixColumnIterator& rhs)
{
    return rhs + i;
}



/**
 * \brief Represents map from marker names to columns in matrix
 */
class MarkerIDMap
{
    std::map<std::string, int> i_map;
    Matrix i_data;



public:
    /**
     * \brief Marker ID Map over a given matrix and from given metadata
     *
     * \param data  Matrix containing data to be marked
     * \param acc   Where to read marker metadata
     * \param key   Id of metadata to read
     */
    MarkerIDMap(Matrix &data, const MetadataAccessor &acc, const std::string &key)
        : i_data(data)
    {
        getMetadata(acc, key, i_map);
    }



    /**
     * \brief Looks up a column with a given name
     *
     * \param key   Id of metadata to read
     * \returns     A MatrixColumnIterator representing the given marker
     */
    MatrixColumnIterator find(const std::string &key)
    {
        return MatrixColumnIterator(i_data, i_map[key]);
    }
};



/**
 * \brief Functor printing a column in a matrix
 */
class Printer
{
public:

    /// Prints the given matrixcolumn
    void operator() (const MatrixColumn &col) const
    {
        for(size_t i = 0; i < col.length(); i++)
            std::cout << int(col[i]) << ' ';
        
        std::cout << std::endl;
    }

    /// Prints the column represented by the given iterator
    void operator() (const MatrixColumnIterator &col) const
    {
        (*this)(*col);
    }
};



/**
 * \brief Functor returning a matrix column with the boolean inverse of its input.
 */
class Invert
{
public:

    /// Computes and returns the elementvise boolean inverse of given column
    MatrixColumn operator() (MatrixColumn col) const
    {
        for(size_t i = 0; i < col.length(); i++)
            col[i] = !col[i];

        return col;
    }
};



/**
 * \brief Functor returning the sum of two matrix columns.
 */
class Sum
{
public:

    /// Computes and returns the elementvise sum of the two vectors
    MatrixColumn operator() (MatrixColumn lhs, MatrixColumn rhs) const
    {
        for(size_t i = 0; i < lhs.length(); i++)
            lhs[i] += rhs[i];

        return lhs;
    }
};



/**
 * \brief Main method exercising and testing above classes.
 */
int main()
{
    int filedesc = 0;
    char *filename = makeTempFile(filedesc);

    {
        SNPFile file(filename, true);
	Matrix m(newReadWriteFileMatrix(file));

        m.resize(5, 5);
        m(0,1) = 1; 
        m(0,2) = 1; 
        m(1,0) = 1; 
        m(2,4) = 1;
        m(4,3) = 1;

        std::map<std::string, int> map;
        map["first"] = 0; //Set name of column 0 in metadata to 'first'
        map["last"] = 4;  //Set name of column 4 in metadata to 'last'

        setMetadata(file, "columnId", map); //Store names
    }

    {
        SNPFile file(filename, true);
	Matrix m(newReadWriteFileMatrix(file));

        std::map<std::string, int> map;
        getMetadata(file, "columnId", map);

        MatrixColumnIterator it(m, map["first"]); //Iterator pointing at element named 'first'
        MatrixColumnIterator end(m, map["last"]); //Iterator pointing at element named 'last' 
        end++; //STL expects the end iterator to point just beynd the last element, not at it.
        MatrixColumnIterator mid = it + 2; //Point to column 2 beyond the one named 'first'
        Printer p;

        std::for_each(it, end, p); //Print matrix from 'first' column to 'end' column

        std::cout << std::endl;

        p(std::find(it, end, *mid)); //Find column equal to mid between first and last and print it

        std::cout << std::endl;

        Invert t;
        std::transform(it, end, it, t); //Boolean invert everything between 'first' and 'last'
        std::for_each(it, end, p); //Print matrix from 'first' column to 'end' column

        Sum s;
	Matrix res(newArrayMatrix(5, 1));
        for(int i = 0; i < res.noRows(); i++)
            res(i, 0) = 0;

        std::cout << std::endl;

        p(std::accumulate(it, end, MatrixColumn(res, 0), s)); //Calculate sum vector of all columns
    }

    removeTempFile(filedesc, filename);
}
