/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2007, 2008
 *               Bioinformatics Research Center,
 *               University of Aarhus, Denmark
 *               Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#ifndef SERIALIZATION_TRAIT_HH_INCLUDED
#define SERIALIZATION_TRAIT_HH_INCLUDED



#include "archive_common.hh"

#include <boost/scoped_array.hpp>

#include <string>
#include <vector>
#include <list>
#include <set>
#include <map>



/**
 * \file serialization_trait.hh
 *
 * \brief Defines convenience macros for creating SerializationTraits
 * and declares SerializationTraits for common STL classes.
 */



namespace BiRC
{
    namespace SNPFile
    {
        //XXXXXXXXXX Convenience macros XXXXXXXXXXXXXXXXXXXXXXXXXX

        /**
         * \brief Defines load and save functions to use a common
         * serialization function.
         *
         * This macro must be used inside the declaration of a
         * SerializationTrair<T> struct.  The common serialization
         * function must have the signature: template<typename A>
         * static void serialize(A &a, T &t), where T is the type to
         * be serialized.
         *
         * \param T Name of type to be serialized.
         */
#define BIRC_SNPFILE_USE_COMMON_SERIALIZATION_FUNCTION(T)       \
        static void load(InputArchive &ia, T &t)                \
        {                                                       \
            serialize(ia, t);                                   \
        }                                                       \
        static void save(OutputArchive &oa, const T &t)         \
        {                                                       \
            serialize(oa, const_cast<T &>(t));                  \
        }

        /**
         * \brief Defines a SerializationTrait<T> to use a common
         * class member function for both serialization and
         * deserialization.
	 *
	 * This version does not export the typename for use with the
         * SNPFile metadata typesystem.  This leaves the type trait
         * specification up to the application programmer.
         *
         * This macro must be used in the global namespace.
         * 
         * The common class member serialization function must have
         * the signature:
	 *
         * template<typename A> static void serialize(A &a, T &t)
	 *
         * where T is the type to be serialized.
         *
         * \param T Name of type to be serialized.
         */
#define BIRC_SNPFILE_INTRUSIVE_SERIALIZATION_NO_TYPE(T)         \
        namespace BiRC { namespace SNPFile {                    \
        template<>                                              \
        struct SerializationTrait<T>                            \
        {                                                       \
            template<typename A>                                \
            static void serialize(A &a, T &t)                   \
            {                                                   \
                t.serialize(a);                                 \
            }                                                   \
            BIRC_SNPFILE_USE_COMMON_SERIALIZATION_FUNCTION(T)   \
        };                                                      \
        }}

        /**
         * \brief Defines a SerializationTrait<T> to use a common
         * class member function for both serialization and
         * deserialization and exports the typename for use with the
         * SNPFile metadata typesystem.
         *
         * The common class member serialization function must have
         * the signature:
	 *
         * template<typename A> static void serialize(A &a, T &t)
	 *
         * where T is the type to be serialized.
         *
         * \param T Name of type to be serialized.
         */
#define BIRC_SNPFILE_INTRUSIVE_SERIALIZATION(T)                 \
        BIRC_SNPFILE_INTRUSIVE_SERIALIZATION_NO_TYPE(T)         \
        BIRC_SNPFILE_EXPORT_TYPE(T)


        /**
         * \brief Defines a SerializationTrait<T> to use a global
         * serialization function template for both serialization and
         * deserialization.
	 *
	 * This version does not export the typename for use with the
         * SNPFile metadata typesystem.  This leaves the type trait
         * specification up to the application programmer.
         *
         * This macro must be used in the global namespace.
         * 
         * The global serialization function must have the signature:
	 *
         * template<typename A> static void serialize(A &a, T &t)
	 *
         * where T is the type to be serialized.
         *
         * \param T Name of type to be serialized.
         */
#define BIRC_SNPFILE_NONINTRUSIVE_SERIALIZATION_NO_TYPE(T)      \
        namespace BiRC { namespace SNPFile {                    \
        template<>                                              \
        struct SerializationTrait<T>                            \
        {                                                       \
            BIRC_SNPFILE_USE_COMMON_SERIALIZATION_FUNCTION(T)   \
        };                                                      \
        }}

        /**
         * \brief Defines a SerializationTrait<T> to use a global
         * serialization function template for both serialization and
         * deserialization.
	 *
	 * This version does not export the typename for use with the
         * SNPFile metadata typesystem.  This leaves the type trait
         * specification up to the application programmer.
         *
         * This macro must be used in the global namespace.
         * 
         * The global serialization function must have the signature:
	 *
         * template<typename A> static void serialize(A &a, T &t)
	 *
         * where T is the type to be serialized.
         *
         * \param T Name of type to be serialized.
         */
#define BIRC_SNPFILE_NONINTRUSIVE_SERIALIZATION(T)              \
        BIRC_SNPFILE_NONINTRUSIVE_SERIALIZATION_NO_TYPE(T)      \
        BIRC_SNPFILE_EXPORT_TYPE(T)







        ///Protects a macro parameter containing one comma.
#define BIRC_SNPFILE_PROTECT_1(A, B) A, B
        ///Protects a macro parameter containing two commas.
#define BIRC_SNPFILE_PROTECT_2(A, B, C) A, B, C
        ///Protects a macro parameter containing three commas.
#define BIRC_SNPFILE_PROTECT_3(A, B, C, D) A, B, C, D






        //XXXXXXXXXXXXXXXXX STL serialization XXXXXXXXXXXXXXXXXXXXXXXXX

        ///Serialization of std::pair.        
        template<typename K, typename T>
        struct SerializationTrait< std::pair< K, T > >
        {
            ///(De)serializes a std::pair.        
            template<typename A>
            static void serialize(A &a, std::pair< K, T > &p)
            {
                a | p.first | p.second;
            }

            BIRC_SNPFILE_USE_COMMON_SERIALIZATION_FUNCTION(BIRC_SNPFILE_PROTECT_1(std::pair< K, T >))
        };



        ///Serialization of std::string.
        template<>
        struct SerializationTrait<std::string>
        {
            ///Deserializes a std::string.
            static void load(InputArchive &ia, std::string &s)
            {
                size_t length;
                ia >> length;
                boost::scoped_array<char> bytes(new char[length + 1]);
                ia.loadBytes(bytes.get(), length);
                bytes[length] = 0;
                s = std::string(bytes.get());
            }

            ///Serializes a std::string.
            static void save(OutputArchive &oa, const std::string &s)
            {
                oa << s.length();
                oa.saveBytes(s.c_str(), s.length());
            }
        };



        /**
         * \brief Deserialized an STL or similar type.
	 *
         * Requires c.insert(c.end(), X) works as expected. Also
         * requires T::size_type and T::value_type.
         *
         * \param ia Archive to load from.
         * \param c Container to store.
         */
        template<typename T>
        void loadSTL(InputArchive &ia, T &c)
        {
            typename T::size_type size;
            ia >> size;
            for(typename T::size_type i = 0; i < size; i++)
            {
                typename T::value_type temp;
                ia >> temp;
                c.insert(c.end(), temp);
            }
        }



        /**
         * \brief Serializes an STL or similar type.
	 *
         * Requires c.size(), c.begin(), c.end() and a T::const_iterator.
         * 
         * \param oa Output archive to save to.
         * \param c Container to store.
         */
        template<typename T>
        void saveSTL(OutputArchive &oa, const T &c)
        {
            oa << c.size();
            typename T::const_iterator it = c.begin(), end = c.end();
            for(; it != end; ++it)
                oa << *it;
        }



        /**
         * \brief Creates a SerializationTrait<T> using the loadSTL
         * and saveSTL functions.
         *
         * This macro should be used as
         * namespace BiRC{namespace SNPFile{ template<> BIRC_SNPFILE_MAKE_STL_SERIALIZATION_TRAIT(T) }}
         * where you can T can depend on template parameters specified
         * in the template<> part. If the type T contain commas you
         * can use the BIRC_SNPFILE_PROTECT macros.
         *
         * \param T Type to serialize.
         */
#define BIRC_SNPFILE_MAKE_STL_SERIALIZATION_TRAIT(T)            \
            struct SerializationTrait< T >                      \
            {                                                   \
                static void load(InputArchive &ia, T &c)        \
                {                                               \
                    loadSTL(ia, c);                             \
                }                                               \
                static void save(OutputArchive &oa, const T &c) \
                {                                               \
                    saveSTL(oa, c);                             \
                }                                               \
            };



        template<typename T>
        BIRC_SNPFILE_MAKE_STL_SERIALIZATION_TRAIT(std::vector<T>);

        template<typename T>
        BIRC_SNPFILE_MAKE_STL_SERIALIZATION_TRAIT(std::list<T>);

        template<typename T>
        BIRC_SNPFILE_MAKE_STL_SERIALIZATION_TRAIT(std::set<T>);



        ///SerializationTrait for std::map.
        template<typename K, typename T>
        struct SerializationTrait< std::map< K, T > >
        {
            ///Deserialization of a std::map.
            static void load(InputArchive &ia, std::map< K, T > &m)
            {
                typename std::map< K, T >::size_type size;
                ia >> size;
                for(typename std::map< K, T >::size_type i = 0; i < size; i++)
                {
                    K key;
                    T data;
                    ia >> key >> data;
                    m[key] = data;
                }
            }
        
            ///Serialization of a std::map.
            static void save(OutputArchive &oa, const std::map< K, T > &m)
            {
                saveSTL(oa, m);
            }
        };
    }
}



#endif
