/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus, Denmark
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */




#include "snpfile.hh"

#include "byte_order_handler.hh"

#include <map>
#include <iostream>
#include <string>
#include <stdexcept>

#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <cassert>
#include <cerrno>

#include <arpa/inet.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include <fcntl.h>

#include "metadata_access.hh"

/** \file
 *
 * \brief Low-level file I/O code.
 */


namespace {
    /**
     * \brief Rounds from up to nearest multiple of to.
     *
     * Finds minimal r such that from <= r*to. Returns r*to.
     *
     * \param from   Number to round up.
     * \param to     Number to round up to a multiple of.
     *
     * \returns      from rounded up.
     */
    template<typename T> T
    roundUp(T from, T to)
    {
	from += to - 1;
	from /= to;
	from *= to;
	return from;
    }
}


namespace BiRC {
    namespace SNPFile {





        //*************** Declaration of SNPFileImpl ***************

        /**
         * \brief Class representing a snp file. Stores a matrix and supports
         *        arbitrary metadata.
         */
        class SNPFile::SNPFileImpl : public MetadataAccessor
        {
            /// Copy constructor. Not implemented. Do not copy snpfiles.
            SNPFileImpl(const SNPFileImpl &snpFile);

            /// Assignment operator. Not implemented. Do not assign snpfiles.
            SNPFileImpl &operator= (SNPFileImpl &rhs);



        protected:

            /// Type for map from metadata keys to data
            typedef std::map<const std::string, MetadataBuffer *> metadata_map_t;



            /// File descriptor of snp file
            int i_fileDescriptor;

            /// Size of current file
            size_t i_fileSize;

            /// Name of file
            std::string i_filename;

            /// Address of mmaped file data
            char *i_address;

            /// Number of columns in matrix
            int32_t i_noCols;

            /// Number of rows in matrix
            int32_t i_noRows;

            /// Holds metadata
            metadata_map_t i_metadata;

            /// Is this file closed?
            bool i_isClosed;

            /// Magic word. Used to identify SNPFiles.
            static const uint32_t MAGIC_WORD;

            /// Current file version
            static const uint32_t CURRENT_VERSION;



            /// Size of file header
            size_t headerSize() const
            {
                return sizeof(MAGIC_WORD) + sizeof(CURRENT_VERSION) + sizeof(i_noRows) + sizeof(i_noCols);
            }

            /// Size of contained matrix
            size_t matrixSize() const
            {
                return roundUp(sizeof(allele_t) * static_cast<size_t>(i_noRows) * static_cast<size_t>(i_noCols), size_t(4));
            }

            /// Size of metadata header
            size_t calcMetadataHeaderSize() const;

            /// Size of metadata. Warning: O(n) time.
            size_t calcMetadataSize() const;

            /// Minimum size of sound snp file
            size_t minimumFileSize() const
            {
                return headerSize() + sizeof(uint32_t);
            }



            /// Common code shared by constructors. Should only be
            /// called from those.
            void init(int fileDescriptor, bool allowWriting);

            /// The value for the mmap prot argument
            virtual int mmapProtValue() const = 0;



            /// Simply closes the file, utility used by subclasses.
            void closeFile();

            /// Does most of destruction, necessary because
            /// destruction is common for subclasses, but need to be
            /// done in their destructor and not in this' classes.
            void preDestructor();

        public:
            /**
             * \brief Constructs a SNPFileImpl from a file, given a name.
             *
             * \param filename Name of file to open.
             *
             * \exception SNPFileError if the file could not be opened and read.
             */
            SNPFileImpl(const std::string &filename);

            /**
             * \brief Ensures data is written to disk, closes file and
             * frees allocated memory.
             *
             * WARNING: This destructor may not throw exceptions and hence
             * will fail silently if unable to access the filesystem. Use
             * the close method to close the file in a safe way.
             */
            virtual ~SNPFileImpl();



            /**
             * \brief Ensures that all data is written to disk.
             *
             * \exception SNPFileError If the file could not be written.
             */
            virtual void commit() = 0;

            /**
             * \brief Ensures that all data is written to disk and
             * closes file.
             *
             * Once a SNPFile has been closed it cannot be opened
             * again and no more methods should be called on it.
             *
             * This has exactly same semantics as the destructor,
             * except this methods will throw appropriate exceptions,
             * whereas the destructor will choke those exceptions
             * resulting in silent data loss.
             *
             * \exception SNPFileError If the file could not be saved
             * and closed.
             */
            virtual void close() = 0;

            /**
             * \brief  Test if this file is read-only
             *
             * \returns true iff this file is read-only
             */
            virtual bool isReadOnly() = 0;

            /**
             * \brief   Access this files matrix data.
             *          Use this method for both reading and writing data.
             *
             * \returns Pointer to memory where the matrix is mmapped.
             */
            virtual allele_t *accessMatrixData() = 0;

            /**
             * \brief   Access this files matrix data.
             *          Use this method for reading data.
             *
             * \returns Pointer to memory where the matrix is mmapped.
             */
            const allele_t *accessMatrixData() const;

            /**
             * \brief Get width of stored matrix.
             *
             * \returns Number of columns in stored matrix.
             */
            int32_t noCols() const
            {
                return i_noCols;
            }

            /**
             * \brief Get height of stored matrix.
             *
             * \returns Number of rows in stored matrix.
             */
            int32_t noRows() const
            {
                return i_noRows;
            }

            /**
             * \brief Sets width and height of matrix.  All data
             * currently in the matrix will (generally) be lost.
             *
             * \param noCols New width of matrix.
             * \param noRows New height of matrix.
             *
             * \exception SNPFileError If the file could not be resized.
             */
            virtual void resizeMatrix(int32_t noRows, int32_t noCols) = 0;



            /**
             * \brief Get a MetadataBuffer associated with a given
             * key.
             *
             * If no buffer is associated with the given key a
             * MetadataError exception is thrown.
             *
             * \param key   Which buffer to get.
             *
             * \returns A metadatabuffer associated to given key.
             *
             * \exception MetadataMissingError If no metadata is
             * associated with given key.
             */
            virtual const MetadataBuffer& getMetadataBuffer(const std::string &key) const;

            /**
             * \brief Get a MetadataBuffer associated with a given
             * key.
             *
             * If no buffer is associated with the given key an empty
             * one is created.
             *
             * \param key   Which buffer to get.
             *
             * \returns A metadatabuffer associated to given key.
             */
            virtual MetadataBuffer& getMetadataBuffer(const std::string &key) = 0;
    
            /**
             * \brief       Checks if metadata exists.
             *
             * \param key   What metadata to test.
             *
             * \returns     true if has a valid mapping.
             */
            virtual bool containsMetadata(const std::string &key) const;

            /**
             * \brief       Gets the size, in bytes, of stored metadata.
             *              If no metadata with key exist, exception is thrown.
             *
             * \param key   Which metadata to access.
             *
             * \returns     Size of metadata in bytes.
             *
             * \exception MetadataMissingError If no metadata is associated
             * with given key.
             */
            virtual size_t getMetadataSize(const std::string &key) const;

            /**
             * \brief Removes metadata from file.  If metadata with
             *        this key don't exist, exception is thrown.
             *
             * \param key What data to remove.
             *
             * \exception MetadataMissingError If no metadata is associated
             *            with given key.
             */
            virtual void removeMetadata(const std::string &key) = 0;

            /**
             * \brief Class for wrapping a
             *        metadata_map_t::const_iterator into a
             *        MetadataKeyInnerIterator.
             */
            class MapMetadataKeyInnerIterator : public MetadataKeyInnerIterator
            {
                metadata_map_t::const_iterator inner;

            public:
                /// Creates a new MapMetadataKeyInnerIterator wrapping a given Map iterator.
                MapMetadataKeyInnerIterator(const metadata_map_t::const_iterator &inner)
                    : inner(inner)
                {
                }

                /// Virtual destructor.
                virtual ~MapMetadataKeyInnerIterator() {}

                virtual MetadataKeyInnerIterator *copy() const
                {
                    return new MapMetadataKeyInnerIterator(inner);
                }

                virtual const std::string &operator*() const
                {
                    return inner->first;
                }

                virtual MetadataKeyInnerIterator &operator++()
                {
                    ++inner;
                    return *this;
                }

                virtual bool operator==(const MetadataKeyInnerIterator &rhs) const
                {
                    const MapMetadataKeyInnerIterator *other
                        = dynamic_cast<const MapMetadataKeyInnerIterator *>(&rhs);
                    return other && inner == other->inner;
                }
            };

            /// Returns iterator pointing to beginning of metadata keys.
            MetadataKeyIterator metadataKeysBegin() const
            {
                return MetadataKeyIterator(MapMetadataKeyInnerIterator(i_metadata.begin()));
            }

            /// Returns iterator pointing past end of metadata keys.
            MetadataKeyIterator metadataKeysEnd() const
            {
                return MetadataKeyIterator(MapMetadataKeyInnerIterator(i_metadata.end()));
            }



        protected:
            /**
             * \brief Used to signal changes to data pointer
             *
             * Whenever the file is re-memory-mapped the matrix data
             * pointer may change. This signal is used to communicate
             * this to users of the matrix.
             */
            mutable signal_t i_dataPointerChangeSignal;            



        public:

            /**
             * \brief Adds a listener to be notified when matrix
             * pointer changes.
             *
             * The SNPFile may, for whatever reason, have to remap the
             * matrix, which may cause the pointer returned by
             * accesMatrixData to change. Use this method to add a
             * subscriber and receive notifications when this happens.
             * It is the subscribers responsibility to unsubscribe
             * when it is garbage collected. This can be done by
             * explicitly calling removeMatrixDataListener, by calling
             * disconnect on the returned connection or by deriving
             * from boost::signals::trackable.  See the boost::signals
             * documentation for details.
             *
             * \param subscriber Functor to be called when pointer changes.
             *
             * \returns          Object representing connection.
             */
            boost::signals::connection addMatrixDataListener(signal_t::slot_function_type subscriber) const
            {
                return i_dataPointerChangeSignal.connect(subscriber);
            }

            /**
             * \brief Removes a matrix data subscriber.
             *
             * See documentation of addMatrixDataListener for details.
             *
             * \param subscriber Connection to subscriber to remove.
             */
            void removeMatrixDataListener(boost::signals::connection subscriber) const
            {
                subscriber.disconnect();
            }
        };




        //******************** Declaration of ReadWriteSNPFileImpl ********************

        /**
         * \brief Class providing an implementation of SNPFile,
         * allowing both reading and writing
         */

        class SNPFile::ReadWriteSNPFileImpl : public SNPFile::SNPFileImpl
        {
            /**
             * \brief Changes the size of the entire file.
             *
             * Changes the size of the file on disk.  Does not effect
             * metadata, since that is buffered in memory but will
             * delete matrix data if matrix does not fit in new
             * filesize.
             *
             * \param newSize   New size of file.
             *
             * \exception SNPFileError if the file could not be resized.
             */
            void resizeFile(size_t newSize);

        protected:
            /// The value for the mmap prot argument
            int mmapProtValue() const;



        public:
            /// Creates a new ReadWriteSNPFileImpl open on given filename.
            ReadWriteSNPFileImpl(const std::string &filename, bool createFile, int mode);

            /// Virtual destructor.
            virtual ~ReadWriteSNPFileImpl();

            virtual void commit();

            virtual void close();

            virtual bool isReadOnly();

            virtual allele_t *accessMatrixData();

            virtual void resizeMatrix(int32_t noRows, int32_t noCols);

            virtual MetadataBuffer& getMetadataBuffer(const std::string &key);
    
            virtual void removeMetadata(const std::string &key);
        };




        //******************** Declaration of ReadOnlySNPFileImpl ********************

        /**
         * \brief Class providing a read-only implementation of SNPFile.
         */

        class SNPFile::ReadOnlySNPFileImpl : public SNPFile::SNPFileImpl
        {
        protected:
            /// The value for the mmap prot argument
            int mmapProtValue() const;



        public:
            /// Creates a new ReadOnlySNPFileImpl open on a given filename.
            ReadOnlySNPFileImpl(const std::string &filename);

            /// Virtual destructor.
            virtual ~ReadOnlySNPFileImpl();

            virtual void commit();

            virtual void close();

            virtual bool isReadOnly();

            virtual allele_t *accessMatrixData();

            virtual void resizeMatrix(int32_t noRows, int32_t noCols);

            virtual MetadataBuffer& getMetadataBuffer(const std::string &key);
    
            virtual void removeMetadata(const std::string &key);
        };




        //*************** Misc. Error handling ***************

        namespace {
            /// Throws an appropriate exception if isError
            inline void handleStdLibError(bool isError, const std::string &errorMsg)
            {
                if(isError)
                    throw SNPFileError(errorMsg, strerror(errno));
            }

            /// Throws an appropriate exception if isError
            inline void handleError(bool isError, const std::string &errorMsg)
            {
                if(isError)
                    throw SNPFileError(errorMsg);
            }

            /// Throws an appropriate exception if file is not found.
            inline void handleFileNotFound(const std::string &filename)
            {
                throw SNPFileNotFoundError("Cannot open file \"" + filename + "\". File not found.");
            }

            /// Throws an appropriate exception if isError
            inline void handleMetadataMissingError(bool isError, const std::string &errorMsg, const std::string &key)
            {
                if(isError)
                    throw MetadataMissingError(errorMsg, "Cannot find metadata with key \"" + key + "\"");
            }

            /// Throws an approptiate exception if i_isClosed
            inline void checkClosed(bool i_isClosed, const std::string &filename)
            {
                if(i_isClosed)
                    throw SNPFileClosedError("Trying to use a closed file: \"" + filename + "\"");
            }

            /// Throws an appropriate exception if accessing read-only file
            inline void handleReadOnlyError(const std::string &what, const std::string &filename)
            {
                throw SNPFileReadOnlyError("Cannot " + what, "Only read access permitted on \"" + filename + "\"");
            }
        }




        //*************** Definition of SNPFileImpl members ***************

        const uint32_t SNPFile::SNPFileImpl::MAGIC_WORD = 0xB10550C;
        const uint32_t SNPFile::SNPFileImpl::CURRENT_VERSION = 0;



        SNPFile::SNPFileImpl::SNPFileImpl(const std::string &filename)
            : i_filename(filename)
        {
        }



        void SNPFile::SNPFileImpl::init(int fileDescriptor, bool allowWriting)
        {
            if(fileDescriptor == -1)
            {
                if(errno == ENOENT)
                    handleFileNotFound(i_filename);
                else
                    handleStdLibError(true, "Cannot open file \"" + i_filename + "\"");
            }

            i_fileDescriptor = fileDescriptor;

            i_isClosed = false;



            struct stat fileStat;
            handleStdLibError(fstat(fileDescriptor, &fileStat) == -1, "Unable to open file \"" + i_filename + "\": Cannot stat");
            i_fileSize = static_cast<size_t>(fileStat.st_size);


            if(i_fileSize == 0 && allowWriting) //If new file.
            {
                //Create file.
                char fileContents[minimumFileSize()];
                WriteOrderHandler newFile(fileContents);
                newFile << MAGIC_WORD << CURRENT_VERSION << uint32_t(0) << uint32_t(0) << uint32_t(0);
                i_fileSize = minimumFileSize();
                handleStdLibError(write(fileDescriptor, fileContents, i_fileSize) == -1, "Unable to create new file \"" + i_filename + "\": Cannot write");
                handleStdLibError(fsync(fileDescriptor) == -1, "Unable to create new file: Cannot fsync");
            }

            handleError(i_fileSize < minimumFileSize(), "Unable to open file \"" + i_filename + "\": Illegal size for a snp file");

            i_address = static_cast<char *>(mmap(0, i_fileSize, mmapProtValue(), MAP_SHARED, fileDescriptor, 0));

            handleStdLibError(i_address == MAP_FAILED, "Unable to open file \"" + i_filename + "\": Cannot mmap");



            ReadOrderHandler file(i_address);
            uint32_t magicWord;
            uint32_t version;
            file >> magicWord >> version >> i_noRows >> i_noCols;
            handleError(magicWord != MAGIC_WORD, "Unable to open file \"" + i_filename + "\": Wrong format");
            handleError(version > CURRENT_VERSION, "Unable to open file \"" + i_filename + "\": Unknown version");
            handleError(i_noRows < 0 || i_noCols < 0, "Unable to open file \"" + i_filename + "\": Illegal size of matrix");

            file.setPosition(i_address + headerSize() + matrixSize());

            uint32_t metadataEntries;
            file >> metadataEntries;

            std::string keys[metadataEntries];
            uint32_t addrs[metadataEntries];
            
            std::map<std::string, std::string> typeInfo;
            
            for(uint32_t i = 0; i < metadataEntries; i++)
            {
                keys[i] = std::string(file.getPosition());
                
                int len = keys[i].size() + 1;
                file.setPosition(file.getPosition() + len);
                file >> addrs[i];
            }

            for(uint32_t i = 0; i < metadataEntries; i++)
            {
                size_t size;
                if(i < metadataEntries - 1) {
                    size = addrs[i + 1] - addrs[i];
                } else {
                    size = i_fileSize - addrs[i];
                }
                
                MetadataBuffer *buffer = new MetadataBuffer(i_address + addrs[i], size);
                i_metadata[keys[i]] = buffer;
            }
            //all buffers read, restore types
            
            if(containsMetadata(BiRC::SNPFile::TYPE_INFO_KEY)) {
                getMetadata< std::map<std::string, std::string> >(*this, BiRC::SNPFile::TYPE_INFO_KEY, typeInfo);
                for(uint32_t i = 0; i < metadataEntries; i++)
                {
                    i_metadata[keys[i]]->setType(typeInfo[keys[i]]);
                }    
            }
            
        }



        SNPFile::SNPFileImpl::~SNPFileImpl()
        {
            metadata_map_t::iterator it;
            for(it = i_metadata.begin(); it != i_metadata.end(); ++it)
                delete it->second;
        }



        void SNPFile::SNPFileImpl::preDestructor()
        {
            //FIXME: Need better exception handling.

            try
            {
                if(!i_isClosed)
                    close();
            }
            catch(std::runtime_error &r)
            {
                assert(!"Destructor failed");
            }
        }



        namespace {
            //Used to avoid compiler thinking close
            //is a recursive call in the closeFile method.
            inline int stdClose(int fd)
            {
                return close(fd);
            }
        }



        void SNPFile::SNPFileImpl::closeFile()
        {
            handleStdLibError(munmap(i_address, i_fileSize) == -1, "Unable to close file \"" + i_filename + "\": Cannot munmap");
            i_address = NULL;
            i_dataPointerChangeSignal();
            handleStdLibError(stdClose(i_fileDescriptor) == -1, "Unable to close file \"" + i_filename + "\"");
            i_isClosed = true;
        }



        size_t SNPFile::SNPFileImpl::calcMetadataHeaderSize() const
        {
            size_t result = sizeof(uint32_t); //First an int: Count of entries.

            metadata_map_t::const_iterator it;
            for(it = i_metadata.begin(); it != i_metadata.end(); ++it)
            {
                result += it->first.size(); //Size of key.
                result += 1;                //Size of terminating zero in key.
                result += sizeof(uint32_t); //Size of index of metadata
            }

            return result;
        }



        size_t SNPFile::SNPFileImpl::calcMetadataSize() const
        {
            size_t result = calcMetadataHeaderSize();
            metadata_map_t::const_iterator it = i_metadata.begin();
            for(; it != i_metadata.end(); ++it)
                result += it->second->size();
          
            return result;
        }



        const allele_t *SNPFile::SNPFileImpl::accessMatrixData() const
        {
            //These casts are kind of ugly.
            //Internally we simply handle an array of bytes, 
            //elsewhere an allele_t might be anything.
            //We should be fine as long as a byte is 8 bits.
            return reinterpret_cast<allele_t*>(i_address + headerSize());
        }



        const MetadataBuffer& SNPFile::SNPFileImpl::getMetadataBuffer(const std::string &key) const
        {
            checkClosed(i_isClosed, i_filename);
            metadata_map_t::const_iterator it = i_metadata.find(key);
            handleMetadataMissingError(it == i_metadata.end(), "Trying to get nonexisting metadata", key);
            return *(it->second);
        }



        bool SNPFile::SNPFileImpl::containsMetadata(const std::string &key) const
        {
            checkClosed(i_isClosed, i_filename);

            return i_metadata.find(key) != i_metadata.end();
        }



        size_t SNPFile::SNPFileImpl::getMetadataSize(const std::string &key) const
        {
            checkClosed(i_isClosed, i_filename);

            metadata_map_t::const_iterator it = i_metadata.find(key);
            handleMetadataMissingError(it == i_metadata.end(),
                                       "Trying to get size of nonexisting metadata",
                                       key);
            return it->second->size();
        }





        //******************** Definition of ReadWriteSNPFileImpl members ********************

        SNPFile::ReadWriteSNPFileImpl::ReadWriteSNPFileImpl(const std::string &filename,
                                                            bool createFile,
                                                            int mode)
            : SNPFileImpl(filename)
        {
            int flags = 0;
            if(createFile)
                flags = O_RDWR | O_CREAT;
            else
                flags = O_RDWR;

            init(open(filename.c_str(), flags, mode), true);
        }



        SNPFile::ReadWriteSNPFileImpl::~ReadWriteSNPFileImpl()
        {
            preDestructor();
        }



        void SNPFile::ReadWriteSNPFileImpl::resizeFile(size_t newSize)
        {
            if(newSize != i_fileSize) //Don't resize if already right size.
            {
                handleStdLibError(munmap(i_address, i_fileSize) == -1,
                                  "Unable to resize file \"" + i_filename + "\": Cannot munmap");
                handleStdLibError(ftruncate(i_fileDescriptor, newSize) == -1,
                                  "Unable to resize file \"" + i_filename + "\": Cannot truncate");

                i_fileSize = newSize;
                i_address = static_cast<char*>(mmap(0, newSize, mmapProtValue(),
                                                    MAP_SHARED,
                                                    i_fileDescriptor,
                                                    0));

                handleStdLibError(i_address == MAP_FAILED,
                                  "Unable to resize file \"" + i_filename + "\": Cannot mmap");
                i_dataPointerChangeSignal();
            }
        }



        int SNPFile::ReadWriteSNPFileImpl::mmapProtValue() const
        {
            return PROT_READ | PROT_WRITE;
        }



        void SNPFile::ReadWriteSNPFileImpl::commit()
        {
            checkClosed(i_isClosed, i_filename);

            metadata_map_t::iterator it;
            
            //Now create a map of all types, and write it..  really
            //ugly to use the key to get the metadataBuffer, consider
            //making a new iterator we could in theory get the old
            //map(if one exists) and update that, but the complexity
            //isnt worth the speedup

            std::map<std::string, std::string> types;
            for(it = i_metadata.begin(); it != i_metadata.end(); ++it)
            {
                types[it->first] = getMetadataBuffer(it->first).getType();
            }
            types[TYPE_INFO_KEY] = "std::map< std::string, std::string >";
            //now add the map as metadata
            
            setMetadata(*this, TYPE_INFO_KEY, types);
            
            size_t newSize = headerSize() + matrixSize() + calcMetadataSize();
            
            resizeFile(newSize);

            WriteOrderHandler file(i_address);
            //Write header
            file << MAGIC_WORD << CURRENT_VERSION << i_noRows << i_noCols;
            //Matrix is already written, through mmap.
            //Write metadata            
            file.setPosition (i_address + headerSize() + matrixSize());
            file << uint32_t(i_metadata.size());
            char *nextPosition = i_address + headerSize() + matrixSize() + calcMetadataHeaderSize();

            
            for(it = i_metadata.begin(); it != i_metadata.end(); ++it)
            {
                size_t addr = nextPosition - i_address;
                size_t size = it->second->size();
                file << it->first.c_str() << uint32_t(addr);
                memcpy(nextPosition, it->second->data(), size);
                nextPosition += size;        
            }

            //Finally: Commit.
            handleStdLibError(fsync(i_fileDescriptor) == -1, "Unable to write file \"" + i_filename + "\": Cannot fsync.");
        }



        void SNPFile::ReadWriteSNPFileImpl::close()
        {
            commit();
            closeFile();
        }



        bool SNPFile::ReadWriteSNPFileImpl::isReadOnly()
        {
            return false;
        }



        allele_t *SNPFile::ReadWriteSNPFileImpl::accessMatrixData()
        {
            //These casts are kind of ugly.
            //Internally we simply handle an array of bytes, 
            //elsewhere an allele_t might be anything.
            //We should be fine as long as a byte is 8 bits.
            return reinterpret_cast<allele_t*>(i_address + headerSize());
        }



        void SNPFile::ReadWriteSNPFileImpl::resizeMatrix(int32_t noRows, int32_t noCols)
        {
            checkClosed(i_isClosed, i_filename);

            i_noRows = noRows;
            i_noCols = noCols;
            size_t newSize = headerSize() + matrixSize() + calcMetadataSize();
            resizeFile(newSize);
        }



        MetadataBuffer& SNPFile::ReadWriteSNPFileImpl::getMetadataBuffer(const std::string &key)
        {
            checkClosed(i_isClosed, i_filename);
            metadata_map_t::iterator it = i_metadata.find(key);
            if(it == i_metadata.end())
            {
                MetadataBuffer *buf = new MetadataBuffer(0);
                i_metadata[key] = buf;
                return *buf;
            }
            else
            {
                return *(it->second);
            }
        }



        void SNPFile::ReadWriteSNPFileImpl::removeMetadata(const std::string &key)
        {
            checkClosed(i_isClosed, i_filename);

            metadata_map_t::iterator it = i_metadata.find(key);
            handleMetadataMissingError(it == i_metadata.end(), "Trying to remove nonexisting metadata", key);
            delete it->second;
            i_metadata.erase(it);
        }




        //******************** Definition of ReadOnlySNPFileImpl members ********************

        SNPFile::ReadOnlySNPFileImpl::ReadOnlySNPFileImpl(const std::string &filename)
            : SNPFileImpl(filename)
        {
            init(open(filename.c_str(), O_RDONLY), false);
        }



        SNPFile::ReadOnlySNPFileImpl::~ReadOnlySNPFileImpl()
        {
            preDestructor();
        }



        int SNPFile::ReadOnlySNPFileImpl::mmapProtValue() const
        {
            return PROT_READ;
        }



        void SNPFile::ReadOnlySNPFileImpl::commit()
        {
            handleReadOnlyError("commit", i_filename);
        }



        void SNPFile::ReadOnlySNPFileImpl::close()
        {
            checkClosed(i_isClosed, i_filename);
            closeFile();
        }



        bool SNPFile::ReadOnlySNPFileImpl::isReadOnly()
        {
            return true;
        }



        allele_t *SNPFile::ReadOnlySNPFileImpl::accessMatrixData()
        {
            handleReadOnlyError("update matrix", i_filename);
            throw SNPFileError("This line should never execute.");
        }



        void SNPFile::ReadOnlySNPFileImpl::resizeMatrix(int32_t noRows, int32_t noCols)
        {
            handleReadOnlyError("resize matrix", i_filename);
        }



        MetadataBuffer& SNPFile::ReadOnlySNPFileImpl::getMetadataBuffer(const std::string &key)
        {
            handleReadOnlyError("update metadata", i_filename);
            throw SNPFileError("This line should never execute.");
        }



        void SNPFile::ReadOnlySNPFileImpl::removeMetadata(const std::string &key)
        {
            handleReadOnlyError("remove metadata", i_filename);
        }




        //******************** Definition of SNPFile members ********************

        SNPFile::SNPFile(const std::string &filename, bool allowWriting, bool createFile, int mode)
            : pimpl(allowWriting ? static_cast<SNPFileImpl *>(new ReadWriteSNPFileImpl(filename,
                                                                                       createFile,
                                                                                       mode))
                                 : static_cast<SNPFileImpl *>(new ReadOnlySNPFileImpl(filename) ))
        {
        }

        SNPFile::~SNPFile()
        {
            delete pimpl;
        }

        void SNPFile::commit()
        {
            pimpl->commit();
        }

        void SNPFile::close()
        {
            pimpl->close();
        }

        bool SNPFile::isReadOnly()
        {
            return pimpl->isReadOnly();
        }

        allele_t *SNPFile::accessMatrixData()
        {
            return pimpl->accessMatrixData();
        }

        const allele_t *SNPFile::accessMatrixData() const
        {
            return const_cast<const SNPFileImpl *>(pimpl)->accessMatrixData();
        }

        int32_t SNPFile::noCols() const
        {
            return pimpl->noCols();
        }

        int32_t SNPFile::noRows() const
        {
            return pimpl->noRows();
        }

        void SNPFile::resizeMatrix(int32_t noRows, int32_t noCols)
        {
            pimpl->resizeMatrix(noRows, noCols);
        }

        const MetadataBuffer &SNPFile::getMetadataBuffer(const std::string &key) const
        {
            return const_cast<const SNPFileImpl *>(pimpl)->getMetadataBuffer(key);
        }

        MetadataBuffer &SNPFile::getMetadataBuffer(const std::string &key)
        {
            return pimpl->getMetadataBuffer(key);
        }
    
        bool SNPFile::containsMetadata(const std::string &key) const
        {
            return pimpl->containsMetadata(key);
        }

        size_t SNPFile::getMetadataSize(const std::string &key) const
        {
            return pimpl->getMetadataSize(key);
        }

        void SNPFile::removeMetadata(const std::string &key)
        {
            pimpl->removeMetadata(key);
        }

        MetadataKeyIterator SNPFile::metadataKeysBegin() const
        {
            return pimpl->metadataKeysBegin();
        }

        MetadataKeyIterator SNPFile::metadataKeysEnd() const
        {
            return pimpl->metadataKeysEnd();
        }

        boost::signals::connection SNPFile::addMatrixDataListener(signal_t::slot_function_type subscriber) const
        {
            return pimpl->addMatrixDataListener(subscriber);
        }

        void SNPFile::removeMatrixDataListener(boost::signals::connection subscriber) const
        {
            pimpl->removeMatrixDataListener(subscriber);
        }
    }
}
