/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2007, 2008
 *               Bioinformatics Research Center,
 *               University of Aarhus, Denmark
 *               Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#ifndef SNP_FILE_HH_INCLUDED
#define SNP_FILE_HH_INCLUDED

#include "metadata.hh"
#include "common.hh"

#include <boost/signal.hpp>



/** \file
 *
 * \brief Declaration of the SNPFile class, which provides basic
 * access to the SNPFile format.
 */


namespace BiRC {
    namespace SNPFile {

        const std::string TYPE_INFO_KEY = "BiRC::SNPFile::TypeInfo";
        /**
         * \brief Class representing a snp file. Stores a matrix and
         * supports arbitrary metadata.
         */
        class SNPFile : public MetadataAccessor
        {
            class SNPFileImpl;
            class ReadWriteSNPFileImpl;
            class ReadOnlySNPFileImpl;

            SNPFileImpl *pimpl;



            /// Copy constructor. Not implemented. Do not copy snpfiles.
            SNPFile(const SNPFile &snpFile);

            /// Assignment operator. Not implemented. Do not assign snpfiles.
            SNPFile &operator= (SNPFile &rhs);



        public:
            /**
             * \brief Constructs a SNPFile from a file, given a name.
             *
             * \param filename     Name of file to open.
             * \param allowWriting Opens the file in read/write mode
             *                     instead of read-only mode if true.
             *                     Default is false.
             * \param createFile   Whether to create the file if it is
             *                     not found. If both this and the allowWriting
             *                     argument is true the file will be created,
             *                     otherwise an exception is thrown.
             *                     Default is true.
             * \param mode         The permissions, to create the file with if
             *                     both allowWriting and createFile is true.
             *                     This mode argument is passed directly as the
             *                     mode argument to open and follows the same format.
             *                     Default is 00644.
             *        
             * \exception SNPFileNotFoundError If the file was not found and
             *            allowWriting or createFile is false.
             * \exception SNPFileError If the file could not be opened.
             */
            SNPFile(const std::string &filename,
                    bool allowWriting = false,
                    bool createFile = true,
                    int mode = 00644);

            /**
             * \brief Ensures data is written to disk, closes file and
             * frees allocated memory.
             *
             * WARNING: This destructor may not throw exceptions and hence
             * will fail silently if unable to access the filesystem. Use
             * the close method to close the file in a safe way.
             */
            virtual ~SNPFile();



            /**
             * \brief Ensures that all data is written to disk.
             *
             * \exception SNPFileError         If the file could not be
	     *                                 written for other reason.
             * \exception SNPFileReadOnlyError If this file is read-only.
             * \exception SNPFileClosedError   If the file has already
	     *                                 been closed.
             */
            void commit();



            /**
             * \brief Ensures that all data is written to disk and
             * closes file.
             *
             * Once a SNPFile has been closed it cannot be opened
             * again and no more methods should be called on it.
             *
             * This has exactly same semantics as the destructor,
             * except this methods will throw appropriate exceptions,
             * whereas the destructor will choke those exceptions
             * resulting in silent data loss.
             *
             * \exception SNPFileError         If the file could not be
	     *                                 saved and closed for other
	     *                                 reason
             * \exception SNPFileClosedError   If the file has already
	     *                                 been closed.
             */
            void close();



            /**
             * \brief  Test if this file is read-only
             *
             * \returns true iff this file is read-only
             */
            bool isReadOnly();


	    // FIXME: this shouldn't be available here!  Breaks protection
            /**
             * \brief   Access this files matrix data.
             *          Use this method for both reading and writing data.
             *
             * \returns Pointer to memory where the matrix is mmapped.
             *
             * \exception SNPFileReadOnlyError If this file is read-only.
             */
            allele_t *accessMatrixData();

	    // FIXME: this shouldn't be available here!  Breaks protection
            /**
             * \brief   Access this files matrix data.
             *          Use this method for reading data.
             *
             * \returns Pointer to memory where the matrix is mmapped.
             */
            const allele_t *accessMatrixData() const;

	    // FIXME: this should really only be availble through the matrix
            /**
             * \brief Get width of stored matrix.
             *
             * \returns Number of columns in stored matrix.
             */
            int32_t noCols() const;

	    // FIXME: this should really only be availble through the matrix
            /**
             * \brief Get height of stored matrix.
             *
             * \returns Number of rows in sotred matrix.
             */
            int32_t noRows() const;

	    // FIXME: this should really only be availble through the matrix
            /**
             * \brief        Sets width and height of matrix.
             *               All data currently in the matrix
             *               will (generally) be lost.
             *
             * \param noCols  New width of matrix.
             * \param noRows New height of matrix.
             *
             * \exception SNPFileError         If the file could not be resized for other reason.
             * \exception SNPFileClosedError   If the file has been closed.
             * \exception SNPFileReadOnlyError If this file is read-only.
             */
            void resizeMatrix(int32_t noRows, int32_t noCols);



            /**
             * \brief Get a MetadataBuffer associated with a given key.
             *
             * \param key   Which buffer to get.
             *
             * \returns A metadatabuffer associated to given key.
             *
             * \exception MetadataMissingError If no metadata is associated
             * with given key.
             * \exception SNPFileClosedError   If the file has been closed.
             */
            virtual const MetadataBuffer&
            getMetadataBuffer(const std::string &key) const;

            /**
             * \brief Get a MetadataBuffer associated with a given key.
             *
             * If no buffer is associated with the given key an empty
             * one is created.
             *
             * \param key   Which buffer to get.
             *
             * \returns A metadatabuffer associated to given key.
             *
             * \exception SNPFileClosedError   If the file has been closed.
             * \exception SNPFileReadOnlyError If this file is read-only.
             */
            virtual MetadataBuffer& getMetadataBuffer(const std::string &key);
    
            /**
             * \brief       Checks if metadata exists.
             *
             * \param key   What metadata to test.
             *
             * \returns     true iff key has a valid mapping.
             *
             * \exception SNPFileClosedError   If the file has been closed.
             */
            virtual bool containsMetadata(const std::string &key) const;

            /**
             * \brief       Gets the size, in bytes, of a buffer of metadata.
             *
             * \param key   Which metadata to access.
             *
             * \returns     Size of metadata in bytes.
             *
             * \exception MetadataMissingError If no metadata is associated
             * with given key.
             * \exception SNPFileClosedError   If the file has been closed.
             */
            virtual size_t getMetadataSize(const std::string &key) const;

            /**
             * \brief        Removes metadata from file.
             *
             * \param key    What data to remove.
             *
             * \exception MetadataMissingError If no metadata is associated
             * with given key.
             * \exception SNPFileClosedError   If the file has been closed.
             * \exception SNPFileReadOnlyError If this file is read-only.
             */
            virtual void removeMetadata(const std::string &key);

            /// Returns iterator pointing to beginning of metadata keys.
            virtual MetadataKeyIterator metadataKeysBegin() const;

            /// Returns iterator pointing past end of metadata keys.
            virtual MetadataKeyIterator metadataKeysEnd() const;



	    // FIXME: should this be part of the public interface?
            ///Boost signal to signal changes to matrix pointer
            typedef boost::signal0<void> signal_t;

	    // FIXME: should this be part of the public interface?
            /**
             * \brief Adds a listener to be notified when matrix pointer changes.
             *
             * The SNPFile may, for whatever reason, have to remap the
             * matrix, which may cause the pointer returned by
             * accesMatrixData to change. Use this method to add a
             * subscriber and receive notifications when this happens.
             * It is the subscribers responsibility to unsubscribe
             * when it is garbage collected. This can be done by
             * explicitly calling removeMatrixDataListener, by calling
             * disconnect on the returned connection or by deriving
             * from boost::signals::trackable.  See the boost::signals
             * documentation for details.
             *
             * \param subscriber Functor to be called when pointer changes.
             *
             * \returns          Object representing connection.
             */
            boost::signals::connection addMatrixDataListener(signal_t::slot_function_type subscriber) const;



	    // FIXME: should this be part of the public interface?
            /**
             * \brief Removes a matrix data subscriber.
             *
             * See documentation of addMatrixDataListener for details.
             *
             * \param subscriber Connection to subscriber to remove.
             */
            void removeMatrixDataListener(boost::signals::connection subscriber) const;
        };



        /**
         * \brief Exception used when problems occur while accessing a SNPFile
         */
        class SNPFileError : public std::runtime_error
        {
        public:
            /**
             * \brief Creates a new SNPFileError with a given description
             *
             * \param arg Explanation of why error was reported
             */
            SNPFileError(const std::string &arg)
                : std::runtime_error(arg)
            {}

            /**
             * \brief Creates a new SNPFileError with a given description
             *
             * \param what Description of what went wrong
             * \param why  Description of why error happened
             */
            SNPFileError(const std::string &what, const std::string &why)
                : std::runtime_error(what + ": " + why)
            {}
        };



        /**
         * \brief Exception used when trying to access a file that
         * have been closed
         */
        class SNPFileClosedError : public SNPFileError
        {
        public:
            /**
             * \brief Creates a new SNPFileClosedError with a given description
             *
             * \param arg Explanation of why error was reported
             */
            SNPFileClosedError(const std::string &arg)
                : SNPFileError(arg)
            {}


            /**
             * \brief Creates a new SNPFileClosedError with a given description
             *
             * \param what Description of what went wrong
             * \param why  Description of why error happened
             */
            SNPFileClosedError(const std::string &what, const std::string &why)
                : SNPFileError(what, why)
            {}
        };



        /**
         * \brief Exception used when trying to write data to a read-only file
         */
        class SNPFileReadOnlyError : public SNPFileError
        {
        public:
            /**
             * \brief Creates a new SNPFileReadOnlyError with a given description
             *
             * \param arg Explanation of why error was reported
             */
            SNPFileReadOnlyError(const std::string &arg)
                : SNPFileError(arg)
            {}

            /**
             * \brief Creates a new SNPFileReadOnlyError with a given
             * description
             *
             * \param what Description of what went wrong
             * \param why  Description of why error happened
             */
            SNPFileReadOnlyError(const std::string &what, const std::string &why)
                : SNPFileError(what, why)
            {}
        };



        /**
         * \brief Exception used when trying to open a non-existing file.
         */
        class SNPFileNotFoundError : public SNPFileError
        {
        public:
            /**
             * \brief Creates a new SNPFileNotFoundError with a given
             * description
             *
             * \param arg Explanation of why error was reported
             */
            SNPFileNotFoundError(const std::string &arg)
                : SNPFileError(arg)
            {}

            /**
             * \brief Creates a new SNPFileNotFoundError with a given
             * description
             *
             * \param what Description of what went wrong
             * \param why  Description of why error happened
             */
            SNPFileNotFoundError(const std::string &what, const std::string &why)
                : SNPFileError(what, why)
            {}
        };
    }
}



/**
 * \mainpage SNPFile documentation
 *
 *
 *
 * \section about_sec About SNPFile
 * 
 * SNPFile is a software library and file format used for storing SNP
 * data for large scale association mapping or population genetics
 * studies.
 * 
 * SNPFile is released under the <a
 * href="http://www.gnu.org/copyleft/gpl.html">GNU General Public
 * License (GPL)</a>.  There are no restrictions on the use of
 * SNPFile, for commercial or academic use, but the use of the source
 * code, in part or in whole, is restricted according to the GPL.
 *
 *
 *
 * \section overview_sec Overview
 *
 * At the heart of SNPFile is the BiRC::SNPFile::SNPFile
 * class. Instances of this class represents a SNPFile file. A such
 * file contain a matrix and various metadata.  To access the data in
 * the matrix see the MatrixData class and the derivatives thereof.
 *
 * Metadata are kept in buffers of bytes in the file. Low level access
 * to these buffers are defined on SNPFile. The recommended way to
 * access metadata however is through the functions
 * BiRC::SNPFile::fetchMetadata, BiRC::SNPFile::getMetadata and
 * BiRC::SNPFile::setMetadata defined in metadata_access.hh. These
 * methods provide automatic serialization.
 */



#endif
