/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus, Denmark
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */


#include "snpfile.hh"
#include "matrix.hh"
#include "file_matrix.hh"
#include "metadata_access.hh"
using namespace BiRC::SNPFile;

#include <boost/program_options.hpp>
#include <boost/lambda/lambda.hpp>
using namespace boost::lambda;

#include <cstdlib>
#include <cerrno>
#include <iostream>
#include <fstream>


namespace options {
    //FIXME: currently not in use
    //bool verbose = false;
    std::string snpFileName;
    std::string outputFilePrefix;
    int startIdx = -1;
    int endIdx = -1;
}

inline const char trEnc(int i, const std::string &enc)
{
    assert(0 <= i);
    assert(i < 2);
    switch (enc[i]) {
    case 'A': case 'a':
	return '1';
    case 'C': case 'c':
	return '2';
    case 'G': case 'g':
	return '3';
    case 'T': case 't':
	return '4';
    default:
	return enc[i];
    }
}

inline std::string trGeno(int i, const std::string &enc)
{
    std::string encoding;
    switch (i) {
    case 0:
	encoding += trEnc(0,enc);
	encoding += ' ';
	encoding += trEnc(0,enc);
	break;
    case 1:
	encoding += trEnc(1,enc);
	encoding += ' ';
	encoding += trEnc(1,enc);
	break;
    case 2:
	encoding += trEnc(0,enc);
	encoding += ' ';
	encoding += trEnc(1,enc);
	break;
    default:
	assert(i < 0);
	encoding = "0 0";
    }
    return encoding;
}

inline std::string trHaplo(int i, int j, const std::string &enc)
{
    std::string encoding;
    encoding += trEnc(i,enc);
    encoding += ' ';
    encoding += trEnc(j,enc);
    return encoding;
}



namespace po = boost::program_options;
static void printUsage(const char *progname,
		       std::ostream &os,
		       po::options_description &options)
{
	os << "Usage: " << progname
	   << " [options] snpfile output-prefix"
	   << std::endl << std::endl
	   << "To the output-prefix will be appended .gen for the genotype"
	   << " (linkage) file,"
	   << std::endl
	   << "and .pos for the positions file."
	   << std::endl << std::endl
	   << "When specifying start and end index, the start index is included"
	   << std::endl
	   << "in the interval while the end index is not, and the indexing"
	   << std::endl
	   << "is from zero."
	   << std::endl << std::endl
	   << options << std::endl;
}



int
main(int argc, char *argv[])
{
    po::options_description visible("Allowed options");
    po::options_description hidden("Allowed options");
    po::options_description all("Allowed options");
    po::variables_map vm;
    visible.add_options()
        ("help,h", "Produce help message.")
	//FIXME: currently not in use
	//        ("verbose,v","Turns on verbose output.")

        ("start",
         po::value<int>(&options::startIdx),
         "Start column index (if different from first column). "
	 "Indexing starts at zero.")
        ("end",
         po::value<int>(&options::endIdx),
         "End column index (if different from last column). "
	 "Indexing starts at zero.")
	;
    hidden.add_options()
        ("snpfile",
         po::value<std::string>(&options::snpFileName),
         "The SNPFile to convert.")
        ("output",
         po::value<std::string>(&options::outputFilePrefix),
         "Output file prefix.")
	;
    all.add(visible).add(hidden);

    po::positional_options_description p;
    p.add("snpfile", 1);
    p.add("output", 1);

    try {
        po::store(po::command_line_parser(argc, argv)
                  .options(all).positional(p).run(), 
                  vm, false);
        po::notify(vm);
    } catch (std::exception &e) {
        std::cerr << "Error parsing options: " << e.what()
                  << std::endl;
        printUsage(argv[0], std::cerr, visible);
        return EXIT_FAILURE;
    }

    if (vm.count("help")) {
        printUsage(argv[0], std::cout, visible);
        return EXIT_SUCCESS;
    }
    //FIXME: currently not in use
    //options::verbose = vm.count("verbose") > 0;

    // checking options
    if (options::snpFileName.empty())
	{
  	    std::cerr << "Error: No SNP file specified."
                  << std::endl << std::endl;
	    printUsage(argv[0], std::cerr, visible);
	    return EXIT_FAILURE;
	}
    if (options::outputFilePrefix.empty())
	{
  	    std::cerr << "Error: No output prefix specified."
                  << std::endl << std::endl;
	    printUsage(argv[0], std::cerr, visible);
	    return EXIT_FAILURE;
	}

    try
    {
        SNPFile input(options::snpFileName);
	ImmutableMatrix genotypes(newReadOnlyFileMatrix(input));

        std::vector<int> positions;
        std::vector<std::string> markerNames;
        std::vector<std::string> alleleEncoding;
        std::vector<bool> diseaseStatus;
        getMetadata(input, "positions", positions);
        getMetadata(input, "phenotype", diseaseStatus);

        if(input.containsMetadata("markerNames"))
        {
            getMetadata(input, "markerNames", markerNames);
        }
        else
        {
            std::cerr << "Warning: SNPFile does not contain marker names. "
                      << "Names will be generated instead."
                      << std::endl;

            using boost::lexical_cast;
            for (size_t i = 0; i < positions.size(); ++i)
                markerNames.push_back("marker"+lexical_cast<std::string>(i));

        }

        if(input.containsMetadata("alleleEncoding"))
        {
            getMetadata(input, "alleleEncoding", alleleEncoding);
        }
        else
        {
            std::cerr << "Warning: SNPFile does not contain allele encodings. "
                      << "A dummy encoding will be generated."
                      << std::endl;

            for (size_t i = 0; i < positions.size(); ++i)
		alleleEncoding.push_back("12");

        }

        int startIdx = 0, endIdx = genotypes.noCols();

        if (vm.count("start")) startIdx = options::startIdx;
        if (vm.count("end"))   endIdx   = options::endIdx;

        if (startIdx < 0)
        {
            std::cerr << "Start index is before the first column!" 
                      << std::endl;
            return EXIT_FAILURE;
        }
        if (endIdx > genotypes.noCols())
        {
            std::cerr << "End index extends beyond the last column!" 
                      << std::endl;
            return EXIT_FAILURE;
        }
    

        std::ofstream genOut((options::outputFilePrefix+".gen").c_str());
        if (!genOut)
        {
            std::cerr << "Error opening "
                      << options::outputFilePrefix << ".gen."
                      << std::endl;
            return EXIT_FAILURE;
        }
        std::ofstream posOut((options::outputFilePrefix+".pos").c_str());
        if (!posOut)
        {
            std::cerr << "Error opening "
                      << options::outputFilePrefix << ".pos."
                      << std::endl;
            return EXIT_FAILURE;
        }

        // writing positions file
        for (int i = startIdx; i < endIdx; ++i)
        {
            posOut << markerNames[i] << '\t' << positions[i] << std::endl;
        }

	
        // writing genotype file
	if (genotypes.isPhased())
	    {
		for (int i = 0; i < genotypes.noRows()/2; ++i)
		    {
			// pedigreeID individuaID fatherID motherID gender status
			genOut << (1+i) << ' ' << (1+i)
			       << " 0 0 1 " << (1+diseaseStatus[i]);
			for (int j = startIdx; j < endIdx; ++j)
			    genOut << ' ' << trHaplo(genotypes(2*i,j), 
						     genotypes(2*i+1,j), 
						     alleleEncoding[j]);
			genOut << std::endl;
		    }
	    }
	else
	    {
		for (int i = 0; i < genotypes.noRows(); ++i)
		    {
			// pedigreeID individuaID fatherID motherID gender status
			genOut << (1+i) << ' ' << (1+i)
			       << " 0 0 1 " << (1+diseaseStatus[i]);
			for (int j = startIdx; j < endIdx; ++j)
			    genOut << ' ' << trGeno(genotypes(i,j), alleleEncoding[j]);
			genOut << std::endl;
		    }
	    }

    }
    catch(SNPFileNotFoundError &)
    {
        std::cerr << "Cannot open \"" << options::snpFileName << "\". "
                  << "File not found.\n";
    }
    catch(std::exception &e)
    {
        std::cerr << "Error: " << e.what() << std::endl;

        if(unlink((options::outputFilePrefix + ".gen").c_str()) == -1 && errno != ENOENT)
            std::cerr << "Unable to remove file "
                      << options::outputFilePrefix << ".gen"
                      << std::endl;

        if(unlink((options::outputFilePrefix + ".pos").c_str()) == -1 && errno != ENOENT)
            std::cerr << "Unable to remove file "
                      << options::outputFilePrefix << ".pos"
                      << std::endl;
    }


    return EXIT_SUCCESS;
}

