/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus, Denmark
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */


#include "snpfile.hh"
#include "matrix.hh"
#include "file_matrix.hh"
#include "metadata_access.hh"
using namespace BiRC::SNPFile;

#include <boost/program_options.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/lexical_cast.hpp>
using namespace boost::lambda;

#include <cstdlib>
#include <cerrno>
#include <iostream>
#include <fstream>


namespace options {
    //FIXME: currently not in use
    //bool verbose = false;
    std::string snpFileName;
    std::string outputFileName;
    int startIdx = -1;
    int endIdx = -1;
}


inline std::string tr(int i)
{
    if (i < 0)  return std::string("?");
    else        return boost::lexical_cast<std::string>(i);
}


namespace po = boost::program_options;
static void printUsage(const char *progname,
                       std::ostream &os,
                       po::options_description &options)
{
	os << "Usage: " << progname
	   << " [options] snpfile [output-file]"
	   << std::endl << std::endl
	   << std::endl << std::endl
	   << "When specifying start and end index, the start index is included"
	   << std::endl
	   << "in the interval while the end index is not, and the indexing"
	   << std::endl
	   << "is from zero."
	   << std::endl << std::endl
	   << options << std::endl;
}

static void printOutput(const ImmutableMatrix &genotypes, 
                        int startIdx, int endIdx,
                        std::ostream &output)
{
    for (int i = 0; i < genotypes.noRows(); ++i)
	{
	    for (int j = startIdx; j < endIdx; ++j)
		{
		    output << tr(genotypes(i,j)) << ' ';
		}
	    output << std::endl;
	}
}


int
main(int argc, char *argv[])
{
    po::options_description visible("Allowed options");
    po::options_description hidden("Allowed options");
    po::options_description all("Allowed options");
    po::variables_map vm;
    visible.add_options()
        ("help,h", "Produce help message.")
        //FIXME: currently not in use
        //        ("verbose,v","Turns on verbose output.")

        ("start",
         po::value<int>(&options::startIdx),
         "Start column index (if different from first column). "
         "Indexing starts at zero.")
        ("end",
         po::value<int>(&options::endIdx),
         "End column index (if different from last column). "
         "Indexing starts at zero.")
        ;
    hidden.add_options()
        ("snpfile",
         po::value<std::string>(&options::snpFileName),
         "The SNPFile to convert.")
        ("output",
         po::value<std::string>(&options::outputFileName),
         "Output file name.")
        ;
    all.add(visible).add(hidden);

    po::positional_options_description p;
    p.add("snpfile", 1);
    p.add("output", 1);

    try {
        po::store(po::command_line_parser(argc, argv)
                  .options(all).positional(p).run(), 
                  vm, false);
        po::notify(vm);
    } catch (std::exception &e) {
        std::cerr << "Error parsing options: " << e.what()
                  << std::endl;
        printUsage(argv[0], std::cerr, visible);
        return EXIT_FAILURE;
    }

    if (vm.count("help")) {
        printUsage(argv[0], std::cout, visible);
        return EXIT_SUCCESS;
    }
    //FIXME: currently not in use
    //options::verbose = vm.count("verbose") > 0;

    // checking options
    if (options::snpFileName.empty())
	{
  	    std::cerr << "Error: No SNP file specified."
                  << std::endl << std::endl;
	    printUsage(argv[0], std::cerr, visible);
	    return EXIT_FAILURE;
	}

    try
    {
        SNPFile input(options::snpFileName);
	ImmutableMatrix genotypes(newReadOnlyFileMatrix(input));

        int startIdx = 0, endIdx = genotypes.noCols();

        if (vm.count("start")) startIdx = options::startIdx;
        if (vm.count("end"))   endIdx   = options::endIdx;

        if (startIdx < 0)
        {
            std::cerr << "Start index is before the first column!" 
                      << std::endl;
            return EXIT_FAILURE;
        }
        if (endIdx > genotypes.noCols())
        {
            std::cerr << "End index extends beyond the last column!" 
                      << std::endl;
            return EXIT_FAILURE;
        }
    

        if (options::outputFileName.empty())
        {
            printOutput(genotypes, startIdx, endIdx, std::cout);
        }
        else
        {
            std::ofstream output((options::outputFileName).c_str());
            if (!output)
            {
                std::cerr << "Error opening "
                          << options::outputFileName
                          << std::endl;
                return EXIT_FAILURE;
            }
            printOutput(genotypes, startIdx, endIdx, output);
        }
    }
    catch(SNPFileNotFoundError &)
    {
        std::cerr << "Cannot open \"" << options::snpFileName << "\". "
                  << "File not found.\n";
    }
    catch(std::exception &e)
    {
        std::cerr << "Error: " << e.what() << std::endl;

        if(unlink(options::outputFileName.c_str()) == -1 && errno != ENOENT)
            std::cerr << "Unable to remove file "
                      << options::outputFileName
                      << std::endl;
    }


    return EXIT_SUCCESS;
}

