/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus, Denmark
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include <iostream>
#include <fstream>
#include <string>

#include "snpfile.hh"
#include "metadata_access.hh"
using namespace BiRC::SNPFile;

#include <boost/program_options.hpp>



namespace options
{
    std::string snpFilename;
    std::string metadataFilename;
    std::string metadataType;
    std::string metadataKey;
}



template <typename T>
std::ostream &operator<< (std::ostream &out, const std::vector<T> &v)
{
    typename std::vector<T>::const_iterator it;
    for(it = v.begin(); it != v.end(); ++it)
        out << *it << std::endl;

    return out;
}



namespace po = boost::program_options;
static void printUsage(const char *progname,
                       std::ostream &os,
                       po::options_description &options)
{
	os << "Usage: " << progname
	   << " [options] metadata-key snpfile"
	   << std::endl
	   << options << std::endl
       << "Recognized types are:" << std::endl
       << "  \"" << typeName< bool >() << "\"" << std::endl
       << "  \"" << typeName< int >() << "\"" << std::endl
       << "  \"" << typeName< double >() << "\"" << std::endl
       << "  \"" << typeName< std::string >() << "\"" << std::endl
       << "  \"" << typeName< std::vector< bool > >() << "\"" << std::endl
       << "  \"" << typeName< std::vector< int > >() << "\"" << std::endl
       << "  \"" << typeName< std::vector< double > >() << "\"" << std::endl
       << "  \"" << typeName< std::vector< std::string > >() << "\"" << std::endl
       << std::endl
       << "Please note white-space is significant in type names." << std::endl;
}



int main(int argc, char **argv)
{
    po::options_description visible("Allowed options");
    po::options_description hidden("Allowed options");
    po::options_description all("Allowed options");
    po::variables_map vm;

    visible.add_options()
        ("help,h", "Produce this help message.")
        ("output-file,o",
         po::value<std::string>(&options::metadataFilename),
         "File to store metadata in.")
        ("metadata-type,t",
         po::value<std::string>(&options::metadataType),
         "Type of metadata. If the type is stored in the snpfile that type is used. "
         "If it is not stored this argument is required.")
        ;

    hidden.add_options()
        ("metadata-key",
         po::value<std::string>(&options::metadataKey),
         "Key to read metadata from.")
        ("snpfile",
         po::value<std::string>(&options::snpFilename),
         "The SNPFile to read from.")
        ;

    all.add(visible).add(hidden);

    po::positional_options_description p;
    p.add("metadata-key", 1);
    p.add("snpfile", 1);

    try
    {
        po::store(po::command_line_parser(argc, argv)
                  .options(all).positional(p).run(), 
                  vm, false);
        po::notify(vm);
    }
    catch(std::exception &e)
    {
        std::cerr << "Error parsing options: " << e.what()
                  << std::endl;
        printUsage(argv[0], std::cerr, visible);
        return EXIT_FAILURE;
    }

    if(vm.count("help"))
    {
        printUsage(argv[0], std::cout, visible);
        return EXIT_SUCCESS;
    }

    if(!vm.count("metadata-key") || !vm.count("snpfile"))
    {
        std::cerr << "Missing required option." << std::endl;
        printUsage(argv[0], std::cerr, visible);
        return EXIT_FAILURE;
    }



    std::ostream *out = 0;
    std::ofstream *outFile = 0;
    bool deleteOutStream = false;

    try
    {
        SNPFile snpFile(options::snpFilename);



        std::string typeFromFile = const_cast<const SNPFile &>(snpFile)
            .getMetadataBuffer(options::metadataKey).getType();

        if(typeFromFile != UNKNOWN_TYPE_NAME && typeFromFile != "")
        {
            if(options::metadataType != "" && options::metadataType != typeFromFile)
                std::cerr << "Warning: Ignoring type argument (=" << options::metadataType
                          << "). Using value stored in SNPFile (=" << typeFromFile << ")." << std::endl;
            options::metadataType = typeFromFile;
        }



        if(options::metadataType == "")
        {
            std::cerr << "Unable to find type of metadata. Please provide it." << std::endl;
            return EXIT_FAILURE;
        }



        if(vm.count("output-file"))
        {
            out = outFile = new std::ofstream(options::metadataFilename.c_str(), std::ios::out);
            deleteOutStream = true;
        }
        else
        {
            out = &std::cout;
            deleteOutStream = false;
        }



        if(isTypeName< bool >(options::metadataType))
            (*out) << fetchMetadata< bool >(snpFile, options::metadataKey) << std::endl;
        else if(isTypeName< int >(options::metadataType))
            (*out) << fetchMetadata< int >(snpFile, options::metadataKey) << std::endl;
        else if(isTypeName< double >(options::metadataType))
            (*out) << fetchMetadata< double >(snpFile, options::metadataKey) << std::endl;
        else if(isTypeName< std::string >(options::metadataType))
            (*out) << fetchMetadata< std::string >(snpFile, options::metadataKey) << std::endl;
        else if(isTypeName< std::vector< bool > >(options::metadataType))
            (*out) << fetchMetadata< std::vector< bool > >(snpFile, options::metadataKey);
        else if(isTypeName< std::vector< int > >(options::metadataType))
            (*out) << fetchMetadata< std::vector< int > >(snpFile, options::metadataKey);
        else if(isTypeName< std::vector< double > >(options::metadataType))
            (*out) << fetchMetadata< std::vector< double > >(snpFile, options::metadataKey);
        else if(isTypeName< std::vector< std::string > >(options::metadataType))
            (*out) << fetchMetadata< std::vector< std::string > >(snpFile, options::metadataKey);
        else
        {
            std::cerr << "Unknown type " << options::metadataType << "." << std::endl;
            printUsage(argv[0], std::cerr, visible);
            return EXIT_FAILURE;
        }



        snpFile.close();
    }
    catch(std::exception &ex)
    {
        std::cerr << "Error occurred: " << ex.what() << std::endl;
        return EXIT_FAILURE;
    }

    if(outFile != 0)
        outFile->close();

    if(deleteOutStream)
        delete out;

    return EXIT_SUCCESS;
}
