/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus, Denmark
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include <iostream>
#include <string>

#include "snpfile.hh"
#include "metadata.hh"
#include "metadata_access.hh"
#include "matrix.hh"
#include "file_matrix.hh"



using namespace BiRC::SNPFile;



namespace
{
    void printMatrixInformation(const SNPFile &file)
    {
	ImmutableMatrix matrix(newReadOnlyFileMatrix(file));
    
        std::cout << "  Matrix:" << std::endl
                  << "    Rows: " << matrix.noRows() << std::endl
                  << "    Columns: " << matrix.noCols() << std::endl
                  << "    Phased?: " << (matrix.isPhased() ? "phased" : "unphased") << std::endl;
    }



    template<typename T>
    void tryToPrintVectorInfo(const SNPFile &file, const std::string &key, const std::string &type)
    {
        if(isTypeName< std::vector<T> >(type))
        {
            std::vector<T> v;
            getMetadata(file, key, v);
            std::cout << "      Vector length: " << v.size() << std::endl;
        }
    }



    void printMetadataInformation(const SNPFile &file, const std::string &key)
    {
        const MetadataBuffer &buffer = file.getMetadataBuffer(key);

        std::cout << "    " << key << std::endl;

        const std::string &typeName = buffer.getType();

        if(typeName != "")
        {
            std::cout << "      Type: " << typeName << std::endl;
            const std::string vectorStart = "std::vector< ";
            if(typeName.substr(0, vectorStart.length()) == vectorStart)
            {
                tryToPrintVectorInfo<bool>         (file, key, typeName);
                tryToPrintVectorInfo<unsigned int> (file, key, typeName);
                tryToPrintVectorInfo<int>          (file, key, typeName);
                tryToPrintVectorInfo<char>         (file, key, typeName);
                tryToPrintVectorInfo<double>       (file, key, typeName);
                tryToPrintVectorInfo<float>        (file, key, typeName);
                tryToPrintVectorInfo<std::string>  (file, key, typeName);
            }
        }
    }
}



int main(int argc, char **argv)
{
    if(argc <= 1)
    {
        std::cerr << "Error: No input file specified." << std::endl;
        std::cerr << std::endl;
        std::cerr << "Usage: " << argv[0] << " inputfiles..." << std::endl;
        std::cerr << std::endl;
        std::cerr << "Prints information on metadata in inputfile." << std::endl;
        return EXIT_FAILURE;
    }

    for(int i = 1; i < argc; i++)
    {
        const std::string filename(argv[i]);
        std::cout << "File " << filename << ":" << std::endl;

        try
        {
            SNPFile file(filename);

            printMatrixInformation(file);

            std::cout << "  Metadata:" << std::endl;
        
            MetadataKeyIterator it = file.metadataKeysBegin();

            for(; it != file.metadataKeysEnd(); ++it)
                printMetadataInformation(file, *it);

            file.close();
        }
        catch(SNPFileError &er)
        {
            std::cerr << "Error: " << er.what() << std::endl;
        }
        catch(MetadataError &er)
        {
            std::cerr << "Error: " << er.what() << std::endl;
        }

        std::cout << std::endl;
    }

    return EXIT_SUCCESS;
}
