/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus, Denmark
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <algorithm>

#include <sys/stat.h>

#include "snpfile.hh"
#include "metadata_access.hh"
#include "type_trait.hh"
using namespace BiRC::SNPFile;

#include <boost/program_options.hpp>
#include <boost/lexical_cast.hpp>



namespace
{
    enum Type {T_BOOL, T_INT, T_DOUBLE, T_STRING,
               T_VECTOR_BOOL, T_VECTOR_INT, T_VECTOR_DOUBLE, T_VECTOR_STRING,
               T_UNKNOWN};


    const int TYPE_COUNT = int(T_UNKNOWN);

    std::string typeNames[TYPE_COUNT] = {typeName<bool>(),
                                         typeName<int>(),
                                         typeName<double>(),
                                         typeName<std::string>(),
                                         typeName< std::vector<bool> >(),
                                         typeName< std::vector<int> >(),
                                         typeName< std::vector<double> >(),
                                         typeName< std::vector<std::string> >()};



    const std::string &typeToName(Type t)
    {
        return typeNames[t];
    }



    const Type nameToType(const std::string &n)
    {
        return Type(std::find(typeNames, typeNames + TYPE_COUNT, n) - typeNames);
    }



    std::ostream &operator<< (std::ostream &out, Type t)
    {
        return out << typeToName(t);
    }



    std::istream &operator>> (std::istream &in, Type &t)
    {
        std::string tmp;
        in >> tmp;
        if(tmp == "std::vector<")
        {
            for(int i = 0; in && i < 2; i++)
            {
                std::string tmp2;
                in >> tmp2;
                tmp += " " + tmp2;
            }
        }

        t = nameToType(tmp);

        return in;
    }



    //Warning: Breaks easily!
    Type vectorOf(Type t)
    {
        return Type(t + 4);
    }



    //Warning: Breaks easily!
    Type basetype(Type v)
    {
        return Type(v - 4);
    }








    inline bool isWhiteSpace(char c)
    {
        return c == ' ' || c == '\t';
    }



    inline bool contains(const std::string &s, char c)
    {
        return s.find(c, 0) != std::string::npos;
    }



    void trim(std::string &s)
    {
        while(!s.empty() && isWhiteSpace(s[0]))
            s.erase(0, 1);

        while(!s.empty() && isWhiteSpace(s[s.length() - 1]))
            s.erase(s.length() - 1, 1);
    }









    void parseTypeToken(std::string token,
                        int &column,
                        std::map<int, Type> &result)
    {
        size_t equalIdx = token.find('=', 0);

        try
        {
            if(equalIdx != std::string::npos)
            {
                column = boost::lexical_cast<int>(token.substr(0, equalIdx));
                token.erase(0, equalIdx + 1);
                trim(token);
            }
        }
        catch(boost::bad_lexical_cast &)
        {
            std::cerr << "Unable to parse \"" << token.substr(0, equalIdx) << "\" as column index." << std::endl;
            exit(EXIT_FAILURE);
        }

        if((result[column] = nameToType(token)) == T_UNKNOWN)
        {
            std::cerr << "Unable to parse \"" << token << "\" as type." << std::endl;
            exit(EXIT_FAILURE);
        }
    }



    void parseTypeString(std::string string,
                         std::map<int, Type> &result)
    {
        int column = 0;
        while(!string.empty())
        {
            while(!string.empty() && isWhiteSpace(string[0]))
                string.erase(0, 1);

            size_t idx = 0;
            int brackets = 0;

            for(; idx < string.length() && (brackets > 0 || !isWhiteSpace(string[idx])); idx++)
            {
                if(string[idx] == '<')
                    brackets++;
                else if(string[idx] == '>')
                    brackets--;
            }

            if(idx != 0)
            {
                parseTypeToken(string.substr(0, idx), column, result);
                string.erase(0, idx);
            }

            column++;
        }
    }



    void parseTypeStrings(const std::vector<std::string> &strings,
                          std::map<int, Type> &result)
    {
        std::vector<std::string>::const_iterator it;
        for(it = strings.begin(); it != strings.end(); ++it)
            parseTypeString(*it, result);
    }









    struct MetadataColumn
    {
        Type type;
        std::string key;
        std::vector< std::string > values;



        MetadataColumn()
            : type(T_BOOL)
        {}
    };









    bool getLine(std::istream &in, std::string &result)
    {
        do
        {
            result.clear();

            if(!in)
                return false;

            std::getline(in, result);
            size_t commentPos = result.find('#', 0);
            if(commentPos != std::string::npos)
                result.erase(commentPos);

            trim(result);
        } while(result.empty());

        return true;
    }



    std::vector<std::string> tokenize(std::string &s, std::string separators)
    {
        std::vector< std::string > result;
        size_t start = 0;
        for(size_t i = 0; i < s.length(); i++)
        {
            if(contains(separators, s[i]))
            {
                std::string temp = s.substr(start, i - start);
                trim(temp);
                result.push_back(temp);

                while(i < s.length() && contains(separators, s[i]))
                    i++;

                start = i;
            }
        }

        if(start != s.length())
        {
            std::string temp = s.substr(start);
            trim(temp);
            result.push_back(temp);
        }

        return result;
    }



    void parseData(std::istream &in, const std::string &header, std::vector< MetadataColumn > &data)
    {
        std::string read;
        bool success;
        int lineNu = 0;

        if(header != "")
        {
            read = header;
        }
        else
        {
            lineNu++;
            success = getLine(in, read);
            if(!success)
            {
                std::cerr << "Unexpected end of input." << std::endl;
                exit(EXIT_FAILURE);
            }
        }

        std::vector< std::string > keys = tokenize(read, " \t");
        std::vector< std::string >::const_iterator it;
        for(it = keys.begin(); it != keys.end(); ++it)
        {
            MetadataColumn column;
            column.key = *it;
            data.push_back(column);
        }

        lineNu++;
        success = getLine(in, read);

        while(success)
        {
            std::vector< std::string > vals = tokenize(read, " \t");

            if(vals.size() != keys.size())
            {
                std::cerr << "Error on line " << lineNu << " of input." << std::endl;
                std::cerr << "Number of columns (" << vals.size()
                          << ") does not match number in header (" << keys.size()
                          << ")." << std::endl;
                exit(EXIT_FAILURE);
            }

            for(size_t i = 0; i < data.size(); i++)
                data[i].values.push_back(vals[i]);
            
            lineNu++;
            success = getLine(in, read);
        }
    }








    template<typename T>
    bool isType(const std::string &value)
    {
        try
        {
            boost::lexical_cast<T>(value);
            return true;
        }
        catch(boost::bad_lexical_cast &)
        {
        }

        return false;
    }



    Type getType(const std::string &value)
    {
        if(isType<int>(value))
            return T_INT;
        else if(isType<double>(value))
            return T_DOUBLE;
        else if(value == "true" || value == "false")
            return T_BOOL;
        else
            return T_STRING;
    }



    //WARNING: Breaks easily
    inline Type mostGeneral(Type left, Type right)
    {
        if(left == T_UNKNOWN)
            return right;

        if(right == T_UNKNOWN)
            return left;

        if(left == T_BOOL || right == T_BOOL)
        {
            if(left == T_BOOL && right == T_BOOL)
                return T_BOOL;
            else
                return T_STRING;
        }
        else
        {
            return left > right ? left : right;
        }
    }



    void deduceType(MetadataColumn &column)
    {
        if(column.values.empty())
        {
            column.type = T_VECTOR_INT;
            return;
        }

        Type bestType = getType(column.values[0]);
        std::vector< std::string >::const_iterator it;
        for(it = column.values.begin();
            it != column.values.end();
            ++it)
            bestType = mostGeneral(bestType, getType(*it));

        if(column.values.size() != 1)
            bestType = vectorOf(bestType);

        column.type = bestType;
    }









    template<typename T>
    T convert(const std::string &val)
    {
        return boost::lexical_cast<T>(val);
    }



    template<>
    bool convert(const std::string &val)
    {
        bool isFalse = val == "false" || val == "0" || val == "no";
        return !isFalse;
    }



    template<typename T>
    std::vector< T > convertVector(const std::vector< std::string > &vals)
    {
        std::vector< T > result;
        std::vector< std::string >::const_iterator it;
        for(it = vals.begin(); it != vals.end(); ++it)
            result.push_back(convert< T >(*it));
        return result;
    }



    void writeData(SNPFile &snpFile, const MetadataColumn &column)
    {
        switch(column.type)
        {
            case T_BOOL:
                assert(column.values.size() == 1);
                setMetadata(snpFile, column.key, convert< bool >(column.values[0]));
                break;

            case T_INT:
                assert(column.values.size() == 1);
                setMetadata(snpFile, column.key, convert< int >(column.values[0]));
                break;

            case T_DOUBLE:
                assert(column.values.size() == 1);
                setMetadata(snpFile, column.key, convert< double >(column.values[0]));
                break;

            case T_STRING:
                assert(column.values.size() == 1);
                setMetadata(snpFile, column.key, column.values[0]);
                break;

            case T_VECTOR_BOOL:
                setMetadata(snpFile, column.key, convertVector< bool >(column.values));
                break;

            case T_VECTOR_INT:
                setMetadata(snpFile, column.key, convertVector< int >(column.values));
                break;

            case T_VECTOR_DOUBLE:
                setMetadata(snpFile, column.key, convertVector< double >(column.values));
                break;

            case T_VECTOR_STRING:
                setMetadata(snpFile, column.key, column.values);
                break;

            default:
                std::cerr << "Internal error: Unknown type " << column.type << std::endl;
                assert(!"Internal error.");
        }               
    }









    namespace
    {
        std::vector<std::string> breakLines(std::string x, int lineLength)
        {
            std::vector<std::string> retVal;
            while(int(x.length()) > lineLength)
            {
                int idx = lineLength;
                for(; idx >= 0 && x[idx] != ' '; idx--);
                if(idx == 0)
                    idx = lineLength;
                retVal.push_back(x.substr(0, idx));
                x.erase(0, idx + 1);
            }

            retVal.push_back(x);

            return retVal;
        }



        void printAsLines(int startCol, int endCol, const std::string &x)
        {
            std::string sep = "";
            for(int i = 0; i < startCol; i++)
                sep += " ";
            
            const std::vector<std::string> lines = breakLines(x, endCol - startCol);
            std::vector<std::string>::const_iterator it;
            for(it = lines.begin(); it != lines.end(); ++it)
                std::cout << sep << *it << std::endl;
        }
    }



    namespace options
    {
        std::string snpFilename;
        std::string metadataFilename;
        std::vector<std::string> metadataTypeStrings;
        std::string header;
        std::string values;
    }



    namespace po = boost::program_options;
    static void printUsage(const char *progname,
                           std::ostream &os,
                           po::options_description &options)
    {
        os << "Usage: " << progname
           << " [options] snpfile"
           << std::endl
           << options << std::endl
           << std::endl
           << "If neither -i or -v is given input is read from stdin." << std::endl
           << std::endl
           << "Data is parsed the following way:" << std::endl;

        printAsLines(2, 80,
                     "The data is read line by line. First any # characters are found "
                     "and anything on a line following a # is considered a comment and "
                     "removed. "
                     "If the line is now empty it is ignored. "
                     "The line is now split into tokens separated by any number of spaces or tabs. "
                     "All lines is expected to have the same number of tokens. "
                     "The first non-empty line is taken as the keys to store the metadata under. "
                     "All remaining lines are considered data. "
                     "The types of the data are deduced the following way. "
                     "If C++ is able to interpret all data in a column as int the type is int. "
                     "Otherwise, if C++ is able to interpret all data of a column as double the type is double. "
                     "Otherwise, if all data in a column is either the string true or the "
                     "string false the type is bool. "
                     "Otherwise the type is string. "
                     "If there is exactly one line of data the final types are simply these primitive types. "
                     "Otherwise the data is vectors of these primitive types.");

        os << std::endl
           << "The format for the -t argument is as follows" << std::endl
           << "  -t \"[column=]type...\"" << std::endl
           << "where type is one of:" << std::endl;

        for(int i = 0; i < TYPE_COUNT; i++)
            os << "  " << Type(i) << std::endl;

        os << std::endl;

        printAsLines(2, 80,
                     "if column is given it is the type of that column that is set. "
                     "The columns are enumerated from 0. "
                     "Otherwise if no type is preceeding this it is the type of the first column that is set. "
                     "Otherwise the type is set of the column following the column whose "
                     "type was previously set. "
                     "For example:");
        os << std::endl;
        os << "    -t \"std::vector< int > 2=std::vector< bool > std::vector< std::string >\"" << std::endl;
        os << std::endl;
        printAsLines(2, 80,
                     "would set the type of the first column as a vector of ints, the type of the third "
                     "column as a vector of bool and the type of the fourth column as a vector of strings, "
                     "while the type of the second column and any other columns would be attempted deduced. "
                     "Please note that whitespace and quotes are significant.");
                
    }
}



int main(int argc, char **argv)
{
    po::options_description visible("Allowed options");
    po::options_description hidden("Allowed options");
    po::options_description all("Allowed options");
    po::variables_map vm;

    visible.add_options()
        ("help,h", "Produce this help message.")
        ("force-types,t",
         po::value< std::vector<std::string> >(&options::metadataTypeStrings),
         "Force types of given columns. If types is not given for some column"
         " the type is deduced for that column.")
        ("header,k",
         po::value<std::string>(&options::header),
         "Use this line for header/metadata keys. First line of input is now "
         "interpreted as data")
        ("values,v",
         po::value<std::string>(&options::values),
         "Use this line for data input. (Mutually exclusive with -i)")
        ("input-file,i",
         po::value<std::string>(&options::metadataFilename),
         "File to read data from. (Mututlly exclusive with -v)")
        ;

    hidden.add_options()
        ("snpfile",
         po::value<std::string>(&options::snpFilename),
         "The SNPFile to update.")
        ;

    all.add(visible).add(hidden);

    po::positional_options_description p;
    p.add("snpfile", 1);

    try
    {
        po::store(po::command_line_parser(argc, argv)
                  .options(all).positional(p).run(), 
                  vm, false);
        po::notify(vm);
    }
    catch(std::exception &e)
    {
        std::cerr << "Error parsing options: " << e.what()
                  << std::endl;
        printUsage(argv[0], std::cerr, visible);
        return EXIT_FAILURE;
    }

    if(vm.count("help"))
    {
        printUsage(argv[0], std::cout, visible);
        return EXIT_SUCCESS;
    }

    if(!vm.count("snpfile"))
    {
        std::cerr << "Missing required option." << std::endl;
        printUsage(argv[0], std::cerr, visible);
        return EXIT_FAILURE;
    }

    if(vm.count("values") && vm.count("input-file"))
    {
        std::cerr << "Reading from file and reading for command line are mutually exclusive." << std::endl;
        printUsage(argv[0], std::cerr, visible);
        return EXIT_FAILURE;
    }



    std::istream *in = 0;
    std::ifstream *inFile = 0;
    bool deleteInStream = false;

    try
    {
        std::map<int, Type> givenTypes;
        parseTypeStrings(options::metadataTypeStrings, givenTypes);



        SNPFile snpFile(options::snpFilename, true, false);



        if(vm.count("input-file"))
        {
            struct stat dummyStat;
            if(stat(options::metadataFilename.c_str(), &dummyStat))
            {
                std::cerr << "Unable to read input file "
                          << options::metadataFilename
                          << "." << std::endl;
                printUsage(argv[0], std::cerr, visible);
                return EXIT_FAILURE;
            }

            in = inFile = new std::ifstream(options::metadataFilename.c_str(), std::ios::in);
            deleteInStream = true;
        }
        else if(vm.count("values"))
        {
            std::istringstream *inString = new std::istringstream();
            inString->str(options::values);
            in = inString;
            deleteInStream = true;
        }
        else
        {
            in = &std::cin;
            deleteInStream = false;
        }



        std::vector< MetadataColumn > data;
        parseData(*in, options::header, data);

        for(size_t i = 0; i < data.size(); i++)
        {
            MetadataColumn &column = data[i];
            if(givenTypes.find(i) != givenTypes.end())
                column.type = givenTypes[i];
            else
                deduceType(column);

            writeData(snpFile, column);
        }

        snpFile.close();
    }
    catch(std::exception &ex)
    {
        std::cerr << "Error occurred: " << ex.what() << std::endl;
        return EXIT_FAILURE;
    }

    if(inFile != 0)
        inFile->close();

    if(deleteInStream)
        delete in;

    return EXIT_SUCCESS;
}
