/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus, Denmark
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include <iostream>
#include <string>

#include "snpfile.hh"
#include "matrix.hh"
#include "file_matrix.hh"
#include "metadata_access.hh"

#include <boost/program_options.hpp>


using namespace BiRC::SNPFile;



template<typename E>
void splitVectorMetadata(const SNPFile &inFile, int splitIdx, 
                         const std::string &key,
                         SNPFile &outFile1, SNPFile &outFile2)
{
    if(!inFile.containsMetadata(key))
        return;

    std::vector<E> in;
    getMetadata(inFile, key, in);
    std::vector<E> out1;
    std::vector<E> out2;

    for(int i = 0; i < splitIdx; i++)
        out1.push_back(in[i]);
    for(int i = splitIdx; i < inFile.noCols(); i++)
        out2.push_back(in[i]);

    setMetadata(outFile1, key, out1);
    setMetadata(outFile2, key, out2);
}



void copyMetadata(const SNPFile &inFile,
                  const std::string &key,
                  SNPFile &outFile1, SNPFile &outFile2)
{
    const MetadataBuffer &inBuffer = inFile.getMetadataBuffer(key);

    MetadataBuffer &outBuffer1 = outFile1.getMetadataBuffer(key);
    outBuffer1.reset(inBuffer.data(), inBuffer.size());
    MetadataBuffer &outBuffer2 = outFile2.getMetadataBuffer(key);
    outBuffer2.reset(inBuffer.data(), inBuffer.size());
}



void splitMatrix(const SNPFile &inFile, int splitIdx, 
                 SNPFile &outFile1, SNPFile &outFile2)
{
    ImmutableMatrix inMatrix(newReadOnlyFileMatrix(inFile));
    Matrix outMatrix1(newReadWriteFileMatrix(outFile1));
    Matrix outMatrix2(newReadWriteFileMatrix(outFile2));

    outMatrix1.resize(inMatrix.noRows(), splitIdx);
    outMatrix2.resize(inMatrix.noRows(), inFile.noCols() - splitIdx);


    for(int i = 0; i < splitIdx; i++)
        for(int j = 0; j < inMatrix.noRows(); j++)
            outMatrix1(j, i) = inMatrix(j, i);

    for(int i = 0; i < inMatrix.noCols() - splitIdx; i++)
        for(int j = 0; j < inMatrix.noRows(); j++)
            outMatrix2(j, i) = inMatrix(j, splitIdx + i);
}



namespace options
{
    bool verbose = false;
    std::string filename;
    std::string outFilename1;
    std::string outFilename2;
    int splitIdx = -1;
}


namespace po = boost::program_options;



static void printUsage(const std::string &progname,
                       std::ostream &os,
                       po::options_description &options)
{
    os << "Usage: " << progname << " [options] snpfile n outfile1 outfile2" << std::endl
       << options << std::endl
       << "Splits snpfile into two on column n. Columns 0 through n-1 goes into file"    << std::endl
       << "outfile1 and all other columns goes into outfile2. MarkerNames and positions" << std::endl
       << "metadata will also be split. All other metadata is copied."                   << std::endl;
}



int main(int argc, char **argv)
{
    po::options_description visible("Allowed options");
    po::options_description hidden("Allowed options");
    po::options_description all("Allowed options");
    po::variables_map vm;
    visible.add_options()
        ("help,h", "Produce help message.")
        //("verbose,v","Turns on verbose output.")
        ;
    hidden.add_options()
        ("filename",
         po::value<std::string>(&options::filename),
         "Input file.")
        ("splitIdx",
         po::value<int>(&options::splitIdx),
         "Where to split file.")
        ("outFilename1",
         po::value<std::string>(&options::outFilename1),
         "Output file 1.")
        ("outFilename2",
         po::value<std::string>(&options::outFilename2),
         "Output file 2.")
        ;
    all.add(visible).add(hidden);

    po::positional_options_description p;
    p.add("filename", 1);
    p.add("splitIdx", 1);
    p.add("outFilename1", 1);
    p.add("outFilename2", 1);

    try
    {
        po::store(po::command_line_parser(argc, argv)
                  .options(all).positional(p).run(), 
                  vm, false);
        po::notify(vm);
    }
    catch (std::exception &e)
    {
        std::cerr << "Error parsing options: " << e.what()
                  << std::endl;
        printUsage(argv[0], std::cerr, visible);
        return EXIT_FAILURE;
    }

    if (vm.count("help"))
    {
        printUsage(argv[0], std::cout, visible);
        return EXIT_SUCCESS;
    }



    // checking options
    if (options::filename.empty())
	{
  	    std::cerr << "Error: No inputfile specified."
                  << std::endl << std::endl;
	    printUsage(argv[0], std::cerr, visible);
	    return EXIT_FAILURE;
	}
    if (options::splitIdx == -1)
	{
  	    std::cerr << "Error: No column to split on specified."
                  << std::endl << std::endl;
	    printUsage(argv[0], std::cerr, visible);
	    return EXIT_FAILURE;
	}
    if (options::outFilename1.empty() || options::outFilename2.empty())
	{
  	    std::cerr << "Error: No output file specified."
                  << std::endl << std::endl;
	    printUsage(argv[0], std::cerr, visible);
	    return EXIT_FAILURE;
	}



    try
    {
        SNPFile inFile(options::filename);
        if(!(0 <= options::splitIdx && options::splitIdx <= inFile.noCols()))
        {
            std::cerr << "Error: Index to split on out of range. "
                      << "Should be in [" << 0 << ", " << inFile.noCols() << "["
                      << " and is " << options::splitIdx << "." << std::endl;
            return EXIT_FAILURE;
        }

        SNPFile outFile1(options::outFilename1, true, true);
        SNPFile outFile2(options::outFilename2, true, true);


        splitVectorMetadata<int>(inFile, options::splitIdx, "positions", outFile1, outFile2);
        splitVectorMetadata<std::string>(inFile, options::splitIdx, "markerNames", outFile1, outFile2);


        MetadataKeyIterator it = inFile.metadataKeysBegin();
        for(; it != inFile.metadataKeysEnd(); ++it)
            if(*it != "positions" && *it != "markerNames")
                copyMetadata(inFile, *it, outFile1, outFile2);

        splitMatrix(inFile, options::splitIdx,
                    outFile1, outFile2);

        inFile.close();
        outFile1.close();
        outFile2.close();
    }
    catch(std::exception &e)
    {
        std::cerr << "Error: " << e.what() << std::endl;
        return EXIT_FAILURE;
    }



    return EXIT_SUCCESS;
}
