/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2008 by Bioinformatics Research Center,
 *                       University of Aarhus, Denmark
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */



#include "iterators.hh"
#include "matrix.hh"

#include <boost/scoped_ptr.hpp>
#include <vector>
#include <map>



/** \file
 * \brief Tabulate a matrix.
 */



namespace BiRC
{
    namespace SNPFile
    {
        //XXXXXXXXXXXXXXXXXXXX Convenience definitions XXXXXXXXXXXXXXXXXXXX
        /// The type of the computed counts.
        typedef unsigned int count_t;



        //XXXXXXXXXXXXXXXXXXXX Categorizers XXXXXXXXXXXXXXXXXXXX
        ///Categorizes cells based on the (genotype) value stored in them.
        struct GenotypeCategorizer
        {
            typedef allele_t category_type;

            allele_t operator() (int row, int col, allele_t val) const
            {
                return val;
            }
        };



        ///Categorizes cells based on the (haplotype) value stored in them.
        typedef GenotypeCategorizer HaplotypeCategorizer;



        ///Categorizes cells based on what row it is in.
        struct RowCategorizer
        {
            typedef int category_type;

            int operator() (int row, int col, allele_t val) const
            {
                return row;
            }
        };



        ///Categorizes cells based on what column it is in.
        struct ColumnCategorizer
        {
            typedef int category_type;

            int operator() (int row, int col, allele_t val) const
            {
                return col;
            }
        };



        /**
         * \brief Trait describing how to use a given type as a transform.
         *
         * A TransformUsageTrait should define typedefs for from_type and to_type
         * for the types transformed from and to respectively. Besides that i should
         * implement a function:<br/>
         * static to_type use(const transform_type &transform, const from_type &from)<br/>
         * which does the actual transformation given the transform and the data to transform.
         */
        template<typename Transform>
        struct TransformUsageTrait
        {
        };



        ///Handle vector: Map index to value at index.
        template<typename T>
        struct TransformUsageTrait< std::vector<T> >
        {
            typedef typename std::vector<T> transform_type;
            typedef typename transform_type::size_type from_type;
            typedef typename transform_type::value_type to_type;

            static to_type use(const transform_type &transform,
                               const from_type &from)
            {
                return transform[from];
            }
        };



        ///Handle map: Map key to value.
        template<typename K, typename T>
        struct TransformUsageTrait< std::map<K, T> >
        {
            typedef typename std::map<K, T> transform_type;
            typedef K from_type;
            typedef T to_type;

            static to_type use(const transform_type &transform,
                               const from_type &from)
            {
                return transform.find(from)->second;
            }
        };



        ///Handle function: Map value by applying function.
        template<typename FromType, typename ToType>
        struct TransformUsageTrait<ToType (*)(FromType)>
        {
            typedef ToType (*transform_type)(FromType);
            typedef FromType from_type;
            typedef ToType to_type;

            static to_type use(const transform_type &transform,
                               const from_type &from)
            {
                return (*transform)(from);
            }
        };



        ///Handle & function: Map value by applying function.
        template<typename FromType, typename ToType>
        struct TransformUsageTrait<ToType (*)(FromType &)>
        {
            typedef ToType (*transform_type)(FromType &);
            typedef FromType from_type;
            typedef ToType to_type;

            static to_type use(const transform_type &transform,
                               const from_type &from)
            {
                return (*transform)(from);
            }
        };



        ///Handle pointer: Lookup in array. Map index to value at index.
        template<typename T>
        struct TransformUsageTrait< T* >
        {
            typedef T* transform_type;
            typedef size_t from_type;
            typedef T to_type;

            static to_type use(const transform_type &transform,
                               const from_type &from)
            {
                return transform[from];
            }
        };



        ///Categorizes a cell by transforming the category it is already in by some other definition.
        template<typename Categorizer,
                 typename Transform>
        class CategoryTransformation
        {
            typedef TransformUsageTrait<Transform> Usage;
        public:
            typedef typename Usage::to_type category_type;

        private:
            const Categorizer &categorizer; 
            const Transform &transform;

        public:
            CategoryTransformation(const Categorizer &categorizer,
                                   const Transform &transform)
                : categorizer(categorizer),
                  transform(transform)
            {
            }

            category_type operator() (int row, int col, int val) const
            {
                return Usage::use(transform, categorizer(row, col, val));
            }
        };



        ///Categorizes a cell by combining the result of two other categorizations.
        template<typename RowCategorizer, typename ColumnCategorizer>
        class TwoWayCategorizer
        {
        public:
            typedef std::pair<typename RowCategorizer::category_type,
                              typename ColumnCategorizer::category_type> category_type;

        private:
            boost::scoped_ptr<RowCategorizer>    rowCategorizerInstance;
            boost::scoped_ptr<ColumnCategorizer> columnCategorizerInstance;
            const RowCategorizer    &rowCategorizer;
            const ColumnCategorizer &columnCategorizer;

        public:
            TwoWayCategorizer(const RowCategorizer &rowCategorizer, const ColumnCategorizer &columnCategorizer)
                : rowCategorizerInstance(0),
                  columnCategorizerInstance(0),
                  rowCategorizer(rowCategorizer),
                  columnCategorizer(columnCategorizer)
            {
            }

            TwoWayCategorizer()
                : rowCategorizerInstance(new RowCategorizer()),
                  columnCategorizerInstance(new ColumnCategorizer()),
                  rowCategorizer(*(rowCategorizerInstance.get())),
                  columnCategorizer(*(columnCategorizerInstance.get()))
            {
            }

            category_type operator() (int row, int col, allele_t val) const
            {
                return std::make_pair(rowCategorizer(row, col, val), columnCategorizer(row, col, val));
            }
        };



        //XXXXXXXXXXXXXXXXXXXX Functions XXXXXXXXXXXXXXXXXXXX
        ///Counts how many cells in each category.
        template<typename Categorizer>
        std::map<typename Categorizer::category_type, count_t>
        tabulate(const MatrixView &m, const Categorizer &categorizer)
        {
            typename std::map<typename Categorizer::category_type, count_t> retVal;

            for(int col = 0; col < m.noCols(); col++)
            {
                for(int row = 0; row < m.noRows(); row++)
                    retVal[categorizer(row, col, m(row, col))]++;
            }

            return retVal;
        }



        ///Counts how many cells in each field in the NxM table given by the two categorizers.
        template<typename RowCategorizer, typename ColumnCategorizer>
        std::map<std::pair<typename RowCategorizer::category_type,
                           typename ColumnCategorizer::category_type>,
                 count_t>
        tabulate(const MatrixView &m, const RowCategorizer &rowCat, const ColumnCategorizer &colCat)
        {
            const TwoWayCategorizer<RowCategorizer, ColumnCategorizer> categorizer(rowCat, colCat);
            return tabulate(m, categorizer);
        }



        ///Transforms one categorizer by a given transformation.
        template<typename Categorizer,
                 typename Transform>
        CategoryTransformation<Categorizer, Transform> transform(const Categorizer &categorizer,
                                                                 const Transform &transform)
        {
            return CategoryTransformation<Categorizer, Transform>(categorizer, transform);
        }



        ///Counts how many of each genotype in matrix.
        inline
        std::map<allele_t, count_t>
        countGenotypes(const MatrixView &m)
        {
            return tabulate(m, GenotypeCategorizer());
        }



        ///Counts how many of each genotype for each category of rows.
        template<typename T>
        std::map<std::pair<typename TransformUsageTrait<T>::to_type, allele_t>, count_t>
        countGenotypesForSampleCategory(const MatrixView &m, const T &t)
        {
            return tabulate(m, CategoryTransformation<RowCategorizer, T>(RowCategorizer(), t), GenotypeCategorizer());
        }



        ///Counts how many of each genotype for each category of columns.
        template<typename T>
        std::map<std::pair<typename TransformUsageTrait<T>::to_type, allele_t>, count_t>
        countGenotypesForSNPCategory(const MatrixView &m, const T &t)
        {
            return tabulate(m, CategoryTransformation<ColumnCategorizer, T>(ColumnCategorizer(), t), GenotypeCategorizer());
        }
    }
}
