/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2008 by Bioinformatics Research Center,
 *                       University of Aarhus, Denmark
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */



#undef NDEBUG			// enable asserts always

#include "array_matrix.hh"
#include "tabulate.hh"



using namespace BiRC::SNPFile;




bool isMissingFunc1(int val)
{
    return val == MISSING;
}




bool isMissingFunc2(const int val)
{
    return val == MISSING;
}




bool isMissingFunc3(const int &val)
{
    return val == MISSING;
}



int main()
{
    Matrix m = newArrayMatrix(3, 4);
    for(int i = 0; i < m.noCols(); i++)
    {
        for(int j = 0; j < m.noRows(); j++)
            m(j, i) = (j * m.noCols() + i + 1) % 3;
    }

    m(0, 1) = MISSING;
    m(1, 0) = MISSING;
    m(2, 1) = MISSING;

    MatrixView v = m.view();

    //std::cout << m << std::endl;

    std::map<allele_t, count_t> genotypeCounts = countGenotypes(v);
    assert(genotypeCounts[HOMOZYGOTE_0] == 4);
    assert(genotypeCounts[HOMOZYGOTE_1] == 3);
    assert(genotypeCounts[HETEROZYGOTE] == 2);
    assert(genotypeCounts[MISSING] == 3);
    
    std::map<std::pair<int, allele_t>, count_t> counts2 = tabulate(v, RowCategorizer(), GenotypeCategorizer());
    assert(counts2[std::make_pair(0, HOMOZYGOTE_1)] == 2);
    assert(counts2[std::make_pair(1, MISSING)] == 1);
    assert(counts2[std::make_pair(2, HETEROZYGOTE)] == 1);
    
    std::map<std::pair<int, allele_t>, count_t> counts3 = tabulate(v, ColumnCategorizer(), GenotypeCategorizer());
    assert(counts3[std::make_pair(0, HETEROZYGOTE)] == 0);
    assert(counts3[std::make_pair(1, MISSING)] == 2);
    assert(counts3[std::make_pair(2, MISSING)] == 0);
    assert(counts3[std::make_pair(3, HOMOZYGOTE_0)] == 1);
    
    std::vector<std::string> gender;
    gender.push_back("boy");
    gender.push_back("boy");
    gender.push_back("girl");
    
    std::map<std::pair<std::string, allele_t>, count_t> counts4 = countGenotypesForSampleCategory(v, gender);
    assert(counts4[std::make_pair("boy", MISSING)] == 2);
    assert(counts4[std::make_pair("girl", MISSING)] == 1);
    
    int markerNum[] = {31, 42, 43, 67};
    
    std::map<std::pair<int, allele_t>, count_t> counts5 = countGenotypesForSNPCategory(v, static_cast<int *>(markerNum));
    assert(counts5[std::make_pair(31, HOMOZYGOTE_0)] == 1);
    assert(counts5[std::make_pair(31, HOMOZYGOTE_1)] == 1);
    assert(counts5[std::make_pair(31, HETEROZYGOTE)] == 0);
    assert(counts5[std::make_pair(31, MISSING)]      == 1);

    std::map<int, bool> isMissing;
    isMissing[HOMOZYGOTE_0] = false;
    isMissing[HOMOZYGOTE_1] = false;
    isMissing[HETEROZYGOTE] = false;
    isMissing[MISSING] = true;
    std::map<bool, count_t> counts6 = tabulate(v, transform(GenotypeCategorizer(), isMissing));
    assert(counts6[true] == 3);
    assert(counts6[false] == 9);

    std::map<bool, count_t> counts7 = tabulate(v, transform(GenotypeCategorizer(), &isMissingFunc1));
    assert(counts7[true] == 3);
    assert(counts7[false] == 9);

    std::map<bool, count_t> counts8 = tabulate(v, transform(GenotypeCategorizer(), &isMissingFunc2));
    assert(counts8[true] == 3);
    assert(counts8[false] == 9);

    std::map<bool, count_t> counts9 = tabulate(v, transform(GenotypeCategorizer(), &isMissingFunc3));
    assert(counts9[true] == 3);
    assert(counts9[false] == 9);
}
