/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus, Denmark
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#undef NDEBUG			// enable asserts always

#include "matrix.hh"
#include "file_matrix.hh"
#include "byte_order_handler.hh"
#include "snpfile.hh"

#include "test_util.hh"

#include <iostream>
#include <sstream>
#include <cassert>
#include <cstring>


using namespace BiRC::SNPFile;



int
main()
{
    try
    {
        const char *filename = "test_snpfile.snp";

        const allele_t matrixTestData[] = {1, 2, 3, 5, 7, 42, 14};
        const size_t matrixTestDataLength = sizeof(matrixTestData) / sizeof(allele_t);

        const size_t nRepeats = 4;

        const std::string key0 = "Key 0";
        const std::string key1 = "Key 1";
        const std::string key2 = "Key 2";
        const char metadata0[] = "Hello world"; // 12 bytes.
        const char metadata1[] = "foo"; // 4 bytes.
        const char metadata2[] = "bar"; // 4 bytes.



        //Low-level stuff:

        SNPFile file(filename);
        const SNPFile &constFile = file;

        assert(size_t(file.noCols()) == nRepeats);
        assert(size_t(file.noRows()) == matrixTestDataLength);
        for(size_t i = 0; i < nRepeats - 1; i++)
        {
            const allele_t *dataPtr = constFile.accessMatrixData();
            assert(memcmp(dataPtr + matrixTestDataLength*i,
                          matrixTestData,
                          sizeof(matrixTestData)) == 0);
        }

        assert(file.containsMetadata(key0));
        assert(file.containsMetadata(key1));
        assert(file.containsMetadata(key2));

        MetadataKeyIterator keyIt = file.metadataKeysBegin();

        bool key0Found = false;
        bool key1Found = false;
        for( ; keyIt != file.metadataKeysEnd(); keyIt++ ) {
            if(*keyIt == key0)
                key0Found = true;
            else if(*keyIt == key1)
                key1Found = true;
        }
        assert(key0Found && key1Found);       
        
        assert(file.getMetadataSize(key0) == sizeof(metadata0));
        assert(file.getMetadataSize(key1) == sizeof(metadata1));
        assert(file.getMetadataSize(key2) == sizeof(metadata2));

        assert(strcmp(constFile.getMetadataBuffer(key0).data(), metadata0) == 0);
        assert(strcmp(constFile.getMetadataBuffer(key1).data(), metadata1) == 0);
        assert(strcmp(constFile.getMetadataBuffer(key2).data(), metadata2) == 0);



        //Matrix stuff:
	ImmutableMatrix matrix(newReadOnlyFileMatrix(file));

        for(size_t i = 0; i < matrixTestDataLength; i++)
            assert(matrix(i, matrix.noCols() - 1) == allele_t(40 + i));

        //FIXME: matrix.readHiding();
        MatrixView view = matrix.view();

        std::ostringstream stringStream;
        std::string expected;

        expected =
            "1 1 40 \n"
            "2 2 41 \n"
            "3 3 42 \n"
            "42 42 45 \n"
            "14 14 46 \n";

        stringStream << view;

        assert(expected == stringStream.str());
        stringStream.str("");

    }
    catch(std::runtime_error &ex)
    {
        std::cout << ex.what() << std::endl;
        assert(false);
    }

    return 0;
}
