/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus, Denmark
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include "snpfile.hh"
#include "matrix.hh"
#include "file_matrix.hh"
#include "metadata_access.hh"

using namespace BiRC::SNPFile;

#include <cstdlib>
#include <cerrno>
#include <fstream>
#include <iostream>
#include <boost/program_options.hpp>
#include <boost/lexical_cast.hpp>




/// Exception thrown if positions read in are not in increasing order.
struct UnsortedPositions : std::logic_error {
    UnsortedPositions() : std::logic_error("Positions must be sorted.") {}
};

/// Exception thrown if loading haplotypes failed.
struct ParseError : std::logic_error {
    ParseError() : std::logic_error("Error parsing haplotype data.") {}
};

/// Exception thrown if a requested file could not be opened.
struct UnableToOpenFile : std::logic_error {
    UnableToOpenFile() : std::logic_error("Unable to open file.") {}
};


namespace {
    template <typename Itr>
    bool isSorted(Itr begin, Itr end)
    {
        if (begin+1 >= end) return true;
        for (Itr i = begin; i < end-1; ++i)
            if (*i > *(i+1)) return false;
        return true;
    }
}

static std::vector<int> 
readPositions(std::istream &is)
{
    std::vector<int> result;
    double floatPos;
    int    intPos;

    while (is >> floatPos) {
        intPos = static_cast<int>(floatPos);
        if (intPos != floatPos)
            std::cerr << "Warning: support for floating point positions is "
                      << "deprecated, rounding " << floatPos << " to "
                      << intPos << '.' << std::endl;
        result.push_back(intPos);
    }
    if (!isSorted(result.begin(), result.end())) throw UnsortedPositions();

    return result;
}

std::vector<int> 
readPositions(const char *fname)
{
    std::ifstream file(fname);
    if (!file) throw UnableToOpenFile();
    return readPositions(file);
}


namespace {
    void split(const std::string& str, std::vector<std::string>& tokens)
    {
        const std::string& delimiters = " \t";
        std::string::size_type lastPos = str.find_first_not_of(delimiters, 0);
        std::string::size_type pos = str.find_first_of(delimiters, lastPos);

        while (std::string::npos != pos || std::string::npos != lastPos)
	    {
            tokens.push_back(str.substr(lastPos, pos - lastPos));
            lastPos = str.find_first_not_of(delimiters, pos);
            pos = str.find_first_of(delimiters, lastPos);
	    }
    }

    template<typename toType> void castLine(const std::vector<std::string>::const_iterator from,
                                            const std::vector<std::string>::const_iterator to,
                                            std::vector<toType> &dest)
    {
        for(std::vector<std::string>::const_iterator i = from; i != to; ++i)
            dest.push_back(boost::lexical_cast<toType>(*i));
    }
}


template<typename Phenotype>
static void
readHaplotypes(std::istream &is, Matrix &matrix, bool unphased, std::vector<Phenotype> &values)
{
    std::vector< std::vector<int> > haplotypes;
    std::vector<Phenotype> tmp_values;
    while(is)
	{
	    std::string line; 
	    std::vector<std::string> tokens;

	    getline(is, line);
	    split(line, tokens);

	    if(tokens.size() == 0) continue; // empty line

	    std::vector<int> haplotype;

	    castLine(tokens.begin() + 1, tokens.end(), 
                 haplotype);

	    tmp_values.push_back(boost::lexical_cast<Phenotype>(tokens[0]));

	    if (haplotypes.size() > 0)
            if (haplotypes[0].size() != haplotype.size())
                throw ParseError();

	    haplotypes.push_back(haplotype);
	}

    if (haplotypes.empty()) throw ParseError();

    int noHaplotypes = (int)haplotypes.size();
    int noMarkers    = (int)haplotypes[0].size();

    int sizeMultiplier = unphased ? 1 : 2;

    matrix.resize(noHaplotypes, noMarkers);
    values.resize(noHaplotypes/sizeMultiplier);
    for (int i = 0; i < noHaplotypes/sizeMultiplier; ++i)
        values[i] = tmp_values[sizeMultiplier*i];

    for (int i = 0; i < noHaplotypes; ++i)
	{
	    for (int j = 0; j < noMarkers; ++j)
            matrix(i,j) = haplotypes[i][j];
	}

    matrix.setPhased(!unphased);
}


static std::vector<std::string>
readMarkerNames(std::istream &is)
{
    std::vector<std::string> result;
    while (is) {
        std::string name;
        is >> name;
        if (!name.empty()) result.push_back(name);
    }
    return result;
}

namespace options {
    bool verbose = false;
    std::string positionsFileName;
    std::string haplotypesFileName;
    std::string markerNamesFileName;
    std::string outputFileName;
    bool  quantitative = false;
    bool  unphased = true;
}


namespace po = boost::program_options;
static void printUsage(const char *progname,
                       std::ostream &os,
                       po::options_description &options)
{
    os << "Usage: " << progname
       << " [options] positions genotypes output"
       << std::endl
       << options << std::endl;
}

int main(int argc, char** argv)
{
    po::options_description visible("Allowed options");
    po::options_description hidden("Allowed options");
    po::options_description all("Allowed options");
    po::variables_map vm;
    visible.add_options()
        ("help,h", "Produce help message.")
        ("verbose,v","Turns on verbose output.")
	
        ("markers,m",
         po::value<std::string>(&options::markerNamesFileName),
         "File containing a list of space-separated marker names.")

        ("quantitative,q", "QTL mapping (default false).")

        ("unphased,u", 
         po::value<bool>(&options::unphased)->default_value(true),
         "Input is unphased (default true).")
        ;
    hidden.add_options()
        ("positions",
         po::value<std::string>(&options::positionsFileName),
         "File containing marker positions.")
        ("genotypes",
         po::value<std::string>(&options::haplotypesFileName),
         "File containing marker data.")
        ("output",
         po::value<std::string>(&options::outputFileName),
         "Output file.")
        ;
    all.add(visible).add(hidden);

    po::positional_options_description p;
    p.add("positions", 1);
    p.add("genotypes", 1);
    p.add("output", 1);

    try {
        po::store(po::command_line_parser(argc, argv)
                  .options(all).positional(p).run(), 
                  vm, false);
        po::notify(vm);
    } catch (std::exception &e) {
        std::cerr << "Error parsing options: " << e.what()
                  << std::endl;
        printUsage(argv[0], std::cerr, visible);
        return EXIT_FAILURE;
    }

    if (vm.count("help")) {
        printUsage(argv[0], std::cout, visible);
        return EXIT_SUCCESS;
    }
    options::verbose = vm.count("verbose") > 0;
    options::quantitative = vm.count("quantitative") > 0;

    // checking options
    if (options::positionsFileName.empty())
	{
  	    std::cerr << "Error: No position file specified."
                  << std::endl << std::endl;
	    printUsage(argv[0], std::cerr, visible);
	    return EXIT_FAILURE;
	}
    if (options::haplotypesFileName.empty())
	{
  	    std::cerr << "Error: No haplotype file specified."
                  << std::endl << std::endl;
	    printUsage(argv[0], std::cerr, visible);
	    return EXIT_FAILURE;
	}
    if (options::outputFileName.empty())
	{
  	    std::cerr << "Error: No output file specified."
                  << std::endl << std::endl;
	    printUsage(argv[0], std::cerr, visible);
	    return EXIT_FAILURE;
	}


    // checking files
    std::ifstream positionsFile(options::positionsFileName.c_str());
    if (!positionsFile)
	{
	    std::cerr << "Error opening " 
                  << options::positionsFileName << std::endl;
	    return EXIT_FAILURE;
	}
    std::ifstream haplotypesFile(options::haplotypesFileName.c_str());
    if (!haplotypesFile)
	{
	    std::cerr << "Error opening "
                  << options::haplotypesFileName << std::endl;
	    return EXIT_FAILURE;
	}
    std::auto_ptr<SNPFile> outputFile;
    try {
        SNPFile *ofile = new SNPFile(options::outputFileName, true);
        outputFile = std::auto_ptr<SNPFile>(ofile);
    } catch (std::exception &e) {
        std::cerr << "Error accessing " << options::outputFileName
                  << " : " << e.what() << std::endl;
        return EXIT_FAILURE;
    }
    
    try
    {
        std::vector<int> positions = readPositions(positionsFile);
        std::vector<std::string> markerNames;
        if (options::markerNamesFileName.empty())
        {
            using boost::lexical_cast;
            for (size_t i = 0; i < positions.size(); ++i)
                markerNames.push_back("marker"+lexical_cast<std::string>(i));
        }
        else
        {
            std::ifstream markersFile(options::markerNamesFileName.c_str());
            if (!markersFile)
            {
                std::cerr << "Error opening " 
                          << options::markerNamesFileName
                          << std::endl;
                return EXIT_FAILURE;
            }
            markerNames = readMarkerNames(markersFile);
        }


        setMetadata(*outputFile, "positions", positions);
        setMetadata(*outputFile, "markerNames", markerNames);


        Matrix matrix(newReadWriteFileMatrix(*outputFile));

        if(options::quantitative)
        {
            std::vector<double> values;
            readHaplotypes(haplotypesFile, matrix, options::unphased, values);
            setMetadata(*outputFile, "phenotype", values);
        }
        else
        {
            std::vector<bool> status;
            readHaplotypes(haplotypesFile, matrix, options::unphased, status);
            setMetadata(*outputFile, "phenotype", status);
        }
        setMetadata(*outputFile, "quantitative", options::quantitative);
        setMetadata(*outputFile, "case/control", !options::quantitative);

        outputFile->close();

    }
    catch(std::exception &e)
    {
        std::cerr << "Error: " << e.what() << std::endl;

        if(unlink(options::outputFileName.c_str()) == -1 && errno != ENOENT)
            std::cerr << "Unable to remove file "
                      << options::outputFileName
                      << std::endl;
    }



    return EXIT_SUCCESS;
}
