/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2008 by Bioinformatics Research Center,
 *                       University of Aarhus, Denmark
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */
#include "timing.hh"

#include <string>
#include <iomanip>
#include <fstream>



namespace BiRC
{
    namespace SNPFile
    {
        TimePoint::TimePoint()
        {
            std::ifstream procStat("/proc/stat");
            std::string cpu;
            procStat >> cpu >> user >> nice >> system >> idle >> io >> irq >> softIrq;
            procStat.close();
        }



        TimeInterval operator- (const TimePoint &to, const TimePoint &from)
        {
            return TimeInterval(from, to);
        }



        TimeInterval::TimeInterval(const TimePoint &from, const TimePoint &to)
            : dUser(to.user - from.user),
              dNice(to.nice - from.nice),
              dSystem(to.system - from.system),
              dIdle(to.idle - from.idle),
              dIo(to.io - from.io),
              dIrq(to.irq - from.irq),
              dSoftIrq(to.softIrq - from.softIrq)
        {
        }



        namespace
        {
            std::ostream &maybePrintTime(std::ostream &out, const std::string description, double totalTime, double time)
            {
                double fraction = 100 * time / totalTime;
                if(fraction > 0.5)
                {
                    out << std::left << std::setw(30) << (description + " time :")
                        << std::right << std::setw(8) << time << "s. "
                        << "(" << std::setprecision(3) << fraction << "%)\n";
                }

                return out;
            }
        }



        std::ostream &operator<< (std::ostream &out, const TimeInterval &ti)
        {
            double totalTime
                = ti.getUserTime()
                + ti.getNiceTime()
                + ti.getSystemTime()
                //+ ti.getIdleTime()
                + ti.getIOWaitTime()
                + ti.getHarwareIRQTime()
                + ti.getSoftwareIRQTime();

            maybePrintTime(out, "User", totalTime, ti.getUserTime());
            maybePrintTime(out, "Low priority user (nice)", totalTime, ti.getNiceTime());
            maybePrintTime(out, "System", totalTime, ti.getSystemTime());
            //maybePrintTime(out, "Idle", totalTime, ti.getIdleTime());
            maybePrintTime(out, "Waiting for I/O", totalTime, ti.getIOWaitTime());
            maybePrintTime(out, "Waiting for HW IRQ", totalTime, ti.getHarwareIRQTime());
            maybePrintTime(out, "Waiting for SW IRQ", totalTime, ti.getSoftwareIRQTime());
            out << "Total " << totalTime << "s.\n";

            return out;
        }
    }
}
