/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2007, 2008
                 Bioinformatics Research Center,
 *               University of Aarhus, Denmark
 *               Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#ifndef TYPETRAIT_HH_INCLUDED
#define TYPETRAIT_HH_INCLUDED

#include <string>
#include <vector>
#include <map>

/**
 * \file
 *
 * Definition of typetrait for giving a string-representation of the name of a type.
 */

namespace BiRC
{
    namespace SNPFile
    {
        ///The typename of a unknown type.
        const std::string UNKNOWN_TYPE_NAME = "unknown";

        /// Default definition of TypeTrait. Sets typename to unknown.
        template <typename T>
        struct TypeTrait {
            ///Returns the typename of the type T, which is unknown per default.
            static const std::string typeName() { return UNKNOWN_TYPE_NAME; }
        };

        /// TypeTrait for a std::vector.
        template <typename T>
        struct TypeTrait < std::vector<T> >
        {
            ///Returns the typename for a vector, which is "std::vector< T >".
            static const std::string typeName() { 
                return std::string("std::vector< ")
                    + TypeTrait<T>::typeName() 
                    + std::string(" >"); 
            }
        };

        /// TypeTrait for a std::map.
        template <typename T, typename K>
        struct TypeTrait < std::map<T, K> >
        {
            ///Returns the typename for a map, which is "std::map< K, T >".
            static const std::string typeName() { 
                return 
                    std::string("std::map< ")
                    + TypeTrait<T>::typeName() 
                    + std::string(", ")
                    + TypeTrait<K>::typeName()
                    + std::string(" >"); 
            }
        };



        ///Convenience function for getting the name of a type.
        template<typename T>
        inline std::string typeName()
        {
            return TypeTrait<T>().typeName();
        }

        ///Convenience function for testing if a string is the typename of a given type.
        template<typename T>
        inline bool isTypeName(const std::string &s)
        {
            return s == typeName<T>();
        }
    }
}

/**
 * \brief Convenience macro for defining a typetrait for a given type.
 *
 * This macro must be used in the global namespace.
 * It simply sets the typename of type T to "T".
 */
#define BIRC_SNPFILE_EXPORT_TYPE(T)                                 \
    namespace BiRC { namespace SNPFile {                            \
            template <> struct TypeTrait<T> {                       \
            static const std::string typeName() { return #T; }      \
            };                                                      \
        }}


BIRC_SNPFILE_EXPORT_TYPE(bool)

BIRC_SNPFILE_EXPORT_TYPE(int8_t)
BIRC_SNPFILE_EXPORT_TYPE(int16_t)
BIRC_SNPFILE_EXPORT_TYPE(int32_t)
BIRC_SNPFILE_EXPORT_TYPE(int64_t)

BIRC_SNPFILE_EXPORT_TYPE(uint8_t)
BIRC_SNPFILE_EXPORT_TYPE(uint16_t)
BIRC_SNPFILE_EXPORT_TYPE(uint32_t)
BIRC_SNPFILE_EXPORT_TYPE(uint64_t)

BIRC_SNPFILE_EXPORT_TYPE(float)
BIRC_SNPFILE_EXPORT_TYPE(double)
BIRC_SNPFILE_EXPORT_TYPE(long double)

BIRC_SNPFILE_EXPORT_TYPE(std::string)


#endif
